/** ********************************************************************************************************************
 * @file        main.c
 * @author      Briowireless, Inc.
 * @date        2016-10-28
 * @brief       Main program file for a sample that uses all the BitPipe API.
 * @version     API V1
 * @copyright   Briowireless, Inc., 2016, Licensed under BSD license. See LICENSE file for more details.
 *
 ** *******************************************************************************************************************/

// Include directives  -------------------------------------------------------------------------------------------------

#include <bp_api.h>
#include <../platform.h>                // Depends on the include path. Example: -I../../platforms/win32/
#include <../sample_common.h>           // Common code for all samples.


// Preprocessor definitions --------------------------------------------------------------------------------------------

#define UART_PORT_BITPIPE                   "\\\\.\\COM11"      // Port on which the BitPipe is found (for win32 platform_init).

#define SIM_PIN                             ""                  // When SIM card is locked using a PIN
#define SIM_PUK                             ""                  // When too many bad PINs were tried, the PIN Unlock Key is necessary.

#define APN_CONFIG_HOST                     "m2minternet.apn"
#define APN_CONFIG_USER                     ""
#define APN_CONFIG_PASS                     ""

#define AUTONOMOUS_MODE_CFG_SECURE          0                   // Not secure.
#define AUTONOMOUS_MODE_CFG_KA_IVAL         0                   // Keep-alive interval => Disabled.
#define AUTONOMOUS_MODE_CFG_PORT            1883
#define AUTONOMOUS_MODE_CFG_HOST            "104.196.16.233"
#define AUTONOMOUS_MODE_CFG_USER            ""
#define AUTONOMOUS_MODE_CFG_PASS            ""
#define AUTONOMOUS_MODE_CFG_GPFX            ""                  // Global prefix.

#define HTTP_REMOTE_HOST                    "httpbin.org"
#define HTTP_REMOTE_PORT                    80
#define HTTP_REMOTE_SECURE                  BP_API_IP_HTTP_INIT_CONNECTION__SECURE_NO
#define HTTP_PATH_QUERY                     "/post"
#define HTTP_METHOD                         BP_API_IP_HTTP_INIT_CONNECTION__METHOD_POST
#define HTTP_ADD_HEADERS                    "MyCustomHeader: A simple value\\0d\\0aMyOtherHeader: AnotherValue"
#define HTTP_POST_DATA                      "This data was posted to the web server, and replied back (because that's what the server at httpbin.org does when POSTing to /post)"
#define HTTP_POST_CONTENT_LEN               (sizeof(HTTP_POST_DATA)-1)      // Only used when method is POST

// -------- MQTT: If you want a public MQTT server, check out cloudmqtt.com for trials.
#define MQTT_1_REMOTE_HOST                  "m10.cloudmqtt.com"
#define MQTT_1_REMOTE_PORT                  24952
//#define MQTT_1_REMOTE_PORT                  14952
//#define MQTT_1_SECURE                       0
#define MQTT_1_SECURE                       1
#define MQTT_1_CLEAN_SESSION                1
#define MQTT_1_CLIENT_ID                    "MqttSample__Cli1"
#define MQTT_1_USER                         "secret user name"
#define MQTT_1_PASS                         "secret password"
#define MQTT_1_KA_IVAL                      300/5
#define MQTT_1_SUB_TOP1                     "c1/1"
#define MQTT_1_SUB_TOP2                     "c1/2/#"
#define MQTT_1_SUB_TOP3                     "c1/3/+/good"
#define MQTT_1_PUBLISH_1_TOPIC              "some/very/interesting/topic"
#define MQTT_1_PUBLISH_1_QOS                0
#define MQTT_1_PUBLISH_1_RETAIN_IND         0
#define MQTT_1_PUBLISH_1_MSG_DATA           "This is the message payload from the BitPipe.\nHurray."

#define SMS_1_RECIPIENT_MSISDN              "5555555555"        // SMS recipient mobile phone number. Do not prefix with '+'.
#define SMS_1_ENCODING                      BP_API_SMS_SETUP_TRANSMISSION__ENCODING_7BITS_GSM0338
#define SMS_1_MESSAGE                       "Eric is not at home.\nHe is currently waiting the subway at the station."


// Local variables  ----------------------------------------------------------------------------------------------------

// Buffers used by SDK.
static uint8_t deser_buf[260 + MESSAGE_PROTOCOL_CONTEXT_SIZE + MESSAGE_PROTOCOL_HEADER_SIZE + MESSAGE_PROTOCOL_FOOTER_SIZE];
static uint8_t ser_buf[260];


// Local functions  ----------------------------------------------------------------------------------------------------

void exit_low_power_mode(void)
{
    uart_write((uint8_t*)"A", 1);           // Send a character to the BitPipe. This will wake it up from low power mode.
}


// ---------------------------------------------------------------------------------------------------------------------

// SAMPLE MAIN ENTRY POINT

// ---------------------------------------------------------------------------------------------------------------------


int main(void)
{
    bp_api_init_params init_params;                 // BitPipe Host SDK initialization parameters.
    mp_seq_nr seq_nr;                               // Request sequence number
    uint8_t status;                                 // Request execution status.
    uint16_t rx_data_len, consumed_len;             // UART read variables.
    uint8_t uart_rx_buf[100];                       // UART read buffer.
    uint32_t wait_radio_ready_iter = 0;
    uint32_t wait_network_regd_iter = 0;
    uint32_t wait_http_post_write_ready_iter = 0;
    uint32_t wait_mqtt_conn_estab_iter = 0;
    uint32_t wait_mqtt_publication_iter = 0;
    uint32_t wait_sms_incoming_iter = 0;
    time_t request_timeout = 0;                     // Used to detect request timeouts.
    time_t sms_power_off_backoff = 0;


    // Initialize Platform.
    status = platform_init(UART_PORT_BITPIPE);
    if(status != 0)
    {
        P_ERR("Failed to initialize platform: Status=%d.\n", status);
        return 1;
    }


    // Initialize SDK layer.
    init_params.app_callbacks =         &sample_callbacks;      // From common sample code (common_sample.h).
    init_params.uart_write =            &uart_write;            // From platform.
    init_params.deserialize_buf =       deser_buf;              // Local to this sample.
    init_params.deserialize_buf_len =   sizeof(deser_buf);      // Local to this sample.
    init_params.serialize_buf =         ser_buf;                // Local to this sample.
    init_params.serialize_buf_len =     sizeof(ser_buf);        // Local to this sample.

    status = bp_api_init(&init_params);
    if(status != 0)
    {
        P_ERR("Failed to initialize BitPipe API SDK: Status=%d.\n", status);
        platform_deinit();
        return 1;
    }


    // Sample state machine.
    while(!sample_state.done)
    {
        // -------------------------------------------------------------------------------------------------------------
        // ---- This could be placed in a separate thread, if your plaform supports threads.                         ---

        // Check if we've received data from the BitPipe, and process it.
        rx_data_len = uart_read(uart_rx_buf, sizeof(uart_rx_buf));
        while(rx_data_len > 0)
        {
            // This MAY/WILL callback into application callbacks.
            // - The function stops processing UART data when a message is found.
            bp_api_process_incoming(uart_rx_buf, rx_data_len, &consumed_len);

            // Might not have processed all UART data if a message was found.
            rx_data_len -= consumed_len;
        }

        // ---                                                                                                       ---
        // -------------------------------------------------------------------------------------------------------------


        if(!sample_state.awaiting_reply)
        {
            switch(sample_state.step)
            {
            case 0:     // Set UART port speed.
                status = bp_api_sys_set_uart_speed(&seq_nr, BP_API_SYS_SET_UART_SPEED__SPEED_115200);
                break;

            case 1:     // Configure Maximum message payload size.
                status = bp_api_sys_set_max_payload_sz(&seq_nr, 260);
                break;

            case 2:     // Get HW version.
                status = bp_api_devi_get_hw_version(&seq_nr);
                break;

            case 3:     // Get FW version.
                status = bp_api_devi_get_fw_version(&seq_nr);
                break;

            case 4:     // Get Product Name
                status = bp_api_devi_get_product_name(&seq_nr);
                break;

            case 5:     // Get VIN / VIO voltages
                status = bp_api_devi_get_vin_vio_voltage(&seq_nr);
                break;

            case 6:     // Get junction temperature
                status = bp_api_devi_get_junction_temp(&seq_nr);
                break;

            case 7:     // Get autonomous mode configuration.
                status = bp_api_devi_get_autonomous_mode_config(&seq_nr);
                break;

            case 8:     // Set autonomous mode configuration.
                status = bp_api_devi_set_autonomous_mode_config(&seq_nr, AUTONOMOUS_MODE_CFG_SECURE,
                                                                AUTONOMOUS_MODE_CFG_KA_IVAL,
                                                                AUTONOMOUS_MODE_CFG_PORT,
                                                                sizeof(AUTONOMOUS_MODE_CFG_HOST)-1, (uint8_t*)AUTONOMOUS_MODE_CFG_HOST,
                                                                sizeof(AUTONOMOUS_MODE_CFG_USER)-1, (uint8_t*)AUTONOMOUS_MODE_CFG_USER,
                                                                sizeof(AUTONOMOUS_MODE_CFG_PASS)-1, (uint8_t*)AUTONOMOUS_MODE_CFG_PASS,
                                                                sizeof(AUTONOMOUS_MODE_CFG_GPFX)-1, (uint8_t*)AUTONOMOUS_MODE_CFG_GPFX);
                break;

            case 9:     // Power-on Radio.
                status = bp_api_mobi_power_on_radio(&seq_nr,
                                                    BP_API_PWRON_RADIO__PWRON_METHOD_DCDC,
                                                    BP_API_PWRON_RADIO__BACKOFF_DISABLED);
                break;

            case 10:    // Get Radio Status.
                status = bp_api_mobi_get_radio_status(&seq_nr);
                break;

            case 11:    // Wait for radio to be ready.
                if(bp_status.radio_ready)
                {
                    if(wait_radio_ready_iter > 0)
                        P_INFO("Power-on-Radio to Ready took %lds\n", time(NULL) - bp_status.pwr_on_start);
                    sample_state.step ++;
                }
                else
                {
                    if(wait_radio_ready_iter == 0)
                        P_INFO("------- Waiting for radio to be ready ...");

                    if((++wait_radio_ready_iter % 50) == 0)
                        P_INFO(".");
                }
                status = 0xFF;
                break;

            case 12:    // Get Radio Status (again).
                status = bp_api_mobi_get_radio_status(&seq_nr);
                break;

            case 13:     // Check SIM card status.
                if(BP_API_RADIO_START_NOTIFY__SIM_NOT_INSERTED == bp_status.sim_status)
                {
                    sample_state.done = 1;
                    P_WARN("  +++No SIM card detected in the BitPipe. Aborting sample.\n");
                }
                else if(BP_API_RADIO_START_NOTIFY__SIM_DENIED_NEED_PIN == bp_status.sim_status)
                {
                    status = bp_api_mobi_set_sim_access_config(NULL, sizeof(SIM_PIN)-1, (uint8_t*)SIM_PIN);
                }
                else if(BP_API_RADIO_START_NOTIFY__SIM_DENIED_NEED_PUK == bp_status.sim_status)
                {
                    status = bp_api_mobi_set_sim_access_config(NULL, sizeof(SIM_PUK)-1, (uint8_t*)SIM_PUK);
                }
                else    // SIM is OK.
                {
                    sample_state.step++;
                    status = 0xFF;
                }
                break;

            case 14:    // Get IMEI
                status = bp_api_mobi_get_imei(&seq_nr);
                break;

            case 15:    // Get ICCID
                status = bp_api_mobi_get_iccid(&seq_nr);
                break;

            case 16:    // Get IMSI
                status = bp_api_mobi_get_imsi(&seq_nr);
                break;

            case 17:    // Get own phone number (only when information present on SIM card).
                status = bp_api_mobi_get_own_phone_number(&seq_nr);
                break;

            case 18:    // Get APN Configuration
                status = bp_api_mobi_get_apn_config(&seq_nr);
                break;

            case 19:    // Set APN Configuration
                status = bp_api_mobi_set_apn_config(&seq_nr,
                                                    sizeof(APN_CONFIG_HOST)-1, (uint8_t*)APN_CONFIG_HOST,
                                                    sizeof(APN_CONFIG_USER)-1, (uint8_t*)APN_CONFIG_USER,
                                                    sizeof(APN_CONFIG_PASS)-1, (uint8_t*)APN_CONFIG_PASS);
                break;

            case 20:    // Get Mobile connection status.
                status = bp_api_mobi_get_mobile_conn_status(&seq_nr);
                break;

            case 21:    // Wait for network registration.
                if(bp_status.network_registered)
                {
                    if(wait_network_regd_iter > 0)
                        P_INFO("\n");
                    sample_state.step ++;
                }
                else
                {
                    if(wait_network_regd_iter == 0)
                        P_INFO("------- Waiting for network registration ...");

                    if((++wait_network_regd_iter % 50) == 0)
                        P_INFO(".");
                }
                status = 0xFF;
                break;

            case 22:    // Get Mobile connection status (again).
                status = bp_api_mobi_get_mobile_conn_status(&seq_nr);
                break;

            case 23:    // Get Mobile date and time
                status = bp_api_mobi_get_mobile_date_time(&seq_nr);
                break;

            case 24:    // Initiate HTTP connection.
                status = bp_api_ip_http_init_connection(&seq_nr, HTTP_REMOTE_SECURE, HTTP_METHOD,
                                                        sizeof(HTTP_REMOTE_HOST)-1, (uint8_t*)HTTP_REMOTE_HOST,
                                                        HTTP_REMOTE_PORT,
                                                        sizeof(HTTP_PATH_QUERY)-1, (uint8_t*)HTTP_PATH_QUERY,
                                                        sizeof(HTTP_ADD_HEADERS)-1, (uint8_t*)HTTP_ADD_HEADERS,
                                                        HTTP_POST_CONTENT_LEN);
                break;

            case 25:    // Wait for HTTP connection to be ready to send POST (Only if method is POST)
                if(HTTP_METHOD == BP_API_IP_HTTP_INIT_CONNECTION__METHOD_POST)
                {
                    if(sample_state.conn_status_http == IP_CONN_STATUS_ESTABLISHED)
                    {
                        if(wait_http_post_write_ready_iter > 0)
                            P_INFO("\n");
                        sample_state.step ++;
                    }
                    else
                    {
                        if(wait_http_post_write_ready_iter == 0)
                            P_INFO("------- Waiting for HTTP POST WRITE READY ...");

                        if((++wait_http_post_write_ready_iter % 50) == 0)
                            P_INFO(".");
                    }
                }
                else
                {
                    // Method is not POST, no need to wait on this notification.
                    // In fact, it will never be sent by the BitPipe.
                    sample_state.step++;
                }
                status = 0xFF;
                break;

            case 26:        // Send POST DATA (Only if method is POST).
                if(HTTP_METHOD == BP_API_IP_HTTP_INIT_CONNECTION__METHOD_POST  &&
                        sample_state.http_post_data_idx < HTTP_POST_CONTENT_LEN)                 // Check if we're done.
                {
                    uint32_t remaining = HTTP_POST_CONTENT_LEN - sample_state.http_post_data_idx;
                    sample_state.http_post_last_pay_len = remaining % 256;   // Can send at most 255 bytes per chunk.
                    uint8_t write_op = (remaining - sample_state.http_post_last_pay_len) > 0 ?
                                BP_API_IP_HTTP_POST_WRITE__WRITEOP_WRITE_TO_BUFFER_AND_SEND :
                                BP_API_IP_HTTP_POST_WRITE__WRITEOP_WRITE_TO_BUFFER_AND_SEND_COMPLETE;

                    status = bp_api_ip_http_post_write(&seq_nr, sample_state.conn_id_http, write_op,
                                                       sample_state.http_post_last_pay_len,
                                                       (uint8_t*)HTTP_POST_DATA + sample_state.http_post_data_idx);
                }
                else    // METHOD is not POST, or all data posted successfully.
                {
                    sample_state.step++;
                    status = 0xFF;
                }
                break;

            case 27:        // Wait for connection to be closed (here, we expect the web server to close the connection)
                if(IP_CONN_STATUS_CLOSED == sample_state.conn_status_http)
                {
                    sample_state.step++; // Method is not POST, move to next step.
                }
                status = 0xFF;
                break;

            case 28:        // MQTT: Initiate connection.
                status = bp_api_mqtt_init_connection(&seq_nr, MQTT_1_SECURE, MQTT_1_CLEAN_SESSION,
                                                     sizeof(MQTT_1_REMOTE_HOST)-1, (uint8_t*)MQTT_1_REMOTE_HOST,
                                                     MQTT_1_REMOTE_PORT,
                                                     sizeof(MQTT_1_CLIENT_ID)-1, (uint8_t*)MQTT_1_CLIENT_ID,
                                                     sizeof(MQTT_1_USER)-1, (uint8_t*)MQTT_1_USER,
                                                     sizeof(MQTT_1_PASS)-1, (uint8_t*)MQTT_1_PASS,
                                                     MQTT_1_KA_IVAL);
                break;

            case 29:        // MQTT: Wait for connection to be established.
                if(sample_state.conn_status_mqtt == IP_CONN_STATUS_CLOSED)
                {
                    P_WARN("MQTT connection closed before being established. Aborting sample.\n");
                    sample_state.done = 1;
                }
                else if(sample_state.conn_status_mqtt == IP_CONN_STATUS_ESTABLISHED)
                {
                    if(wait_mqtt_conn_estab_iter > 0)
                        P_INFO("\n");
                    sample_state.step ++;
                }
                else
                {
                    if(wait_mqtt_conn_estab_iter == 0)
                        P_INFO("------- Waiting for MQTT CONNECTION to be ESTABLISHED ...");

                    if((++wait_mqtt_conn_estab_iter % 50) == 0)
                        P_INFO(".");
                }
                status = 0xFF;
                break;

            case 30:        // MQTT: Subscribe to topics.
                status = bp_api_mqtt_subscribe_topics(&seq_nr, sample_state.conn_id_mqtt, 0, 0, 0,
                                                      sizeof(MQTT_1_SUB_TOP1) -1, (uint8_t*)MQTT_1_SUB_TOP1,
                                                      sizeof(MQTT_1_SUB_TOP2) -1, (uint8_t*)MQTT_1_SUB_TOP2,
                                                      sizeof(MQTT_1_SUB_TOP3) -1, (uint8_t*)MQTT_1_SUB_TOP3);
                break;

            case 31:        // MQTT: Wait for a publication from the server to a topic we subscribed.
                if(sample_state.mqtt_publication_received)
                {
                    if(wait_mqtt_publication_iter > 0)
                        P_INFO("\n");
                    sample_state.step ++;
                }
                else
                {
                    if(wait_mqtt_publication_iter == 0)
                        P_INFO("------- Waiting for publication from MQTT Server ...");

                    if((++wait_mqtt_publication_iter % 100) == 0)
                        P_INFO(".");
                }
                status = 0xFF;
                break;

            case 32:        // MQTT: Do a PUBLISH.
                status = bp_api_mqtt_publish(&seq_nr, sample_state.conn_id_mqtt, MQTT_1_PUBLISH_1_QOS, MQTT_1_PUBLISH_1_RETAIN_IND, 1,
                                             sizeof(MQTT_1_PUBLISH_1_TOPIC) - 1, (uint8_t*)MQTT_1_PUBLISH_1_TOPIC,
                                             sizeof(MQTT_1_PUBLISH_1_MSG_DATA) -1, (uint8_t*)MQTT_1_PUBLISH_1_MSG_DATA);
                break;

            case 33:        // MQTT: Unsubscribe from topic(s).
                status = bp_api_mqtt_unsubscribe_topics(&seq_nr, sample_state.conn_id_mqtt,
                                                        sizeof(MQTT_1_SUB_TOP1) -1, (uint8_t*)MQTT_1_SUB_TOP1,
                                                        sizeof(MQTT_1_SUB_TOP2) -1, (uint8_t*)MQTT_1_SUB_TOP2,
                                                        sizeof(MQTT_1_SUB_TOP3) -1, (uint8_t*)MQTT_1_SUB_TOP3);
                break;

            case 34:        // MQTT: Disconnect.
                status = bp_api_mqtt_disconnect(&seq_nr, sample_state.conn_id_mqtt);
                break;

            case 35:        // SMS: Setup transmission.
                status = bp_api_sms_setup_transmission(&seq_nr, 0, SMS_1_ENCODING, sizeof(SMS_1_RECIPIENT_MSISDN)-1, (uint8_t*)SMS_1_RECIPIENT_MSISDN);
                break;

            case 36:        // SMS: Write text message.
                status = bp_api_sms_message_append_send(&seq_nr, 0, sizeof(SMS_1_MESSAGE)-1, (uint8_t*)SMS_1_MESSAGE);
                break;

            case 37:        // SMS: Send text message.
                status = bp_api_sms_message_append_send(&seq_nr, 1, 0, NULL);
                break;

            case 38:        // SMS: Wait for an incoming message.
                if(sample_state.sms_received)
                {
                    if(wait_sms_incoming_iter > 0)
                        P_INFO("\n");
                    sample_state.step ++;
                }
                else
                {
                    if(wait_sms_incoming_iter == 0)
                        P_INFO("------- Waiting for incoming SMS ...");

                    if((++wait_sms_incoming_iter % 100) == 0)
                        P_INFO(".");
                }
                status = 0xFF;
                break;

            case 39:        // Wait 2 seconds.
                // Note: Shutting down the radio very fast after receiving an SMS may prevent the radio
                //       from acknowledging the received SMS.
                if(sms_power_off_backoff == 0)
                    sms_power_off_backoff = time(NULL) + 2;
                else if(time(NULL) > sms_power_off_backoff)
                    sample_state.step ++;
                status = 0xFF;
                break;

            case 40:     // Get FW version (again, before powering-off the radio).
                status = bp_api_devi_get_fw_version(&seq_nr);
                break;

            case 41:        // Power-off the Radio
                status = bp_api_mobi_power_off_radio(&seq_nr, BP_API_PWROFF_RADIO__PWROFF_METHOD_GRACEFUL);
                break;

            case 42:        // Set LOW POWER MODE.
                status = bp_api_devi_set_low_power_mode(&seq_nr);
                break;

            case 43:        // Exit from low power mode after 5 seconds.
                if(time(NULL) > bp_status.low_power_entry_ts + 5)
                {
                    P_INFO(">>> Exiting low power mode ...");
                    exit_low_power_mode();
                    P_INFO(" Done!\n");
                    sample_state.step++;
                }
                status = 0xFF;
                break;

            case 44:     // Get FW version (again, after exiting low power mode).
                status = bp_api_devi_get_fw_version(&seq_nr);
                break;

            case 45:
                sample_state.done = 1;
                printf("\nDone!\n");
                break;
            }

            if(status == 0xFF)
            {
            }
            else if(status != 0)
            {
                P_WARN("Failed to send request to BitPipe. Status=%d.\n", status);
                platform_deinit();
                return 2;
            }
            else
            {
                sample_state.awaiting_reply = 1;
                request_timeout = time(NULL) + 30;       // 30 second default timeout.
            }
        }
        else
        {
            // Check for sent request timeout.
            if(time(NULL) >= request_timeout)
            {
                P_WARN("    +++Request timed out. Got no reply from BitPipe.\n");

                // This will trigger a re-send of the request (with a different sequence number).
                sample_state.awaiting_reply = 0;
            }
        }
    }

    platform_deinit();
    return 0;
}

