/** ********************************************************************************************************************
 * @file        main.c
 * @author      Briowireless, Inc.
 * @date        2016-10-28
 * @brief       Main program file for a sample that uses all the BitPipe API.
 * @version     API V1
 * @copyright   Briowireless, Inc., 2016, Licensed under BSD license. See LICENSE file for more details.
 *
 ** *******************************************************************************************************************/

// Include directives  -------------------------------------------------------------------------------------------------

#include <bp_api.h>
#include <../platform.h>                // Depends on the include path. Example: -I../../platforms/win32/
#include <../sample_common.h>           // Common code for all samples.


// Preprocessor definitions --------------------------------------------------------------------------------------------

#define UART_PORT_BITPIPE                   "\\\\.\\COM11"      // Port on which the BitPipe is found (for win32 platform_init).

#define SIM_PIN                             ""                  // When SIM card is locked using a PIN
#define SIM_PUK                             ""                  // When too many bad PINs were tried, the PIN Unlock Key is necessary.

#define APN_CONFIG_HOST                     "m2minternet.apn"
#define APN_CONFIG_USER                     ""
#define APN_CONFIG_PASS                     ""

#define HTTP_REMOTE_HOST                    "httpbin.org"
#define HTTP_REMOTE_PORT                    80
#define HTTP_REMOTE_SECURE                  BP_API_IP_HTTP_INIT_CONNECTION__SECURE_NO
#define HTTP_PATH_QUERY                     "/get"
#define HTTP_METHOD                         BP_API_IP_HTTP_INIT_CONNECTION__METHOD_GET
#define HTTP_ADD_HEADERS                    "MyCustomHeader: A simple value\\0d\\0aMyOtherHeader: AnotherValue"
#define HTTP_POST_DATA                      "This data was posted to the web server, and replied back (because that's what the server at httpbin.org does when POSTing to /post)"
#define HTTP_POST_CONTENT_LEN               (sizeof(HTTP_POST_DATA)-1)      // Only used when method is POST


// Local variables  ----------------------------------------------------------------------------------------------------

// Buffers used by SDK.
static uint8_t deser_buf[260 + MESSAGE_PROTOCOL_CONTEXT_SIZE + MESSAGE_PROTOCOL_HEADER_SIZE + MESSAGE_PROTOCOL_FOOTER_SIZE];
static uint8_t ser_buf[260];


// Local functions  ----------------------------------------------------------------------------------------------------

void exit_low_power_mode(void)
{
    uart_write((uint8_t*)"A", 1);           // Send a character to the BitPipe. This will wake it up from low power mode.
}


// ---------------------------------------------------------------------------------------------------------------------

// SAMPLE MAIN ENTRY POINT

// ---------------------------------------------------------------------------------------------------------------------


int main(void)
{
    bp_api_init_params init_params;                 // BitPipe Host SDK initialization parameters.
    mp_seq_nr seq_nr;                               // Request sequence number
    uint8_t status;                                 // Request execution status.
    uint16_t rx_data_len, consumed_len;             // UART read variables.
    uint8_t uart_rx_buf[100];                       // UART read buffer.
    uint32_t wait_radio_ready_iter = 0;
    uint32_t wait_network_regd_iter = 0;
    uint32_t wait_http_post_write_ready_iter = 0;
    time_t request_timeout = 0;                     // Used to detect request timeouts.


    // Initialize Platform.
    status = platform_init(UART_PORT_BITPIPE);
    if(status != 0)
    {
        P_ERR("Failed to initialize platform: Status=%d.\n", status);
        return 1;
    }


    // Initialize SDK layer.
    init_params.app_callbacks =         &sample_callbacks;      // From common sample code (common_sample.h).
    init_params.uart_write =            &uart_write;            // From platform.
    init_params.deserialize_buf =       deser_buf;              // Local to this sample.
    init_params.deserialize_buf_len =   sizeof(deser_buf);      // Local to this sample.
    init_params.serialize_buf =         ser_buf;                // Local to this sample.
    init_params.serialize_buf_len =     sizeof(ser_buf);        // Local to this sample.

    status = bp_api_init(&init_params);
    if(status != 0)
    {
        P_ERR("Failed to initialize BitPipe API SDK: Status=%d.\n", status);
        platform_deinit();
        return 1;
    }


    // Sample state machine.
    while(!sample_state.done)
    {
        // -------------------------------------------------------------------------------------------------------------
        // ---- This could be placed in a separate thread, if your plaform supports threads.                         ---

        // Check if we've received data from the BitPipe, and process it.
        rx_data_len = uart_read(uart_rx_buf, sizeof(uart_rx_buf));
        while(rx_data_len > 0)
        {
            // This MAY/WILL callback into application callbacks.
            // - The function stops processing UART data when a message is found.
            bp_api_process_incoming(uart_rx_buf, rx_data_len, &consumed_len);

            // Might not have processed all UART data if a message was found.
            rx_data_len -= consumed_len;
        }

        // ---                                                                                                       ---
        // -------------------------------------------------------------------------------------------------------------


        if(!sample_state.awaiting_reply)
        {
            switch(sample_state.step)
            {
            case 0:     // Configure Maximum message payload size.
                status = bp_api_sys_set_max_payload_sz(&seq_nr, 260);
                break;

            case 1:     // Get HW version.
                status = bp_api_devi_get_hw_version(&seq_nr);
                break;

            case 2:     // Get FW version.
                status = bp_api_devi_get_fw_version(&seq_nr);
                break;

            case 3:     // Power-on Radio.
                status = bp_api_mobi_power_on_radio(&seq_nr,
                                                    BP_API_PWRON_RADIO__PWRON_METHOD_DCDC,
                                                    BP_API_PWRON_RADIO__BACKOFF_DISABLED);
                break;

            case 4:    // Wait for radio to be ready.
                if(bp_status.radio_ready)
                {
                    if(wait_radio_ready_iter > 0)
                        P_INFO("Power-on-Radio to Ready took %lds\n", time(NULL) - bp_status.pwr_on_start);
                    sample_state.step ++;
                }
                else
                {
                    if(wait_radio_ready_iter == 0)
                        P_INFO("------- Waiting for radio to be ready ...");

                    if((++wait_radio_ready_iter % 50) == 0)
                        P_INFO(".");
                }
                status = 0xFF;
                break;

            case 5:     // Check SIM card status.
                if(BP_API_RADIO_START_NOTIFY__SIM_NOT_INSERTED == bp_status.sim_status)
                {
                    sample_state.done = 1;
                    P_WARN("  +++No SIM card detected in the BitPipe. Aborting sample.\n");
                }
                else if(BP_API_RADIO_START_NOTIFY__SIM_DENIED_NEED_PIN == bp_status.sim_status)
                {
                    status = bp_api_mobi_set_sim_access_config(NULL, sizeof(SIM_PIN)-1, (uint8_t*)SIM_PIN);
                }
                else if(BP_API_RADIO_START_NOTIFY__SIM_DENIED_NEED_PUK == bp_status.sim_status)
                {
                    status = bp_api_mobi_set_sim_access_config(NULL, sizeof(SIM_PUK)-1, (uint8_t*)SIM_PUK);
                }
                else    // SIM is OK.
                {
                    sample_state.step++;
                    status = 0xFF;
                }
                break;

            case 6:    // Wait for network registration.
                if(bp_status.network_registered)
                {
                    if(wait_network_regd_iter > 0)
                        P_INFO("\n");
                    sample_state.step ++;
                }
                else
                {
                    if(wait_network_regd_iter == 0)
                        P_INFO("------- Waiting for network registration ...");

                    if((++wait_network_regd_iter % 50) == 0)
                        P_INFO(".");
                }
                status = 0xFF;
                break;

            case 7:    // Get Mobile connection status (again).
                status = bp_api_mobi_get_mobile_conn_status(&seq_nr);
                break;

            case 8:    // Initiate HTTP connection.
                status = bp_api_ip_http_init_connection(&seq_nr, HTTP_REMOTE_SECURE, HTTP_METHOD,
                                                        sizeof(HTTP_REMOTE_HOST)-1, (uint8_t*)HTTP_REMOTE_HOST,
                                                        HTTP_REMOTE_PORT,
                                                        sizeof(HTTP_PATH_QUERY)-1, (uint8_t*)HTTP_PATH_QUERY,
                                                        sizeof(HTTP_ADD_HEADERS)-1, (uint8_t*)HTTP_ADD_HEADERS,
                                                        HTTP_POST_CONTENT_LEN);
                break;

            case 9:    // Wait for HTTP connection to be ready to send POST (Only if method is POST)
                if(HTTP_METHOD == BP_API_IP_HTTP_INIT_CONNECTION__METHOD_POST)
                {
                    if(sample_state.conn_status_http == IP_CONN_STATUS_ESTABLISHED)
                    {
                        if(wait_http_post_write_ready_iter > 0)
                            P_INFO("\n");
                        sample_state.step ++;
                    }
                    else
                    {
                        if(wait_http_post_write_ready_iter == 0)
                            P_INFO("------- Waiting for HTTP POST WRITE READY ...");

                        if((++wait_http_post_write_ready_iter % 50) == 0)
                            P_INFO(".");
                    }
                }
                else
                {
                    // Method is not POST, no need to wait on this notification.
                    // In fact, it will never be sent by the BitPipe.
                    sample_state.step++;
                }
                status = 0xFF;
                break;

            case 10:        // Send POST DATA (Only if method is POST).
                if(HTTP_METHOD == BP_API_IP_HTTP_INIT_CONNECTION__METHOD_POST  &&
                        sample_state.http_post_data_idx < HTTP_POST_CONTENT_LEN)                 // Check if we're done.
                {
                    uint32_t remaining = HTTP_POST_CONTENT_LEN - sample_state.http_post_data_idx;
                    sample_state.http_post_last_pay_len = remaining % 256;   // Can send at most 255 bytes per chunk.
                    uint8_t write_op = (remaining - sample_state.http_post_last_pay_len) > 0 ?
                                BP_API_IP_HTTP_POST_WRITE__WRITEOP_WRITE_TO_BUFFER_AND_SEND :
                                BP_API_IP_HTTP_POST_WRITE__WRITEOP_WRITE_TO_BUFFER_AND_SEND_COMPLETE;

                    status = bp_api_ip_http_post_write(&seq_nr, sample_state.conn_id_http, write_op,
                                                       sample_state.http_post_last_pay_len,
                                                       (uint8_t*)HTTP_POST_DATA + sample_state.http_post_data_idx);
                }
                else    // METHOD is not POST, or all data posted successfully.
                {
                    sample_state.step++;
                    status = 0xFF;
                }
                break;

            case 11:        // Wait for connection to be closed (here, we expect the web server to close the connection)
                if(IP_CONN_STATUS_CLOSED == sample_state.conn_status_http)
                {
                    sample_state.step++; // Method is not POST, move to next step.
                }
                status = 0xFF;
                break;

            case 12:        // Power-off the Radio
                status = bp_api_mobi_power_off_radio(&seq_nr, BP_API_PWROFF_RADIO__PWROFF_METHOD_GRACEFUL);
                break;

            case 13:
                sample_state.done = 1;
                printf("\nDone!\n");
                break;
            }

            if(status == 0xFF)
            {
            }
            else if(status != 0)
            {
                P_WARN("Failed to send request to BitPipe. Status=%d.\n", status);
                platform_deinit();
                return 2;
            }
            else
            {
                sample_state.awaiting_reply = 1;
                request_timeout = time(NULL) + 30;       // 30 second default timeout.
            }
        }
        else
        {
            // Check for sent request timeout.
            if(time(NULL) >= request_timeout)
            {
                P_WARN("    +++Request timed out. Got no reply from BitPipe.\n");

                // This will trigger a re-send of the request (with a different sequence number).
                sample_state.awaiting_reply = 0;
            }
        }
    }

    platform_deinit();
    return 0;
}

