/** ********************************************************************************************************************
 * @file        main.c
 * @author      Briowireless, Inc.
 * @date        2016-10-28
 * @brief       Main program file for SMS Sample.
 * @version     API V1
 * @copyright   Briowireless, Inc., 2016, Licensed under BSD license. See LICENSE file for more details.
 *
 * This sample illustrates how to send a SMS message and receive a SMS message. It goes through the following steps:
 *  - Power on the radio (and wait until ready)
 *  - Configure the APN
 *  - Query the mobile network status (and wait until registered on mobile network).
 *  - Sends a SMS message.
 *  - Waits for an incoming SMS message.
 *  - Power-off the radio.
 *
 ** *******************************************************************************************************************/

// Include directives  -------------------------------------------------------------------------------------------------

#include <bp_api.h>
#include <../platform.h>                // Depends on the include path. Example: -I../../platforms/win32/
#include <../sample_common.h>           // To manage timeouts, and statistics.


// Preprocessor definitions --------------------------------------------------------------------------------------------

#define UART_PORT_BITPIPE                   "\\\\.\\COM11"      // Port on which the BitPipe is found (for win32 platform_init).

#define SIM_PIN                             ""                  // When SIM card is locked using a PIN
#define SIM_PUK                             ""                  // When too many bad PINs were tried, the PIN Unlock Key is necessary.

#define APN_CONFIG_HOST                     "m2minternet.apn"
#define APN_CONFIG_USER                     ""
#define APN_CONFIG_PASS                     ""

#define SMS_1_RECIPIENT_MSISDN              "5555555555"        // SMS recipient mobile phone number. Do not prefix with '+'.
#define SMS_1_ENCODING                      0                   // 7-bit GSM 03.38
#define SMS_1_MESSAGE                       "Eric is not at home.\nHe is currently waiting the subway at the station."


// Local variables  ----------------------------------------------------------------------------------------------------

// Buffers used by SDK.
static uint8_t deser_buf[260 + MESSAGE_PROTOCOL_CONTEXT_SIZE + MESSAGE_PROTOCOL_HEADER_SIZE + MESSAGE_PROTOCOL_FOOTER_SIZE];
static uint8_t ser_buf[260];


// ---------------------------------------------------------------------------------------------------------------------

// SAMPLE MAIN ENTRY POINT

// ---------------------------------------------------------------------------------------------------------------------


int main(void)
{
    bp_api_init_params init_params;
    uint8_t status;
    uint16_t rx_data_len, consumed_len;
    uint8_t uart_rx_buf[100];
    uint32_t wait_radio_ready_iter = 0;
    uint32_t wait_network_regd_iter = 0;
    uint32_t wait_sms_incoming_iter = 0;
    time_t request_timeout = 0;
    time_t sms_power_off_backoff = 0;


    // Initialize Platform.
    status = platform_init(UART_PORT_BITPIPE);                  // From platform.
    if(status != 0)
    {
        P_ERR("Failed to initialize platform: Status=%d.\n", status);
        return 1;
    }


    // Initialize SDK layer.
    init_params.app_callbacks =         &sample_callbacks;      // From common sample code (common_sample.h).
    init_params.uart_write =            &uart_write;            // From platform.
    init_params.deserialize_buf =       deser_buf;              // Local to this sample.
    init_params.deserialize_buf_len =   sizeof(deser_buf);      // Local to this sample.
    init_params.serialize_buf =         ser_buf;                // Local to this sample.
    init_params.serialize_buf_len =     sizeof(ser_buf);        // Local to this sample.

    status = bp_api_init(&init_params);
    if(status != 0)
    {
        P_ERR("Failed to initialize BitPipe API SDK: Status=%d.\n", status);
        platform_deinit();
        return 1;
    }


    // Sample state machine.
    while(!sample_state.done)
    {
        // -------------------------------------------------------------------------------------------------------------
        // ---- This could be placed in a separate thread, if your plaform supports threads.                         ---

        // Check if we've received data from the BitPipe, and process it.
        rx_data_len = uart_read(uart_rx_buf, sizeof(uart_rx_buf));      // From platform.
        while(rx_data_len > 0)
        {
            // This MAY/WILL callback into application callbacks.
            // - The function stops processing UART data when a message is found.
            bp_api_process_incoming(uart_rx_buf, rx_data_len, &consumed_len);

            // Might not have processed all UART data if a message was found.
            rx_data_len -= consumed_len;
        }

        // ---                                                                                                       ---
        // -------------------------------------------------------------------------------------------------------------


        if(!sample_state.awaiting_reply)
        {
            switch(sample_state.step)
            {
            case 0:
                // Get HW version.
                status = bp_api_devi_get_hw_version(NULL);
                break;

            case 1:
                // Get FW version.
                status = bp_api_devi_get_fw_version(NULL);
                break;

            case 2:
                // Power-on Radio.
                status = bp_api_mobi_power_on_radio(NULL,
                                                    BP_API_PWRON_RADIO__PWRON_METHOD_DCDC,
                                                    BP_API_PWRON_RADIO__BACKOFF_DISABLED);
                break;

            case 3:
                // Get Radio Status.
                status = bp_api_mobi_get_radio_status(NULL);
                break;

            case 4:
                // Wait for radio to be ready.
                if(bp_status.radio_ready)
                {
                    if(wait_radio_ready_iter > 0)
                        P_INFO("Power-on-Radio to Ready took %lds\n", time(NULL) - bp_status.pwr_on_start);
                    sample_state.step ++;
                }
                else
                {
                    if(wait_radio_ready_iter == 0)
                        P_INFO("------- Waiting for radio to be ready ...");

                    if((++wait_radio_ready_iter % 50) == 0)
                        P_INFO(".");
                }
                status = 0xFF;
                break;

            case 5:     // Check SIM card status.
                if(BP_API_RADIO_START_NOTIFY__SIM_NOT_INSERTED == bp_status.sim_status)
                {
                    sample_state.done = 1;
                    P_WARN("  +++No SIM card detected in the BitPipe. Aborting sample.\n");
                }
                else if(BP_API_RADIO_START_NOTIFY__SIM_DENIED_NEED_PIN == bp_status.sim_status)
                {
                    status = bp_api_mobi_set_sim_access_config(NULL, sizeof(SIM_PIN)-1, (uint8_t*)SIM_PIN);
                }
                else if(BP_API_RADIO_START_NOTIFY__SIM_DENIED_NEED_PUK == bp_status.sim_status)
                {
                    status = bp_api_mobi_set_sim_access_config(NULL, sizeof(SIM_PUK)-1, (uint8_t*)SIM_PUK);
                }
                else    // SIM is OK.
                {
                    sample_state.step++;
                    status = 0xFF;
                }
                break;

            case 6:
                // Get Radio Status (again).
                status = bp_api_mobi_get_radio_status(NULL);
                break;

            case 7:
                // Set APN Configuration (This doesn't need to be set every time).
                status = bp_api_mobi_set_apn_config(NULL,
                                                    sizeof(APN_CONFIG_HOST)-1, (uint8_t*)APN_CONFIG_HOST,
                                                    sizeof(APN_CONFIG_USER)-1, (uint8_t*)APN_CONFIG_USER,
                                                    sizeof(APN_CONFIG_PASS)-1, (uint8_t*)APN_CONFIG_PASS);
                break;

            case 8:
                // Get Mobile connection status.
                status = bp_api_mobi_get_mobile_conn_status(NULL);
                break;

            case 9:
                // Wait for network registration.
                if(bp_status.network_registered)
                {
                    if(wait_network_regd_iter > 0)
                        P_INFO("\n");
                    sample_state.step ++;
                }
                else
                {
                    if(wait_network_regd_iter == 0)
                        P_INFO("------- Waiting for network registration ...");

                    if((++wait_network_regd_iter % 50) == 0)
                        P_INFO(".");
                }
                status = 0xFF;
                break;

            case 10:
                // Get Mobile connection status (again).
                status = bp_api_mobi_get_mobile_conn_status(NULL);
                break;

            case 11:        // SMS: Setup transmission.
                status = bp_api_sms_setup_transmission(NULL, 0, SMS_1_ENCODING, sizeof(SMS_1_RECIPIENT_MSISDN)-1, (uint8_t*)SMS_1_RECIPIENT_MSISDN);
                break;

            case 12:        // SMS: Write text message.
                status = bp_api_sms_message_append_send(NULL, 0, sizeof(SMS_1_MESSAGE)-1, (uint8_t*)SMS_1_MESSAGE);
                break;

            case 13:        // SMS: Send text message.
                status = bp_api_sms_message_append_send(NULL, 1, 0, NULL);
                break;

            case 14:        // SMS: Wait for an incoming message.
                if(sample_state.sms_received)
                {
                    if(wait_sms_incoming_iter > 0)
                        P_INFO("\n");
                    sample_state.step ++;
                }
                else
                {
                    if(wait_sms_incoming_iter == 0)
                        P_INFO("------- Waiting for incoming SMS ...");

                    if((++wait_sms_incoming_iter % 100) == 0)
                        P_INFO(".");
                }
                status = 0xFF;
                break;

            case 15:        // Wait 2 seconds.
                // Note: Shutting down the radio very fast after receiving an SMS may prevent the radio
                //       from acknowledging the received SMS.
                if(sms_power_off_backoff == 0)
                    sms_power_off_backoff = time(NULL) + 2;
                else if(time(NULL) > sms_power_off_backoff)
                    sample_state.step ++;
                status = 0xFF;
                break;

            case 16:
                status = bp_api_mobi_power_off_radio(NULL, BP_API_PWROFF_RADIO__PWROFF_METHOD_GRACEFUL);
                break;

            case 17:
                sample_state.done = 1;
                printf("\nDone!\n");
                break;
            }

            if(status == 0xFF)
            {
            }
            else if(status != 0)
            {
                P_WARN("Failed to send request to BitPipe. Status=%d.\n", status);
                platform_deinit();
                return 2;
            }
            else
            {
                sample_state.awaiting_reply = 1;
                request_timeout = time(NULL) + 30;       // 30 second default timeout.
            }
        }
        else
        {
            // Check for sent request timeout.
            if(time(NULL) >= request_timeout)
            {
                P_WARN("    +++Request timed out. Got no reply from BitPipe.\n");

                // This will trigger a re-send of the request (with a different sequence number).
                sample_state.awaiting_reply = 0;
            }
        }
    }

    platform_deinit();
    return 0;
}

