/** ********************************************************************************************************************
 * @file        sample_common.c
 * @author      Briowireless, Inc.
 * @date        2016-11-03
 * @brief       Common code for all samples. Manages the sample state machine.
 * @copyright   Briowireless, Inc., 2016, Licensed under BSD license. See LICENSE file for more details.
 *
 ** *******************************************************************************************************************/

// Include directives  -------------------------------------------------------------------------------------------------

#include <bp_api.h>
#include <message_defs.h>

/**
 * @defgroup sample_common Sample Common Code
 * Handles BitPipe status and sample execution state.
 * @{
 */

#include <../sample_common.h>


// Preprocessor definitions --------------------------------------------------------------------------------------------

// Global variables ----------------------------------------------------------------------------------------------------

struct __bp_status bp_status = {
    0,
    0xFF,
    0,
    BP_API_IP_INVALID_CONNECTION_ID,
    0,
    0,
};

struct __sample_state sample_state = {
    0, 0, 0, 0, 0, 0, 0, 0,
    IP_CONN_STATUS_CLOSED,
    IP_CONN_STATUS_CLOSED,
    IP_CONN_STATUS_CLOSED,
    BP_API_IP_INVALID_CONNECTION_ID,
    BP_API_IP_INVALID_CONNECTION_ID,
    BP_API_IP_INVALID_CONNECTION_ID,
};

bp_api_callbacks sample_callbacks  = {
    sys_error_report_post,
    sys_set_max_payload_sz_reply,
    sys_set_uart_speed_reply,

    devi_get_hw_version_reply,
    devi_get_fw_version_reply,
    devi_get_product_name_reply,
    devi_get_vin_vio_voltage_reply,
    devi_get_junction_temp_reply,
    devi_set_low_power_reply,
    devi_get_latest_fw_version_reply,
    devi_update_fw_reply,
    devi_get_autonomous_mode_reply,
    devi_set_autonomous_mode_reply,

    mobi_power_on_radio_reply,
    mobi_power_off_radio_reply,
    mobi_radio_start_notify_post,
    mobi_get_radio_status_reply,
    mobi_get_imei_reply,
    mobi_get_iccid_reply,
    mobi_get_imsi_reply,
    mobi_get_own_phone_number_reply,
    mobi_get_apn_config_reply,
    mobi_set_apn_config_reply,
    mobi_set_sim_access_config_reply,
    mobi_get_mobile_conn_status_reply,
    mobi_mobile_conn_status_change_notify_post,
    mobi_get_mobile_date_time_reply,

    ip_tcp_init_connection_reply,
    ip_tcp_connection_established_notify_post,
    ip_tcp_write_append_reply,
    ip_tcp_write_send_reply,
    ip_tcp_rx_notify_request,
    ip_http_init_connection_reply,
    ip_http_post_write_ready_notify_post,
    ip_http_post_write_reply,
    ip_http_rx_notify_request,
    ip_connection_close_reply,
    ip_connection_close_notify_post,

    mqtt_init_connection_reply,
    mqtt_connection_established_notify_post,
    mqtt_subscribe_topics_reply,
    mqtt_unsubscribe_topics_reply,
    mqtt_publish_reply,
    mqtt_publication_rx_notify_request,
    mqtt_disconnect_reply,
    mqtt_disconnect_notify_post,

    sms_setup_transmission_reply,
    sms_message_append_send_reply,
    sms_message_rx_notify_request,
    sms_message_part_notify_request,

};


// Local functions  ----------------------------------------------------------------------------------------------------

static void sample_move_next_step(uint8_t move_next)
{
    if(move_next)
        sample_state.step ++;
    sample_state.awaiting_reply = 0;
}


// Global (extern) function  -------------------------------------------------------------------------------------------


//  SYSTEM

/// @copydoc bp_api_sys_error_report_post
void sys_error_report_post(uint8_t seq_nr, uint8_t facility, uint8_t err_nr)
{
    P_INFO("[%02X] ERROR REPORT! Facility=%02X ErrNr=%02X\n", seq_nr, facility, err_nr);
}

/// @copydoc bp_api_sys_set_max_payload_sz_reply
void sys_set_max_payload_sz_reply(uint8_t seq_nr, uint8_t status)
{
    P_INFO("[%02X] SET MAX MESSAGE PAYLOAD SZ reply. Status=%02X\n", seq_nr, status);

    // Move to next step.
    sample_move_next_step(MSG_STATUS_CODE_SUCCESS == status);
}

/// @copydoc bp_api_sys_set_uart_speed_reply
void sys_set_uart_speed_reply(uint8_t seq_nr, uint8_t status)
{
    P_INFO("[%02X] SET UART SPEED reply. Status=%02X\n", seq_nr, status);

    // Move to next step.
    sample_move_next_step(MSG_STATUS_CODE_SUCCESS == status);
}


//  DEVICE INFORMATION

/// @copydoc bp_api_devi_get_hw_version_reply
void devi_get_hw_version_reply(uint8_t seq_nr, uint8_t status, uint8_t hw_ver_len, uint8_t* hw_ver)
{
    P_INFO("[%02X] GET HW VERSION reply. Status=%02X Version=%.*s\n", seq_nr, status, hw_ver_len, hw_ver);

    // Move to next step.
    sample_move_next_step(MSG_STATUS_CODE_SUCCESS == status);
}

/// @copydoc bp_api_devi_get_fw_version_reply
void devi_get_fw_version_reply(uint8_t seq_nr, uint8_t status, uint16_t boot_ver, uint8_t api_ver,
                                      uint8_t bp_ver_len, uint8_t* bp_ver, uint8_t radio_ver_len, uint8_t* radio_ver)
{
    P_INFO("[%02X] GET FW VERSION reply. Status=%02X BootVer=%02X.%02X API=%02X BitPipeFw=%.*s RadioFw=%.*s\n", seq_nr, status,
           (boot_ver >> 8) & 0xFF, boot_ver & 0xFF, api_ver, bp_ver_len, bp_ver, radio_ver_len, radio_ver);

    // Move to next step.
    sample_move_next_step(MSG_STATUS_CODE_SUCCESS == status);
}

/// @copydoc bp_api_devi_get_product_name_reply
void devi_get_product_name_reply(uint8_t seq_nr, uint8_t status, uint8_t prod_name_len, uint8_t* prod_name)
{
    P_INFO("[%02X] GET PRODUCT NAME reply. Status=%02X ProductName=%.*s\n", seq_nr, status, prod_name_len, prod_name);

    // Move to next step.
    sample_move_next_step(MSG_STATUS_CODE_SUCCESS == status);
}

/// @copydoc bp_api_devi_get_vin_vio_voltage_reply
void devi_get_vin_vio_voltage_reply(uint8_t seq_nr, uint8_t status, uint16_t vin_voltage, uint16_t vio_voltage)
{
    P_INFO("[%02X] GET VIN/VIO VOLTAGES reply. Status=%02X VIN=%dmV VIO=%dmV\n", seq_nr, status, vin_voltage, vio_voltage);

    // Move to next step.
    sample_move_next_step(MSG_STATUS_CODE_SUCCESS == status);
}

/// @copydoc bp_api_devi_get_junction_temp_reply
void devi_get_junction_temp_reply(uint8_t seq_nr, uint8_t status, int8_t junction_temp)
{
    P_INFO("[%02X] GET JUNCTION TEMP reply. Status=%02X TEMP=%dC\n", seq_nr, status, junction_temp);

    // Move to next step.
    sample_move_next_step(MSG_STATUS_CODE_SUCCESS == status);
}

/// @copydoc bp_api_devi_set_low_power_reply
void devi_set_low_power_reply(uint8_t seq_nr, uint8_t status)
{
    P_INFO("[%02X] SET LOW POWER reply. Status=%02X\n", seq_nr, status);

    if(MSG_STATUS_CODE_SUCCESS == status)
    {
        bp_status.low_power_entry_ts = time(NULL);
    }

    // Move to next step.
    sample_move_next_step(MSG_STATUS_CODE_SUCCESS == status);
}

/// @copydoc bp_api_devi_get_latest_fw_version_reply
void devi_get_latest_fw_version_reply(uint8_t seq_nr, uint8_t status, uint8_t fw_ver_len, uint8_t* fw_ver)
{
    P_INFO("[%02X] GET LATEST FW VERSION reply. Status=%02X LatestFwVersion=%*s\n", seq_nr, status, fw_ver_len, fw_ver);

    // Move to next step.
    sample_move_next_step(MSG_STATUS_CODE_SUCCESS == status);
}

/// @copydoc bp_api_devi_update_fw_reply
void devi_update_fw_reply(uint8_t seq_nr, uint8_t status)
{
    P_INFO("[%02X] UPDATE FIRMWARE reply. Status=%02X\n", seq_nr, status);

    // Move to next step.
    sample_move_next_step(MSG_STATUS_CODE_SUCCESS == status);
}

/// @copydoc bp_api_devi_get_autonomous_mode_config_reply
void devi_get_autonomous_mode_reply(uint8_t seq_nr, uint8_t status, uint8_t secure, uint16_t ka_ival, uint16_t port,
                                                      uint8_t host_len, uint8_t* host, uint8_t user_len, uint8_t* user,
                                                      uint8_t pass_len, uint8_t* pass, uint8_t gpfx_len, uint8_t* gpfx)
{
    if(MSG_STATUS_CODE_SUCCESS == status)
    {
        P_INFO("[%02X] GET AUTONOMOUS MODE reply. Status=%02X Secure=%s KA_ival=%d Port=%d Host=%.*s User=%.*s Pass=%.*s Gpfx=%.*s\n",
               seq_nr, status,
               secure == 0 ? "No" : "Yes",
               ka_ival,
               port,
               host_len, host,
               user_len, user,
               pass_len, pass,
               gpfx_len, gpfx);
    }
    else if(BP_API_DEVI_GET_AUTONOMOUS_MODE_CONFIG__ERR_NOT_CONFIGURED == status)
    {
        P_INFO("[%02X] GET AUTONOMOUS MODE reply. Status=%02X - AUTONOMOUS MODE CONFIGURATION IS NOT SET!\n", seq_nr, status);
    }

    // Move to next step.
    sample_move_next_step(MSG_STATUS_CODE_SUCCESS == status || BP_API_DEVI_GET_AUTONOMOUS_MODE_CONFIG__ERR_NOT_CONFIGURED == status);
}

/// @copydoc bp_api_devi_set_autonomous_mode_config_reply
void devi_set_autonomous_mode_reply(uint8_t seq_nr, uint8_t status)
{
    P_INFO("[%02X] SET AUTONOMOUS MODE reply. Status=%02X\n", seq_nr, status);

    // Move to next step.
    sample_move_next_step(MSG_STATUS_CODE_SUCCESS == status);
}


//  MOBILE INFORMATION

/// @copydoc bp_api_mobi_power_on_radio_reply
void mobi_power_on_radio_reply(uint8_t seq_nr, uint8_t status)
{
    P_INFO("[%02X] POWER ON RADIO reply. Status=%02X\n", seq_nr, status);

    if(BP_API_PWRON_RADIO__STATUS_VIN_INCOMPAT_PWRON_METHOD == status)
    {
        P_WARN("    -> VIN voltage is out of bounds for specified power-on method.\n");
        sample_state.done = 1; // Abort sample.
    }
    else if(BP_API_PWRON_RADIO__STATUS_DENIED_FRIENDLY_NETWORK_CHK == status)
    {
        P_WARN("    -> Power-on-radio denied by network-friendly-behavior check.\n");
        // You may not power-on the radio more than 6 times per hour, and must wait at least 5 minutes
        //  between each successful power-on radio requests.

        // Here, start a timer to retry the power-on-radio operation every X seconds, or minutes.
        sample_state.done = 1; // Abort sample.
    }

    if(status == 0)
        bp_status.pwr_on_start = time(NULL);

    // Move to next step.
    sample_move_next_step(MSG_STATUS_CODE_SUCCESS == status || BP_API_PWRON_RADIO__STATUS_ALREADY_STARTED == status);
}

/// @copydoc bp_api_mobi_power_off_radio_reply
void mobi_power_off_radio_reply(uint8_t seq_nr, uint8_t status)
{
    P_INFO("[%02X] POWER OFF RADIO reply. Status=%02X\n", seq_nr, status);

    if(MSG_STATUS_CODE_SUCCESS == status)
    {
        bp_status.radio_ready = 0;
        bp_status.network_registered = 0;
        bp_status.sim_status = 0xFF;
    }

    // Move to next step.
    sample_move_next_step(MSG_STATUS_CODE_SUCCESS == status);
}

/// @copydoc bp_api_mobi_radio_start_notify_post
void mobi_radio_start_notify_post(uint8_t seq_nr, uint8_t notify_reason, uint8_t sim_status)
{
    P_INFO("\n[%02X] RADIO START NOTIFY - Reason=%02X SimStatus=%02X\n", seq_nr, notify_reason, sim_status);

    // Keep track if we're registered on the network.
    bp_status.radio_ready = (BP_API_RADIO_START_NOTIFY__REASON_PWRON_READY == notify_reason ||
                             BP_API_RADIO_START_NOTIFY__REASON_UNEXPECTED_REBOOT == notify_reason);

    bp_status.sim_status = sim_status;
}

/// @copydoc bp_api_mobi_get_radio_status_reply
void mobi_get_radio_status_reply(uint8_t seq_nr, uint8_t status, uint8_t radio_status, uint8_t sim_status)
{
    P_INFO("[%02X] GET RADIO STATUS reply. Status=%02X - RadioStatus=%02X SimStatus=%02X\n", seq_nr, status, radio_status, sim_status);

    bp_status.radio_ready = (BP_API_GET_RADIO_STATUS__RADIO_PWRON_READY == radio_status);

    bp_status.sim_status = sim_status;

    // Move to next step.
    sample_move_next_step(MSG_STATUS_CODE_SUCCESS == status);
}

/// @copydoc bp_api_mobi_get_imei_reply
void mobi_get_imei_reply(uint8_t seq_nr, uint8_t status, uint8_t* imei)
{
    P_INFO("[%02X] GET IMEI reply. Status=%02X - IMEI=%.15s\n", seq_nr, status, imei);

    // Move to next step.
    sample_move_next_step(MSG_STATUS_CODE_SUCCESS == status);
}

/// @copydoc bp_api_mobi_get_iccid_reply
void mobi_get_iccid_reply(uint8_t seq_nr, uint8_t status, uint8_t* iccid)
{
    P_INFO("[%02X] GET ICCID reply. Status=%02X - ICCID=%.20s\n", seq_nr, status, iccid);

    // Move to next step.
    sample_move_next_step(MSG_STATUS_CODE_SUCCESS == status);
}

/// @copydoc bp_api_mobi_get_imsi_reply
void mobi_get_imsi_reply(uint8_t seq_nr, uint8_t status, uint8_t* imsi)
{
    P_INFO("[%02X] GET IMSI reply. Status=%02X - IMSI=%.15s\n", seq_nr, status, imsi);

    // Move to next step.
    sample_move_next_step(MSG_STATUS_CODE_SUCCESS == status);
}

/// @copydoc bp_api_mobi_get_own_phone_number_reply
void mobi_get_own_phone_number_reply(uint8_t seq_nr, uint8_t status, uint8_t msisdn_len, uint8_t* msisdn)
{
    P_INFO("[%02X] GET OWN PHONE NUMBER reply. Status=%02X - MSISDN=%.*s\n", seq_nr, status, msisdn_len, msisdn);

    // Move to next step.
    sample_move_next_step(MSG_STATUS_CODE_SUCCESS == status);
}

/// @copydoc bp_api_mobi_get_apn_config_reply
void mobi_get_apn_config_reply(uint8_t seq_nr, uint8_t status, uint8_t host_len, uint8_t* host,
                                      uint8_t user_len, uint8_t* user, uint8_t pass_len, uint8_t* pass)
{
    P_INFO("[%02X] GET APN CONFIG reply. Status=%02X - Host=%.*s User=%.*s Pass=%.*s\n", seq_nr, status,
           host_len, host, user_len, user, pass_len, pass);

    // Move to next step.
    sample_move_next_step(MSG_STATUS_CODE_SUCCESS == status);
}

/// @copydoc bp_api_mobi_set_apn_config_reply
void mobi_set_apn_config_reply(uint8_t seq_nr, uint8_t status)
{
    P_INFO("[%02X] SET APN CONFIG reply. Status=%02X\n", seq_nr, status);

    // Move to next step.
    sample_move_next_step(MSG_STATUS_CODE_SUCCESS == status);
}

/// @copydoc bp_api_mobi_set_sim_access_config_reply
void mobi_set_sim_access_config_reply(uint8_t seq_nr, uint8_t status)
{
    P_INFO("[%02X] SET SIM ACCESS CONFIG reply. Status=%02X\n", seq_nr, status);

    // Move to next step.
    sample_move_next_step(MSG_STATUS_CODE_SUCCESS == status);
}

/// @copydoc bp_api_mobi_get_mobile_conn_status_reply
void mobi_get_mobile_conn_status_reply(uint8_t seq_nr, uint8_t status, uint8_t conn_ind, uint8_t link_tech,
                                              uint16_t net_lac, uint32_t net_cell_id, uint8_t signal_strength,
                                              uint8_t* net_mcc, uint8_t* net_mnc)
{
    P_INFO("[%02X] GET MOBILE CONN STATUS reply. Status=%02X - ConnectedInd=%d LinkTech=%d NetLAC=%04X NetCellID=%08X SignalStrength=",
           seq_nr, status, conn_ind, link_tech, net_lac, net_cell_id);

    bp_status.network_registered = (BP_API_GET_MOBILE_CONN_STATUS__CONN_IND_REGD_HOME == conn_ind ||
                                    BP_API_GET_MOBILE_CONN_STATUS__CONN_IND_REGD_FOREIGN == conn_ind);

    if(signal_strength == 0xFF)
    {
        printf("Unknown");
    }
    else
    {
        int16_t signal_str_dbm = (int16_t)signal_strength - (link_tech==1?110:link_tech==2?124:link_tech==3?140:0);
        printf("%ddBm", signal_str_dbm);
    }
    P_INFO(" MCC=%s MNC=%.4s\n", net_mcc, net_mnc);

    // Move to next step.
    sample_move_next_step(MSG_STATUS_CODE_SUCCESS == status);
}

/// @copydoc bp_api_mobi_mobile_conn_status_change_notify_post
void mobi_mobile_conn_status_change_notify_post(uint8_t seq_nr, uint8_t conn_ind, uint8_t link_tech,
                                                       uint16_t net_lac, uint32_t net_cell_id)
{
    P_INFO("\n[%02X] MOBILE CONN STATUS CHANGE NOTIFY - ConnectedInd=%d LinkTech=%d NetLAC=%04X NetCellID=%08X\n",
           seq_nr, conn_ind, link_tech, net_lac, net_cell_id);

    bp_status.network_registered = (BP_API_GET_MOBILE_CONN_STATUS__CONN_IND_REGD_HOME == conn_ind ||
                                    BP_API_GET_MOBILE_CONN_STATUS__CONN_IND_REGD_FOREIGN == conn_ind);
}

/// @copydoc bp_api_mobi_get_mobile_date_time_reply
void mobi_get_mobile_date_time_reply(uint8_t seq_nr, uint8_t status, uint8_t year, uint8_t month, uint8_t day,
                                            uint8_t hour, uint8_t minute, uint8_t second)
{
    P_INFO("[%02X] GET MOBILE DATE TIME reply. Status=%02X - %04d-%02d-%02d %02d:%02d:%02d\n", seq_nr, status,
           (uint16_t)year + 2000, month+1, day+1, hour, minute, second);

    // Move to next step.
    sample_move_next_step(MSG_STATUS_CODE_SUCCESS == status);
}



//  IP CONNECTIVITY

/// @copydoc bp_api_ip_tcp_init_connection_reply
void ip_tcp_init_connection_reply(uint8_t seq_nr, uint8_t status, uint8_t tcp_conn_id)
{
    P_INFO("[%02X] TCP INITIATE CONNECTION reply. Status=%02X - TcpConn=%d\n", seq_nr, status, tcp_conn_id);

    if(MSG_STATUS_CODE_SUCCESS == status)
    {
        sample_state.conn_id_tcp = tcp_conn_id;
        sample_state.conn_status_tcp = IP_CONN_STATUS_OPEN;
    }

    if(BP_API_IP_TCP_INIT_CONNECTION__STATUS_NO_TCP_CONN_AVAILABLE == status)
    {
        P_ERR("+++++No TCP connections available. Did you abort a previous test? Reset the BitPipe or power-off the radio. Aborting sample.\n");
        sample_state.done = 1;
    }

    // Move to next step.
    sample_move_next_step(MSG_STATUS_CODE_SUCCESS == status);
}

/// @copydoc bp_api_ip_tcp_connection_established_notify_post
void ip_tcp_connection_established_notify_post(uint8_t seq_nr, uint8_t tcp_conn_id)
{
    P_INFO("\n[%02X] TCP CONNECTION ESTABLISHED NOTIFY. TcpConn=%d\n", seq_nr, tcp_conn_id);

    if(sample_state.conn_id_tcp == tcp_conn_id)
        sample_state.conn_status_tcp = IP_CONN_STATUS_ESTABLISHED;
}

/// @copydoc bp_api_ip_tcp_write_append_reply
void ip_tcp_write_append_reply(uint8_t seq_nr, uint8_t status, uint16_t buf_sz_current, uint16_t buf_sz_max)
{
    P_INFO("[%02X] TCP WRITE APPEND reply. Status=%02X - BufferUsage=%3.2f%%\n", seq_nr, status,
           (buf_sz_max!=0) ? buf_sz_current / buf_sz_max * 100.0: 0);

    // Move to next step.
    sample_move_next_step(MSG_STATUS_CODE_SUCCESS == status);
}

/// @copydoc bp_api_ip_tcp_write_send_reply
void ip_tcp_write_send_reply(uint8_t seq_nr, uint8_t status)
{
    P_INFO("[%02X] TCP WRITE SEND reply. Status=%02X\n", seq_nr, status);

    // Move to next step.
    sample_move_next_step(MSG_STATUS_CODE_SUCCESS == status);
}

/// @copydoc bp_api_ip_tcp_rx_notify_request
void ip_tcp_rx_notify_request(uint8_t seq_nr, uint8_t tcp_conn_id, uint8_t data_len, uint8_t* data)
{
    P_INFO("\n[%02X] TCP RX NOTIFY request. TcpConn=%d DataLen=%d\n", seq_nr, tcp_conn_id, data_len);

    // IMPORTANT: Reply to this request, or else the BitPipe will re-send this data chunk.
    bp_api_ip_tcp_rx_notify_ack(seq_nr);

    sample_state.tcp_data_rx_idx += data_len;

    P_DEBUG("     > ");
#ifdef DBG_PRINT_RXBYTES_HEXADECIMAL
    uint8_t i;
    for(i=0; i<data_len; i++)
        P_DEBUG("%02X ", data[i]);
#else
    P_DEBUG("%.*s", data_len, data);
#endif
    P_DEBUG("\n");
}

/// @copydoc bp_api_ip_http_init_connection_reply
void ip_http_init_connection_reply(uint8_t seq_nr, uint8_t status, uint8_t http_conn_id)
{
    P_INFO("[%02X] HTTP INITIATE CONNECTION reply. Status=%02X - HttpConn=%d\n", seq_nr, status, http_conn_id);

    if(MSG_STATUS_CODE_SUCCESS == status)
    {
        sample_state.conn_id_http = http_conn_id;
        sample_state.conn_status_http = IP_CONN_STATUS_OPEN;
    }
    else if(MSG_STATUS_WARN_BITPIPE_BUSY != status)     // An error occurred.
    {
        P_ERR("  +++Aborting because of error.\n");
        sample_state.done = 1;
    }

    // Move to next step.
    sample_move_next_step(MSG_STATUS_WARN_BITPIPE_BUSY != status);
}

/// @copydoc bp_api_ip_http_post_write_ready_notify_post
void ip_http_post_write_ready_notify_post(uint8_t seq_nr, uint8_t http_conn_id)
{
    P_INFO("\n[%02X] HTTP POST WRITE READY NOTIFY. HttpConn=%d\n", seq_nr, http_conn_id);

    if(sample_state.conn_id_http == http_conn_id)
        sample_state.conn_status_http = IP_CONN_STATUS_ESTABLISHED;  // Can now send the POST payload.
}

/// @copydoc bp_api_ip_http_post_write_reply
void ip_http_post_write_reply(uint8_t seq_nr, uint8_t status)
{
    P_INFO("[%02X] HTTP POST WRITE reply. Status=%02X\n", seq_nr, status);

    if(MSG_STATUS_CODE_SUCCESS == status)
    {
        sample_state.http_post_data_idx += sample_state.http_post_last_pay_len;
    }

    sample_move_next_step(0);   // Do not move to next step here.
}

/// @copydoc bp_api_ip_http_rx_notify_request
void ip_http_rx_notify_request(uint8_t seq_nr, uint8_t http_conn_id, uint8_t data_len, uint8_t* data)
{
    P_INFO("[%02X] HTTP RX NOTIFY request. HttpConn=%d DataLen=%d\n", seq_nr, http_conn_id, data_len);

    // IMPORTANT: Reply to this request, or else the BitPipe will re-send this data chunk.
    bp_api_ip_http_rx_notify_ack(seq_nr);

    P_DEBUG("     > ");
#ifdef DBG_PRINT_RXBYTES_HEXADECIMAL
    uint8_t i;
    for(i=0; i<data_len; i++)
        P_DEBUG("%02X ", data[i]);
#else
    P_DEBUG("%.*s", data_len, data);
#endif
    P_DEBUG("\n");
}

/// @copydoc bp_api_ip_connection_close_reply
void ip_connection_close_reply(uint8_t seq_nr, uint8_t status)
{
    P_INFO("[%02X] IP CONNECTION CLOSE reply. Status=%02X ConnectionId=%02X\n", seq_nr, status, bp_status.ip_close_conn_id);

    if(MSG_STATUS_CODE_SUCCESS == status)
    {
        if(bp_status.ip_close_conn_id == sample_state.conn_id_tcp)
        {
            sample_state.conn_status_tcp = IP_CONN_STATUS_CLOSED;
            sample_state.conn_id_tcp = BP_API_IP_INVALID_CONNECTION_ID;
        }

        else if(bp_status.ip_close_conn_id == sample_state.conn_id_http)
        {
            sample_state.conn_status_http = IP_CONN_STATUS_CLOSED;
            sample_state.conn_id_http = BP_API_IP_INVALID_CONNECTION_ID;
        }
    }

    // Move to next step.
    sample_move_next_step(MSG_STATUS_CODE_SUCCESS == status);
}

/// @copydoc bp_api_ip_connection_close_notify_post
void ip_connection_close_notify_post(uint8_t seq_nr, uint8_t ip_conn_id, uint8_t reason)
{
    if(sample_state.conn_id_tcp == ip_conn_id)
    {
        P_INFO("[%02X] NOTIFICATION: TCP Connection was closed. Reason=%02X\n", seq_nr, reason);

        sample_state.conn_status_tcp = IP_CONN_STATUS_CLOSED;
        sample_state.conn_id_tcp = BP_API_IP_INVALID_CONNECTION_ID;
    }

    else if(sample_state.conn_id_http == ip_conn_id)
    {
        P_INFO("[%02X] NOTIFICATION: HTTP Connection was closed. Reason=%02X\n", seq_nr, reason);

        sample_state.conn_status_http = IP_CONN_STATUS_CLOSED;
        sample_state.conn_id_http = BP_API_IP_INVALID_CONNECTION_ID;
    }
}



//  MQTT API

/// @copydoc bp_api_mqtt_init_connection_reply
void mqtt_init_connection_reply(uint8_t seq_nr, uint8_t status, uint8_t mqtt_conn_id)
{
    P_INFO("[%02X] MQTT INITIATE CONNECTION reply. Status=%02X - MqttConn=%d\n", seq_nr, status, mqtt_conn_id);

    if(MSG_STATUS_CODE_SUCCESS == status)
    {
        sample_state.conn_id_mqtt = mqtt_conn_id;
        sample_state.conn_status_mqtt = IP_CONN_STATUS_OPEN;
    }

    // Move to next step.
    sample_move_next_step(MSG_STATUS_CODE_SUCCESS == status);
}

/// @copydoc bp_api_mqtt_connection_established_notify_post
void mqtt_connection_established_notify_post(uint8_t seq_nr, uint8_t mqtt_conn_id)
{
    P_INFO("\n[%02X] MQTT CONNECTION ESTABLISHED NOTIFY. MqttConn=%d\n", seq_nr, mqtt_conn_id);

    if(sample_state.conn_id_mqtt == mqtt_conn_id)
        sample_state.conn_status_mqtt = IP_CONN_STATUS_ESTABLISHED;
}

/// @copydoc bp_api_mqtt_subscribe_topics_reply
void mqtt_subscribe_topics_reply(uint8_t seq_nr, uint8_t status, uint8_t sub_status)
{
    P_INFO("[%02X] MQTT SUBSCRIBE TOPICS reply. Status=%02X Topic1Subscription=%s Topic2Subscription=%s Topic3Subscription=%s\n",
           seq_nr, status,
           (sub_status & 0x01) ? "Success" : "Fail",
           (sub_status & 0x02) ? "Success" : "Fail",
           (sub_status & 0x04) ? "Success" : "Fail");

    // Move to next step.
    sample_move_next_step(MSG_STATUS_CODE_SUCCESS == status);
}

/// @copydoc bp_api_mqtt_unsubscribe_topics_reply
void mqtt_unsubscribe_topics_reply(uint8_t seq_nr, uint8_t status)
{
    P_INFO("[%02X] MQTT UNSUBSCRIBE TOPICS reply. Status=%02X\n", seq_nr, status);

    // Move to next step.
    sample_move_next_step(MSG_STATUS_CODE_SUCCESS == status);
}

/// @copydoc bp_api_mqtt_publish_reply
void mqtt_publish_reply(uint8_t seq_nr, uint8_t status)
{
    P_INFO("[%02X] MQTT PUBLISH reply. Status=%02X\n", seq_nr, status);

    // Move to next step.
    sample_move_next_step(MSG_STATUS_CODE_SUCCESS == status);
}

/// @copydoc bp_api_mqtt_publication_rx_notify_request
void mqtt_publication_rx_notify_request(uint8_t seq_nr, uint8_t mqtt_conn_id, uint8_t qos, uint8_t retained,
                                               uint8_t complete, uint8_t topic_len, uint8_t* topic,
                                               uint8_t data_len, uint8_t* data)
{
    P_INFO("\n[%02X] MQTT PUBLICATION RX NOTIFY request. MqttConn=%d TopicLen=%d DataLen=%d\n", seq_nr, mqtt_conn_id, topic_len, data_len);

    // IMPORTANT: Reply to this request, or else the BitPipe will re-send this data chunk.
    bp_api_mqtt_publication_rx_notify_ack(seq_nr);

    P_DEBUG("     > Topic: \'%.*s\' qos=%d retained=%d complete=%d\n", topic_len, topic, qos, retained, complete);
#ifdef DBG_PRINT_RXBYTES_HEXADECIMAL
    uint8_t i;
    for(i=0; i<data_len; i++)
        P_DEBUG("%02X ", data[i]);
#else
    P_DEBUG("%.*s", data_len, data);
#endif
    P_DEBUG("\n");

    sample_state.mqtt_publication_received = complete;
}

/// @copydoc bp_api_mqtt_disconnect_reply
void mqtt_disconnect_reply(uint8_t seq_nr, uint8_t status)
{
    P_INFO("[%02X] MQTT DISCONNECT reply. Status=%02X\n", seq_nr, status);

    if(MSG_STATUS_CODE_SUCCESS == status)
    {
        sample_state.conn_status_mqtt = IP_CONN_STATUS_CLOSED;
        sample_state.conn_id_mqtt = BP_API_IP_INVALID_CONNECTION_ID;
    }

    // Move to next step.
    sample_move_next_step(MSG_STATUS_CODE_SUCCESS == status);
}

/// @copydoc bp_api_mqtt_disconnect_notify_post
void mqtt_disconnect_notify_post(uint8_t seq_nr, uint8_t mqtt_conn_id, uint8_t reason)
{
    if(sample_state.conn_id_mqtt == mqtt_conn_id)
    {
        P_INFO("[%02X] NOTIFICATION: MQTT Connection was closed. Reason=%02X\n", seq_nr, reason);

        sample_state.conn_status_mqtt = IP_CONN_STATUS_CLOSED;
        sample_state.conn_id_mqtt = BP_API_IP_INVALID_CONNECTION_ID;
    }
}



//  SMS API

/// @copydoc bp_api_sms_setup_transmission_reply
void sms_setup_transmission_reply(uint8_t seq_nr, uint8_t status)
{
    P_INFO("[%02X] SMS SETUP TRANSMISSION reply. Status=%02X\n", seq_nr, status);

    // Move to next step.
    sample_move_next_step(MSG_STATUS_CODE_SUCCESS == status);
}

/// @copydoc bp_api_sms_message_append_send_reply
void sms_message_append_send_reply(uint8_t seq_nr, uint8_t status)
{
    P_INFO("[%02X] SMS MESSAGE APPEND/SEND reply. Status=%02X\n", seq_nr, status);

    // Move to next step.
    sample_move_next_step(MSG_STATUS_CODE_SUCCESS == status);
}

/// @copydoc bp_api_sms_message_rx_notify_request
void sms_message_rx_notify_request(uint8_t seq_nr, uint8_t encoding, uint8_t year, uint8_t month, uint8_t day,
                                          uint8_t hour, uint8_t minute, uint8_t second, uint8_t tz_info_offset, uint8_t tz_info_sign,
                                          uint8_t msisdn_len, uint8_t* msisdn)
{
    // IMPORTANT: Reply to this request, or else the BitPipe will re-send this notification.
    bp_api_sms_message_rx_notify_ack(seq_nr);

    P_INFO("\n[%02X] SMS MESSAGE RECEPTION: Encoding=%d DateSent=%04d-%02d-%02d.%02dh%02dm%02ds UTC%s%d from %.*s:\n", seq_nr, encoding,
           (uint16_t)year+2000, month+1, day+1, hour, minute, second, tz_info_sign==0?"+":"-", (uint16_t)tz_info_offset / 4, msisdn_len, msisdn);
}

/// @copydoc bp_api_sms_message_part_notify_request
void sms_message_part_notify_request(uint8_t seq_nr, uint8_t last_part_indicator, uint8_t data_len, uint8_t* data)
{
    // IMPORTANT: Reply to this request, or else the BitPipe will re-send this notification.
    bp_api_sms_message_part_notify_ack(seq_nr);

    P_INFO(">%.*s%s\n", data_len, data, last_part_indicator ? "\nEND":"");

    sample_state.sms_received = last_part_indicator;
}

/**
 * @}       // Doxygen sample_common defgroup
 */
