/** ********************************************************************************************************************
 * @file        sample_common.h
 * @author      Briowireless, Inc.
 * @date        2016-11-11
 * @brief       Common defitnitions for all samples.
 * @copyright   Briowireless, Inc., 2016, Licensed under BSD license. See LICENSE file for more details.
 *
 ** *******************************************************************************************************************/
#ifndef __SAMPLE_COMMON_H__
#define __SAMPLE_COMMON_H__


// Include directives  -------------------------------------------------------------------------------------------------

#include <platform_defs.h>
#include <bp_api.h>
#include <time.h>                       // For statistics, and timeouts. Optional.


// Global Type definitions ---------------------------------------------------------------------------------------------

/// @brief IP connection status.
typedef enum __ip_connection_status {
    IP_CONN_STATUS_CLOSED       = 0,
    IP_CONN_STATUS_OPEN         = 1,
    IP_CONN_STATUS_ESTABLISHED  = 2,
} ip_conn_status;


/// @brief BitPipe statuses and connection identifiers.
struct __bp_status
{
    uint8_t         radio_ready;                ///< boolean: Specifies if the radio is started and ready.
    uint8_t         sim_status;                 ///< The status of the SIM card.
    uint8_t         network_registered;         ///< boolean: Are we registered on mobile network (either home or foreign).
    uint8_t         ip_close_conn_id;           ///< Used to know which IP connection id was sent in IP CONNECTION CLOSE request.
    time_t          pwr_on_start;               ///< Statistics: how long it took to start the radio and for it to be ready.
    time_t          low_power_entry_ts;         ///< Timestamp at which low power mode was activated.
};

/// @brief Sample execution state and flags.
struct __sample_state
{
    uint8_t  step;                              ///< Keeps track of where we are in the sample state machine.
    uint8_t  awaiting_reply;                    ///< boolean: Specifies if we're currently waiting on a reply from the BitPipe.
    uint8_t  done;                              ///< boolean: Triggers the end of the sample.

    uint8_t  mqtt_publication_received;         ///< boolean: Have we received an MQTT publication from the server.
    uint8_t  sms_received;                      ///< boolean: Have we received a SMS message.

    uint32_t tcp_data_rx_idx;                   ///< Number of bytes received by TCP RX NOTIFY

    uint8_t  http_post_last_pay_len;            ///< Payload length of the last HTTP POST WRITE request.
    uint32_t http_post_data_idx;                ///< Index in the HTTP POST data to send (in case of large payload).

    ip_conn_status  conn_status_tcp;            ///< Connection status for TCP
    ip_conn_status  conn_status_http;           ///< Connection status for HTTP
    ip_conn_status  conn_status_mqtt;           ///< Connection status for MQTT
    uint8_t         conn_id_tcp;                ///< Connection ID for TCP
    uint8_t         conn_id_http;               ///< Connection ID for HTTP
    uint8_t         conn_id_mqtt;               ///< Connection ID for MQTT
};


// Global variables ----------------------------------------------------------------------------------------------------

extern struct __bp_status       bp_status;          ///< Statuses and connection identifiers - for this sample.
extern struct __sample_state    sample_state;       ///< The current state of the sample.
extern bp_api_callbacks         sample_callbacks;   ///< Sample callbacks that will be registered in the BitPipe Host SDK.


// Forward definitions -------------------------------------------------------------------------------------------------


// THESE ARE APPLICATION CALLBACKS FOR THE SAMPLES

// SYSTEM
void sys_error_report_post(uint8_t seq_nr, uint8_t facility, uint8_t err_nr);
void sys_set_max_payload_sz_reply(uint8_t seq_nr, uint8_t status);
void sys_set_uart_speed_reply(uint8_t seq_nr, uint8_t status);

// DEVICE INFO
void devi_get_hw_version_reply(uint8_t seq_nr, uint8_t status, uint8_t hw_ver_len, uint8_t* hw_ver);
void devi_get_fw_version_reply(uint8_t seq_nr, uint8_t status, uint16_t boot_ver, uint8_t api_ver,
                                      uint8_t bp_ver_len, uint8_t* bp_ver, uint8_t radio_ver_len, uint8_t* radio_ver);
void devi_get_product_name_reply(uint8_t seq_nr, uint8_t status, uint8_t prod_name_len, uint8_t* prod_name);
void devi_get_vin_vio_voltage_reply(uint8_t seq_nr, uint8_t status, uint16_t vin_voltage, uint16_t vio_voltage);
void devi_get_junction_temp_reply(uint8_t seq_nr, uint8_t status, int8_t junction_temp);
void devi_set_low_power_reply(uint8_t seq_nr, uint8_t status);
void devi_get_latest_fw_version_reply(uint8_t seq_nr, uint8_t status, uint8_t fw_ver_len, uint8_t* fw_ver);
void devi_update_fw_reply(uint8_t seq_nr, uint8_t status);
void devi_get_autonomous_mode_reply(uint8_t seq_nr, uint8_t status, uint8_t secure, uint16_t ka_ival, uint16_t port,
                                           uint8_t host_len, uint8_t* host, uint8_t user_len, uint8_t* user,
                                           uint8_t pass_len, uint8_t* pass, uint8_t gpfx_len, uint8_t* gpfx);
void devi_set_autonomous_mode_reply(uint8_t seq_nr, uint8_t status);

// MOBILE INFORMATION
void mobi_power_on_radio_reply(uint8_t seq_nr, uint8_t status);
void mobi_power_off_radio_reply(uint8_t seq_nr, uint8_t status);
void mobi_radio_start_notify_post(uint8_t seq_nr, uint8_t notify_reason, uint8_t sim_status);
void mobi_get_radio_status_reply(uint8_t seq_nr, uint8_t status, uint8_t radio_status, uint8_t sim_status);
void mobi_get_imei_reply(uint8_t seq_nr, uint8_t status, uint8_t* imei);
void mobi_get_iccid_reply(uint8_t seq_nr, uint8_t status, uint8_t* iccid);
void mobi_get_imsi_reply(uint8_t seq_nr, uint8_t status, uint8_t* imsi);
void mobi_get_own_phone_number_reply(uint8_t seq_nr, uint8_t status, uint8_t msisdn_len, uint8_t* msisdn);
void mobi_get_apn_config_reply(uint8_t seq_nr, uint8_t status, uint8_t host_len, uint8_t* host,
                                      uint8_t user_len, uint8_t* user, uint8_t pass_len, uint8_t* pass);
void mobi_set_apn_config_reply(uint8_t seq_nr, uint8_t status);
void mobi_set_sim_access_config_reply(uint8_t seq_nr, uint8_t status);
void mobi_get_mobile_conn_status_reply(uint8_t seq_nr, uint8_t status, uint8_t conn_ind, uint8_t link_tech,
                                              uint16_t net_lac, uint32_t net_cell_id, uint8_t signal_strength,
                                              uint8_t* net_mcc, uint8_t* net_mnc);
void mobi_mobile_conn_status_change_notify_post(uint8_t seq_nr, uint8_t conn_ind, uint8_t link_tech,
                                                       uint16_t net_lac, uint32_t net_cell_id);
void mobi_get_mobile_date_time_reply(uint8_t seq_nr, uint8_t status, uint8_t year, uint8_t month, uint8_t day,
                                            uint8_t hour, uint8_t minute, uint8_t second);

// IP CONNECTIVITY
void ip_tcp_init_connection_reply(uint8_t seq_nr, uint8_t status, uint8_t tcp_conn_id);
void ip_tcp_connection_established_notify_post(uint8_t seq_nr, uint8_t tcp_conn_id);
void ip_tcp_write_append_reply(uint8_t seq_nr, uint8_t status, uint16_t buf_sz_current, uint16_t buf_sz_max);
void ip_tcp_write_send_reply(uint8_t seq_nr, uint8_t status);
void ip_tcp_rx_notify_request(uint8_t seq_nr, uint8_t tcp_conn_id, uint8_t data_len, uint8_t* data);
void ip_http_init_connection_reply(uint8_t seq_nr, uint8_t status, uint8_t http_conn_id);
void ip_http_post_write_ready_notify_post(uint8_t seq_nr, uint8_t http_conn_id);
void ip_http_post_write_reply(uint8_t seq_nr, uint8_t status);
void ip_http_rx_notify_request(uint8_t seq_nr, uint8_t http_conn_id, uint8_t data_len, uint8_t* data);
void ip_connection_close_reply(uint8_t seq_nr, uint8_t status);
void ip_connection_close_notify_post(uint8_t seq_nr, uint8_t ip_conn_id, uint8_t reason);

// MQTT
void mqtt_init_connection_reply(uint8_t seq_nr, uint8_t status, uint8_t mqtt_conn_id);
void mqtt_connection_established_notify_post(uint8_t seq_nr, uint8_t mqtt_conn_id);
void mqtt_subscribe_topics_reply(uint8_t seq_nr, uint8_t status, uint8_t sub_status);
void mqtt_unsubscribe_topics_reply(uint8_t seq_nr, uint8_t status);
void mqtt_publish_reply(uint8_t seq_nr, uint8_t status);
void mqtt_publication_rx_notify_request(uint8_t seq_nr, uint8_t mqtt_conn_id, uint8_t qos, uint8_t retained,
                                               uint8_t complete, uint8_t topic_len, uint8_t* topic,
                                               uint8_t data_len, uint8_t* data);
void mqtt_disconnect_reply(uint8_t seq_nr, uint8_t status);
void mqtt_disconnect_notify_post(uint8_t seq_nr, uint8_t mqtt_conn_id, uint8_t reason);

// SMS
void sms_setup_transmission_reply(uint8_t seq_nr, uint8_t status);
void sms_message_append_send_reply(uint8_t seq_nr, uint8_t status);
void sms_message_rx_notify_request(uint8_t seq_nr, uint8_t encoding,
                                          uint8_t year, uint8_t month, uint8_t day,
                                          uint8_t hour, uint8_t minute, uint8_t second,
                                          uint8_t tz_info_offset, uint8_t tz_info_sign,
                                          uint8_t msisdn_len, uint8_t* msisdn);
void sms_message_part_notify_request(uint8_t seq_nr, uint8_t last_part_indicator,
                                            uint8_t data_len, uint8_t* data);


#endif // __SAMPLE_COMMON_H__
