/** ********************************************************************************************************************
 * @file        bp_api.c
 * @author      Briowireless, Inc.
 * @date        2016-10-28
 * @brief       This file defines the internal function handlers for each message the BitPipe may send to the Host.
 *              It handles serialization and deserialization of binary frames received from and sent to the BitPipe.
 *              It handles invoking the application-registered callback after message reception.
 * @version     API V1
 * @copyright   Briowireless, Inc., 2016, Licensed under BSD license. See LICENSE file for more details.
 *
 ** *******************************************************************************************************************/

// Include directives  -------------------------------------------------------------------------------------------------

#include <bp_api.h>
#include <message_defs.h>


// Preprocessor definitions --------------------------------------------------------------------------------------------

/// @brief X-MACRO which defines message IDs, the handling internal function, the expected message type and the minimal
///        message payload data len for each message.
///
/// @see   The *BitPipe Serial API* document to have more information on messages and their parameters.
///
#define X_BP_RX_MSG_HANDLERS_TABLE \
    X_MSG_HANDLER( MSG_ID_ERROR_REPORT,                         process_rx_error_report_post,                       MSG_STATUS_REQUEST, 2 ) \
    X_MSG_HANDLER( MSG_ID_SET_MAX_MSG_PAYSIZE,                  process_rx_set_max_msg_payload_sz_reply,            MSG_STATUS_REPLY,   0 ) \
    X_MSG_HANDLER( MSG_ID_SET_UART_SPEED,                       process_rx_set_uart_speed_reply,                    MSG_STATUS_REPLY,   0 ) \
    X_MSG_HANDLER( MSG_ID_GET_HW_VERSION,                       process_rx_get_hw_version_reply,                    MSG_STATUS_REPLY,   1 ) \
    X_MSG_HANDLER( MSG_ID_GET_FW_VERSION,                       process_rx_get_fw_version_reply,                    MSG_STATUS_REPLY,   5 ) \
    X_MSG_HANDLER( MSG_ID_GET_PRODUCT_NAME,                     process_rx_get_product_name_reply,                  MSG_STATUS_REPLY,   1 ) \
    X_MSG_HANDLER( MSG_ID_GET_VIN_VIO_VOLTAGE,                  process_rx_get_vin_vio_voltage_reply,               MSG_STATUS_REPLY,   4 ) \
    X_MSG_HANDLER( MSG_ID_GET_JUNCTION_TEMP,                    process_rx_get_junction_temp_reply,                 MSG_STATUS_REPLY,   1 ) \
    X_MSG_HANDLER( MSG_ID_SET_LOW_POWER_MODE,                   process_rx_set_low_power_reply,                     MSG_STATUS_REPLY,   0 ) \
    X_MSG_HANDLER( MSG_ID_GET_LATEST_FW_VERSION,                process_rx_get_latest_fw_version_reply,             MSG_STATUS_REPLY,   1 ) \
    X_MSG_HANDLER( MSG_ID_UPDATE_FIRMWARE,                      process_rx_update_firmware_reply,                   MSG_STATUS_REPLY,   0 ) \
    X_MSG_HANDLER( MSG_ID_GET_AUTONOMOUS_CONFIG,                process_rx_get_autonomous_config_reply,             MSG_STATUS_REPLY,   9 ) \
    X_MSG_HANDLER( MSG_ID_SET_AUTONOMOUS_CONFIG,                process_rx_set_autonomous_config_reply,             MSG_STATUS_REPLY,   0 ) \
    X_MSG_HANDLER( MSG_ID_POWER_ON_RADIO,                       process_rx_power_on_radio_reply,                    MSG_STATUS_REPLY,   0 ) \
    X_MSG_HANDLER( MSG_ID_POWER_OFF_RADIO,                      process_rx_power_off_radio_reply,                   MSG_STATUS_REPLY,   0 ) \
    X_MSG_HANDLER( MSG_ID_RADIO_START_NOTIFY,                   process_rx_radio_start_notify_post,                 MSG_STATUS_REQUEST, 2 ) \
    X_MSG_HANDLER( MSG_ID_GET_RADIO_STATUS,                     process_rx_get_radio_status_reply,                  MSG_STATUS_REPLY,   2 ) \
    X_MSG_HANDLER( MSG_ID_GET_IMEI,                             process_rx_get_imei_reply,                          MSG_STATUS_REPLY,   15 ) \
    X_MSG_HANDLER( MSG_ID_GET_ICCID,                            process_rx_get_iccid_reply,                         MSG_STATUS_REPLY,   20 ) \
    X_MSG_HANDLER( MSG_ID_GET_IMSI,                             process_rx_get_imsi_reply,                          MSG_STATUS_REPLY,   15 ) \
    X_MSG_HANDLER( MSG_ID_GET_OWN_PHONE_NUMBER,                 process_rx_get_own_phone_num_reply,                 MSG_STATUS_REPLY,   1 ) \
    X_MSG_HANDLER( MSG_ID_GET_APN_CONFIG,                       process_rx_get_apn_config_reply,                    MSG_STATUS_REPLY,   3 ) \
    X_MSG_HANDLER( MSG_ID_SET_APN_CONFIG,                       process_rx_set_apn_config_reply,                    MSG_STATUS_REPLY,   0 ) \
    X_MSG_HANDLER( MSG_ID_SET_SIM_ACCESS_CONFIG,                process_rx_set_sim_axs_config_reply,                MSG_STATUS_REPLY,   0 ) \
    X_MSG_HANDLER( MSG_ID_GET_MOBILE_CONN_STATUS,               process_rx_get_mobile_conn_status_reply,            MSG_STATUS_REPLY,   17 ) \
    X_MSG_HANDLER( MSG_ID_MOBILE_CONN_STATUS_NOTIFY,            process_rx_mobile_conn_status_notify_post,          MSG_STATUS_REQUEST, 8 ) \
    X_MSG_HANDLER( MSG_ID_GET_MOBILE_DATE_AND_TIME,             process_rx_get_mobile_date_and_time_reply,          MSG_STATUS_REPLY,   6 ) \
    X_MSG_HANDLER( MSG_ID_TCP_INITIATE_CONNECTION,              process_rx_tcp_init_connection_reply,               MSG_STATUS_REPLY,   1 ) \
    X_MSG_HANDLER( MSG_ID_TCP_CONNECTION_ESTABLISHED_NOTIFY,    process_rx_tcp_connection_established_notify_post,  MSG_STATUS_REQUEST, 1 ) \
    X_MSG_HANDLER( MSG_ID_TCP_WRITE_APPEND,                     process_rx_tcp_write_append_reply,                  MSG_STATUS_REPLY,   4 ) \
    X_MSG_HANDLER( MSG_ID_TCP_WRITE_SEND,                       process_rx_tcp_write_send_reply,                    MSG_STATUS_REPLY,   0 ) \
    X_MSG_HANDLER( MSG_ID_TCP_RX_NOTIFY,                        process_rx_tcp_rx_notify_request,                   MSG_STATUS_REQUEST, 2 ) \
    X_MSG_HANDLER( MSG_ID_HTTP_INITIATE_CONNECTION,             process_rx_http_init_connection_reply,              MSG_STATUS_REPLY,   1 ) \
    X_MSG_HANDLER( MSG_ID_HTTP_POST_WRITE_READY_NOTIFY,         process_rx_http_post_write_ready_notify_post,       MSG_STATUS_REQUEST, 1 ) \
    X_MSG_HANDLER( MSG_ID_HTTP_POST_WRITE,                      process_rx_http_post_write_reply,                   MSG_STATUS_REPLY,   0 ) \
    X_MSG_HANDLER( MSG_ID_HTTP_RX_NOTIFY,                       process_rx_http_rx_notify_request,                  MSG_STATUS_REQUEST, 2 ) \
    X_MSG_HANDLER( MSG_ID_INET_CLOSE,                           process_rx_inet_close_reply,                        MSG_STATUS_REPLY,   0 ) \
    X_MSG_HANDLER( MSG_ID_INET_CLOSE_NOTIFY,                    process_rx_inet_close_notify_post,                  MSG_STATUS_REQUEST, 2 ) \
    X_MSG_HANDLER( MSG_ID_MQTT_INITIATE_CONNECTION,             process_rx_mqtt_init_connection_reply,              MSG_STATUS_REPLY,   1 ) \
    X_MSG_HANDLER( MSG_ID_MQTT_CONNECTION_ESTABLISHED_NOTIFY,   process_rx_mqtt_connection_established_notify_post, MSG_STATUS_REQUEST, 1 ) \
    X_MSG_HANDLER( MSG_ID_MQTT_SUBSCRIBE_TOPICS,                process_rx_mqtt_subscribe_topics_reply,             MSG_STATUS_REPLY,   1 ) \
    X_MSG_HANDLER( MSG_ID_MQTT_UNSUBSCRIBE_TOPICS,              process_rx_mqtt_unsubscribe_topics_reply,           MSG_STATUS_REPLY,   0 ) \
    X_MSG_HANDLER( MSG_ID_MQTT_PUBLISH,                         process_rx_mqtt_publish_reply,                      MSG_STATUS_REPLY,   0 ) \
    X_MSG_HANDLER( MSG_ID_MQTT_PUBLICATION_RX_NOTIFY,           process_rx_mqtt_publication_rx_notify_request,      MSG_STATUS_REQUEST, 4 ) \
    X_MSG_HANDLER( MSG_ID_MQTT_DISCONNECT,                      process_rx_mqtt_disconnect_reply,                   MSG_STATUS_REPLY,   0 ) \
    X_MSG_HANDLER( MSG_ID_MQTT_DISCONNECT_NOTIFY,               process_rx_mqtt_disconnect_notify_post,             MSG_STATUS_REQUEST, 2 ) \
    X_MSG_HANDLER( MSG_ID_SMS_TX_SETUP,                         process_rx_sms_tx_setup_reply,                      MSG_STATUS_REPLY,   0 ) \
    X_MSG_HANDLER( MSG_ID_SMS_TX_APPEND_SEND,                   process_rx_sms_tx_append_send_reply,                MSG_STATUS_REPLY,   0 ) \
    X_MSG_HANDLER( MSG_ID_SMS_RX_NOTIFY,                        process_rx_sms_rx_notify_request,                   MSG_STATUS_REQUEST, 9 ) \
    X_MSG_HANDLER( MSG_ID_SMS_RX_PART_NOTIFY,                   process_rx_sms_rx_part_notify_request,              MSG_STATUS_REQUEST, 2 )

/// @brief Used in every handler for message received from the BitPipe. Ensures the SDK is initialized properly
///        and that the application callback is defined. Also validates the expected message type and data length.
#define RX_HANDLER_BASIC_CHECKS(APP_CB) \
    if(NULL == bp_sdk_params  ||  NULL == bp_sdk_params->app_callbacks || NULL == APP_CB) \
    { \
        return; \
    } \
    \
    if(exp_msg_type != MSG_STATUS_GET_MSGTYPE(msg->status) ||   \
       exp_data_len > msg->data_len)                            \
    { \
        return; \
    }


// Local functions prototypes ------------------------------------------------------------------------------------------

inline void process_rx_message(const mp_message* msg);

/// @brief For defining prototypes for internal message handlers.
#define X_MSG_HANDLER(ID, FCT, EXPECTED_MSG_TYPE, EXPECTED_DATA_LEN)  static void FCT(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len);
    X_BP_RX_MSG_HANDLERS_TABLE
#undef X_MSG_HANDLER


// Local variables -----------------------------------------------------------------------------------------------------

static bp_api_init_params*  bp_sdk_params;          ///< @brief Runtime parameters (Callbacks, function pointers, etc).
static mp_context           bp_mp_ctx;              ///< @brief Message Protocol context, for communication with BitPipe.


// Global variables ----------------------------------------------------------------------------------------------------

mp_seq_nr bp_req_seq_nr = 0x00;                     ///< @brief Sequence number, for outgoing request messages,
                                                    ///         incremented by SDK. Ideally should be initialized randomly.


// Global (extern) function  -------------------------------------------------------------------------------------------

/** ********************************************************************************************************************
 * @brief  Registers the application callbacks in the SDK layer, initializes SDK internals.
 *
 * @param  init_params  Pointer to structure containing SDK initialization data.
 *
 * @retval uint8_t      0 indicates success;
 *                      1 indicates an invalid parameter was supplied;
 *                      2 indicates the deserialization buffer provided doesn't suit 'message_protocol' requirements.
 */
uint8_t bp_api_init(bp_api_init_params* init_params)
{
    mp_status status;

    if(NULL == init_params                  ||
       NULL == init_params->app_callbacks   ||
       NULL == init_params->uart_write      ||
       NULL == init_params->serialize_buf   ||
       NULL == init_params->deserialize_buf)
    {
        return 1;
    }


    // Initialize the internal binary protocol parser.
    status = message_protocol_init(init_params->deserialize_buf, init_params->deserialize_buf_len, &bp_mp_ctx);
    if(MESSAGE_PROTOCOL_OK != status)
    {
        return 2;
    }

    // Keep initialization parameters internally.
    bp_sdk_params = init_params;

    return 0;
}


/** ********************************************************************************************************************
 * @brief  Processes data received from the BitPipe through the UART connection.
 *
 *         If a message is completed, it is processed and will callback into application using registered callback.
 *         Processing of data stops after a message is complete. If this is the case you may re-invoke this function
 *         until all data has been consumed.
 *
 * @param  data_len     Length of data received from the BitPipe.
 * @param  data         Data received from the BitPipe.
 * @param[out] consumed_len Number of bytes consumed from the 'data' buffer, after this function has returned.
 *
 * @retval uint8_t      0 indicates success;
 *                      1 indicates an error occurred while pushing bytes in the binary protocol machine.
 */
uint8_t bp_api_process_incoming(const uint8_t* data, uint16_t data_len, uint16_t *consumed_len)
{
    mp_status status;
    mp_message message;

    for(*consumed_len=0; (*consumed_len)<data_len; (*consumed_len)++)
    {
        status = message_protocol_push_data(bp_mp_ctx, data[*consumed_len]);
        switch(status)
        {
        case MESSAGE_PROTOCOL_OK:
            // Byte processed correctly; message not yet complete.
            break;

        case MESSAGE_PROTOCOL_MESSAGE_READY:
            status = message_protocol_retrieve(bp_mp_ctx, &message);
            if(MESSAGE_PROTOCOL_OK == status)
            {
                // Will validate message and callback within application-defined function.
                process_rx_message(&message);
            }
            else
            {
                P_WARN("bp_api_process_incoming: Failed to retrieve message. Status=%d", status);
            }
            break;

        case MESSAGE_PROTOCOL_ERROR_CHECKSUM:
            P_ERR("bp_api_process_incoming: CHECKSUM error in message from BitPipe. This should not occur. Check UART link.\n");
            return 1;

        case MESSAGE_PROTOCOL_ERROR_BUF_TOO_SMALL:
            P_ERR("bp_api_process_incoming: Received a message from the BitPipe that doesn't fit in the message protocol buffer.\n" \
                  "                         Increase the size of 'deserialize_buf', or use 'bp_api_sys_set_max_payload_sz()' to\n" \
                  "                         limit the size of variable messages sent from the BitPipe.\n");
            return 1;

        case MESSAGE_PROTOCOL_ERROR_INDEX:
        case MESSAGE_PROTOCOL_ERROR_UNKOWN_STATE:
        default:
            P_ERR("bp_api_process_incoming: Invalid message_protocol status after push. State: %d\n", status);
            return 1;
        }
    }

    return 0;
}


/** ********************************************************************************************************************
 * @brief Sends a message to the BitPipe.
 * @see   The *BitPipe Serial API* document to have more information on messages and their parameters.
 *
 * @param  msg_id   Identifier of the message (also known as command identifider). Defines the content of the message.
 * @param  seq_nr   Sequence number of the message, used to detect duplicates at both endpoints.
 * @param  status   Status byte of the message. Bit 7 is used to differentiate REQUESTS / REPLIES.
 * @param  data     Data to send along with the message.
 * @param  data_len Length of the data that accompanies this message.
 *
 * @retval uint8_t  0 indicates the message was sent successfully.
 *                  1 indicates that the SDK layer is not initialized properly.
 *                  2 indicates that the serialization buffer provided is too small to contain message.
 *                  3 indicates an other error occurred while serializing the message (generic).
 */
uint8_t bp_api_send_message(uint8_t msg_id, mp_seq_nr seq_nr, uint8_t status, const uint8_t* data, uint16_t data_len)
{
    mp_message message = { msg_id, seq_nr, status, (uint8_t*)data, data_len };
    mp_status  encode_status;
    uint16_t   msg_len;

    if(NULL == bp_sdk_params->uart_write || NULL == bp_sdk_params->serialize_buf || MESSAGE_PROTOCOL_INVALID_CONTEXT == bp_mp_ctx)
    {
        return 1;
    }

    // This step here will create the full frame within the linear buffer (serialize_buf), along
    //  with escape characters (when required).
    /// POSSIBLE OPTIMIZATION: To save the usage of this buffer, you could directly write to the UART port
    ///                        from within 'message_protocol_serialize'.
    encode_status = message_protocol_serialize(&message, bp_sdk_params->serialize_buf, bp_sdk_params->serialize_buf_len, &msg_len);
    if(MESSAGE_PROTOCOL_OK == encode_status)
    {
        // Write the full-framed binary message from linear buffer to the UART port.
        bp_sdk_params->uart_write(bp_sdk_params->serialize_buf, msg_len);
    }
    else if(MESSAGE_PROTOCOL_ERROR_BUF_TOO_SMALL == encode_status)
    {
        return 2;
    }
    else
    {
        return 3;
    }

    return 0;
}


/** ********************************************************************************************************************
 * @brief  [LOCAL] Checks the message and dispatches it to the correct handler to complete deserialization.
 *                 Those handles will, in turn, invoke the app-registered callback.
 *
 * @param  msg  Message received from the BitPipe.
 *
 * @retval None.
 */
inline void process_rx_message(const mp_message* msg)
{
    switch(msg->cmd_id)
    {

/// @brief For generating switch cases for every known message identifier, and invoking the correct internal message handler.
#define X_MSG_HANDLER(ID, FCT, EXPECTED_MSG_TYPE, EXPECTED_DATA_LEN)  \
    case ID: FCT(msg, EXPECTED_MSG_TYPE, EXPECTED_DATA_LEN); break;

    X_BP_RX_MSG_HANDLERS_TABLE
#undef X_MSG_HANDLER

    default:
        // This SDK doesn't know this message ID. Is the BitPipe FW in sync with this SDK?
        P_WARN("process_rx_message: Don't know this message(ID=%02X). Check BitPipe FW .vs. SDK version.\n", msg->cmd_id);
        break;
    }
}


// ---------------------------------------------------------------------------------------------------------------------
// ---------------------------------------------------------------------------------------------------------------------

//                       LOW-LEVEL MESSAGE HANDLERS FOR MESSAGES RECEIVED FROM THE BITPIPE


// ---------------------------------------------------------------------------------------------------------------------
// ---------------------------------------------------------------------------------------------------------------------

// Copy-paste the following header for each function below:

/** ********************************************************************************************************************
 * @brief  Handles the incoming message from the BitPipe.
 *         Every function part of this group separates the message data into message-specific variables, and invokes
 *         the application callback.
 *
 * @param  msg          Message received from the BitPipe.
 * @param  exp_msg_type The expected message type (Request or response). See X_BP_RX_MSG_HANDLERS_TABLE.
 * @param  exp_data_len The minimum expected data length for this message. See X_BP_RX_MSG_HANDLERS_TABLE.
 *
 * @retval None.
 */


static void process_rx_error_report_post(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->sys_error_report_post);

    bp_sdk_params->app_callbacks->sys_error_report_post(msg->seq_nr, msg->data[0], msg->data[1]);
}

static void process_rx_set_max_msg_payload_sz_reply(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->sys_set_max_payload_sz_reply);

    bp_sdk_params->app_callbacks->sys_set_max_payload_sz_reply(msg->seq_nr, MSG_STATUS_GET_CODE(msg->status));
}

static void process_rx_set_uart_speed_reply(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->sys_set_uart_speed_reply);

    bp_sdk_params->app_callbacks->sys_set_uart_speed_reply(msg->seq_nr, MSG_STATUS_GET_CODE(msg->status));
}

static void process_rx_get_hw_version_reply(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->devi_get_hw_version_reply);

    bp_sdk_params->app_callbacks->devi_get_hw_version_reply(msg->seq_nr, MSG_STATUS_GET_CODE(msg->status),
                                                            msg->data[0], &msg->data[1]);
}

static void process_rx_get_fw_version_reply(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->devi_get_fw_version_reply);

    uint8_t bp_ver_len    = msg->data[3];
    uint8_t radio_ver_len = msg->data[4+bp_ver_len];

    bp_sdk_params->app_callbacks->devi_get_fw_version_reply(msg->seq_nr, MSG_STATUS_GET_CODE(msg->status),
                                                            msg->data[0] << 8 | msg->data[1],
                                                            msg->data[2],
                                                            bp_ver_len, &msg->data[4],
                                                            radio_ver_len, &msg->data[5+bp_ver_len]);
}

static void process_rx_get_product_name_reply(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->devi_get_product_name_reply);

    bp_sdk_params->app_callbacks->devi_get_product_name_reply(msg->seq_nr, MSG_STATUS_GET_CODE(msg->status),
                                                              msg->data[0], &msg->data[1]);
}

static void process_rx_get_vin_vio_voltage_reply(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->devi_get_vin_vio_voltage_reply);

    bp_sdk_params->app_callbacks->devi_get_vin_vio_voltage_reply(msg->seq_nr, MSG_STATUS_GET_CODE(msg->status),
                                                            msg->data[0] << 8 | msg->data[1],
                                                            msg->data[2] << 8 | msg->data[3]);
}

static void process_rx_get_junction_temp_reply(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->devi_get_junction_temp_reply);

    bp_sdk_params->app_callbacks->devi_get_junction_temp_reply(msg->seq_nr, MSG_STATUS_GET_CODE(msg->status),
                                                               (int8_t)msg->data[0]);
}

static void process_rx_set_low_power_reply(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->devi_set_low_power_reply);

    bp_sdk_params->app_callbacks->devi_set_low_power_reply(msg->seq_nr, MSG_STATUS_GET_CODE(msg->status));
}

static void process_rx_get_latest_fw_version_reply(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->devi_get_latest_fw_version_reply);

    bp_sdk_params->app_callbacks->devi_get_latest_fw_version_reply(msg->seq_nr, MSG_STATUS_GET_CODE(msg->status),
                                                                   msg->data[0], &msg->data[1]);
}

static void process_rx_update_firmware_reply(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->devi_update_fw_reply);

    bp_sdk_params->app_callbacks->devi_update_fw_reply(msg->seq_nr, MSG_STATUS_GET_CODE(msg->status));
}

static void process_rx_get_autonomous_config_reply(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->devi_get_autonomous_mode_config_reply);

    uint8_t host_len    = msg->data[5];
    uint8_t user_len    = msg->data[6+host_len];
    uint8_t pass_len    = msg->data[7+host_len+user_len];
    uint8_t gpfx_len    = msg->data[8+host_len+user_len+pass_len];

    bp_sdk_params->app_callbacks->devi_get_autonomous_mode_config_reply(msg->seq_nr, MSG_STATUS_GET_CODE(msg->status),
                                                                        msg->data[0],
                                                                        msg->data[1] << 8 | msg->data[2],
                                                                        msg->data[3] << 8 | msg->data[4],
                                                                        host_len, &msg->data[6],
                                                                        user_len, &msg->data[7+host_len],
                                                                        pass_len, &msg->data[8+host_len+user_len],
                                                                        gpfx_len, &msg->data[9+host_len+user_len+pass_len]);
}

static void process_rx_set_autonomous_config_reply(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->devi_set_autonomous_mode_config_reply);

    bp_sdk_params->app_callbacks->devi_set_autonomous_mode_config_reply(msg->seq_nr, MSG_STATUS_GET_CODE(msg->status));
}

static void process_rx_power_on_radio_reply(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->mobi_power_on_radio_reply);

    bp_sdk_params->app_callbacks->mobi_power_on_radio_reply(msg->seq_nr, MSG_STATUS_GET_CODE(msg->status));
}

static void process_rx_power_off_radio_reply(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->mobi_power_off_radio_reply);

    bp_sdk_params->app_callbacks->mobi_power_off_radio_reply(msg->seq_nr, MSG_STATUS_GET_CODE(msg->status));
}

static void process_rx_radio_start_notify_post(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->mobi_radio_start_notify_post);

    bp_sdk_params->app_callbacks->mobi_radio_start_notify_post(msg->seq_nr, msg->data[0], msg->data[1]);
}

static void process_rx_get_radio_status_reply(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->mobi_get_radio_status_reply);

    bp_sdk_params->app_callbacks->mobi_get_radio_status_reply(msg->seq_nr, MSG_STATUS_GET_CODE(msg->status), msg->data[0], msg->data[1]);
}

static void process_rx_get_imei_reply(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->mobi_get_imei_reply);

    bp_sdk_params->app_callbacks->mobi_get_imei_reply(msg->seq_nr, MSG_STATUS_GET_CODE(msg->status), &msg->data[0]);
}

static void process_rx_get_iccid_reply(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->mobi_get_iccid_reply);

    bp_sdk_params->app_callbacks->mobi_get_iccid_reply(msg->seq_nr, MSG_STATUS_GET_CODE(msg->status), &msg->data[0]);
}

static void process_rx_get_imsi_reply(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->mobi_get_imsi_reply);

    bp_sdk_params->app_callbacks->mobi_get_imsi_reply(msg->seq_nr, MSG_STATUS_GET_CODE(msg->status), &msg->data[0]);
}

static void process_rx_get_own_phone_num_reply(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->mobi_get_own_phone_number_reply);

    bp_sdk_params->app_callbacks->mobi_get_own_phone_number_reply(msg->seq_nr, MSG_STATUS_GET_CODE(msg->status), msg->data[0], &msg->data[1]);
}

static void process_rx_get_apn_config_reply(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->mobi_get_apn_config_reply);

    uint8_t host_len    = msg->data[0];
    uint8_t user_len    = msg->data[1+host_len];
    uint8_t pass_len    = msg->data[2+host_len+user_len];

    bp_sdk_params->app_callbacks->mobi_get_apn_config_reply(msg->seq_nr, MSG_STATUS_GET_CODE(msg->status),
                                                            host_len, &msg->data[1],
                                                            user_len, &msg->data[2+host_len],
                                                            pass_len, &msg->data[3+host_len+user_len]);
}

static void process_rx_set_apn_config_reply(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->mobi_set_apn_config_reply);

    bp_sdk_params->app_callbacks->mobi_set_apn_config_reply(msg->seq_nr, MSG_STATUS_GET_CODE(msg->status));
}


static void process_rx_set_sim_axs_config_reply(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->mobi_set_sim_access_config_reply);

    bp_sdk_params->app_callbacks->mobi_set_sim_access_config_reply(msg->seq_nr, MSG_STATUS_GET_CODE(msg->status));
}

static void process_rx_get_mobile_conn_status_reply(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->mobi_get_mobile_conn_status_reply);

    bp_sdk_params->app_callbacks->mobi_get_mobile_conn_status_reply(msg->seq_nr, MSG_STATUS_GET_CODE(msg->status),
                                                                    msg->data[0], msg->data[1],
                                                                    msg->data[2] << 8 | msg->data[3],
                                                                    msg->data[4] << 24 | msg->data[5] << 16 | msg->data[6] << 8 | msg->data[7],
                                                                    msg->data[8],
                                                                    &msg->data[9],
                                                                    &msg->data[13]);
}

static void process_rx_mobile_conn_status_notify_post(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->mobi_mobile_conn_status_change_notify_post);

    bp_sdk_params->app_callbacks->mobi_mobile_conn_status_change_notify_post(msg->seq_nr,
                                                                             msg->data[0], msg->data[1],
                                                                             msg->data[2] << 8 | msg->data[3],
                                                                             msg->data[4] << 24 | msg->data[5] << 16 | msg->data[6] << 8 | msg->data[7]);
}

static void process_rx_get_mobile_date_and_time_reply(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->mobi_get_mobile_date_time_reply);

    bp_sdk_params->app_callbacks->mobi_get_mobile_date_time_reply(msg->seq_nr, MSG_STATUS_GET_CODE(msg->status),
                                                                  msg->data[0], msg->data[1], msg->data[2],
                                                                  msg->data[3], msg->data[4], msg->data[5]);
}

static void process_rx_tcp_init_connection_reply(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->ip_tcp_init_connection_reply);

    bp_sdk_params->app_callbacks->ip_tcp_init_connection_reply(msg->seq_nr, MSG_STATUS_GET_CODE(msg->status),
                                                               msg->data[0]);
}

static void process_rx_tcp_connection_established_notify_post(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->ip_tcp_connection_established_notify_post);

    bp_sdk_params->app_callbacks->ip_tcp_connection_established_notify_post(msg->seq_nr, msg->data[0]);
}

static void process_rx_tcp_write_append_reply(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->ip_tcp_write_append_reply);

    bp_sdk_params->app_callbacks->ip_tcp_write_append_reply(msg->seq_nr, MSG_STATUS_GET_CODE(msg->status),
                                                            msg->data[0] << 8 | msg->data[1],
                                                            msg->data[2] << 8 | msg->data[3]);
}

static void process_rx_tcp_write_send_reply(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->ip_tcp_write_send_reply);

    bp_sdk_params->app_callbacks->ip_tcp_write_send_reply(msg->seq_nr, MSG_STATUS_GET_CODE(msg->status));
}

static void process_rx_tcp_rx_notify_request(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->ip_tcp_rx_notify_request);

    bp_sdk_params->app_callbacks->ip_tcp_rx_notify_request(msg->seq_nr, msg->data[0], msg->data[1], &msg->data[2]);
}

static void process_rx_http_init_connection_reply(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->ip_http_init_connection_reply);

    bp_sdk_params->app_callbacks->ip_http_init_connection_reply(msg->seq_nr, MSG_STATUS_GET_CODE(msg->status),
                                                                msg->data[0]);
}

static void process_rx_http_post_write_ready_notify_post(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->ip_http_post_write_ready_notify_post);

    bp_sdk_params->app_callbacks->ip_http_post_write_ready_notify_post(msg->seq_nr, msg->data[0]);
}

static void process_rx_http_post_write_reply(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->ip_http_post_write_reply);

    bp_sdk_params->app_callbacks->ip_http_post_write_reply(msg->seq_nr, MSG_STATUS_GET_CODE(msg->status));
}

static void process_rx_http_rx_notify_request(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->ip_http_rx_notify_request);

    bp_sdk_params->app_callbacks->ip_http_rx_notify_request(msg->seq_nr, msg->data[0], msg->data[1], &msg->data[2]);
}

static void process_rx_inet_close_reply(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->ip_connection_close_reply);

    bp_sdk_params->app_callbacks->ip_connection_close_reply(msg->seq_nr, MSG_STATUS_GET_CODE(msg->status));
}

static void process_rx_inet_close_notify_post(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->ip_connection_close_notify_post);

    bp_sdk_params->app_callbacks->ip_connection_close_notify_post(msg->seq_nr, msg->data[0], msg->data[1]);
}

static void process_rx_mqtt_init_connection_reply(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->mqtt_init_connection_reply);

    bp_sdk_params->app_callbacks->mqtt_init_connection_reply(msg->seq_nr, MSG_STATUS_GET_CODE(msg->status),
                                                             msg->data[0]);
}

static void process_rx_mqtt_connection_established_notify_post(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->mqtt_connection_established_notify_post);

    bp_sdk_params->app_callbacks->mqtt_connection_established_notify_post(msg->seq_nr, msg->data[0]);
}

static void process_rx_mqtt_subscribe_topics_reply(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->mqtt_subscribe_topics_reply);

    bp_sdk_params->app_callbacks->mqtt_subscribe_topics_reply(msg->seq_nr, MSG_STATUS_GET_CODE(msg->status),
                                                              msg->data[0]);
}

static void process_rx_mqtt_unsubscribe_topics_reply(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->mqtt_unsubscribe_topics_reply);

    bp_sdk_params->app_callbacks->mqtt_unsubscribe_topics_reply(msg->seq_nr, MSG_STATUS_GET_CODE(msg->status));
}

static void process_rx_mqtt_publish_reply(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->mqtt_publish_reply);

    bp_sdk_params->app_callbacks->mqtt_publish_reply(msg->seq_nr, MSG_STATUS_GET_CODE(msg->status));
}

static void process_rx_mqtt_publication_rx_notify_request(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->mqtt_publication_rx_notify_request);

    uint8_t topic_len   = msg->data[2];
    uint8_t data_len    = msg->data[3+topic_len];

    bp_sdk_params->app_callbacks->mqtt_publication_rx_notify_request(msg->seq_nr, msg->data[0],
                                                                     msg->data[1] & 0x03,
                                                                     (msg->data[1] & 0x04) >> 2,
                                                                     (msg->data[1] & 0x80) >> 7,
                                                                     topic_len, &msg->data[3],
                                                                     data_len, &msg->data[4+topic_len]);
}

static void process_rx_mqtt_disconnect_reply(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->mqtt_disconnect_reply);

    bp_sdk_params->app_callbacks->mqtt_disconnect_reply(msg->seq_nr, MSG_STATUS_GET_CODE(msg->status));
}

static void process_rx_mqtt_disconnect_notify_post(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->mqtt_disconnect_notify_post);

    bp_sdk_params->app_callbacks->mqtt_disconnect_notify_post(msg->seq_nr, msg->data[0], msg->data[1]);
}

static void process_rx_sms_tx_setup_reply(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->sms_setup_transmission_reply);

    bp_sdk_params->app_callbacks->sms_setup_transmission_reply(msg->seq_nr, MSG_STATUS_GET_CODE(msg->status));
}

static void process_rx_sms_tx_append_send_reply(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->sms_message_append_send_reply);

    bp_sdk_params->app_callbacks->sms_message_append_send_reply(msg->seq_nr, MSG_STATUS_GET_CODE(msg->status));
}

static void process_rx_sms_rx_notify_request(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->sms_message_rx_notify_request);

    bp_sdk_params->app_callbacks->sms_message_rx_notify_request(msg->seq_nr, msg->data[0],
                                                                msg->data[1], msg->data[2], msg->data[3],
                                                                msg->data[4], msg->data[5], msg->data[6],
                                                                msg->data[7] & 0x7F, (msg->data[7] & 0x80) >> 7,
                                                                msg->data[8], &msg->data[9]);
}

static void process_rx_sms_rx_part_notify_request(const mp_message* msg, uint8_t exp_msg_type, uint16_t exp_data_len)
{
    RX_HANDLER_BASIC_CHECKS(bp_sdk_params->app_callbacks->sms_message_part_notify_request);

    bp_sdk_params->app_callbacks->sms_message_part_notify_request(msg->seq_nr, msg->data[0],
                                                                  msg->data[1], &msg->data[2]);
}

