/** ********************************************************************************************************************
 * @file        bp_api.h
 * @author      Briowireless, Inc.
 * @date        2016-10-28
 * @brief       Main API definition for the BitPipe Serial SDK.
 *              Defines the public function to initialize the SDK, and process bytes received from the BitPipe via UART.
 *              See each bp_api_* header file for functions to send requests to the BitPipe, and prototypes for callbacks.
 * @version     API V1
 * @copyright   Briowireless, Inc., 2016, Licensed under BSD license. See LICENSE file for more details.
 *
 ** *******************************************************************************************************************/
#ifndef __BP_API_H__
#define __BP_API_H__


// Include directives  -------------------------------------------------------------------------------------------------

#include <bp_api_sys.h>         // System API
#include <bp_api_devi.h>        // Device Information API
#include <bp_api_mobi.h>        // Mobile Information API
#include <bp_api_ip.h>          // IP Connectivity API
#include <bp_api_mqtt.h>        // MQTT API
#include <bp_api_sms.h>         // SMS API


// Preprocessor definitions --------------------------------------------------------------------------------------------

// Global Type definitions ---------------------------------------------------------------------------------------------

/// @brief Application callbacks structure. You need to define one instance of this struct and use it in the
///        bp_api_init_params() function below.
typedef struct __bp_api_callbacks
{
    // SYSTEM
    bp_api_sys_error_report_post                sys_error_report_post;
    bp_api_sys_set_max_payload_sz_reply         sys_set_max_payload_sz_reply;
    bp_api_sys_set_uart_speed_reply             sys_set_uart_speed_reply;

    // DEVICE INFORMATION
    bp_api_devi_get_hw_version_reply            devi_get_hw_version_reply;
    bp_api_devi_get_fw_version_reply            devi_get_fw_version_reply;
    bp_api_devi_get_product_name_reply          devi_get_product_name_reply;
    bp_api_devi_get_vin_vio_voltage_reply       devi_get_vin_vio_voltage_reply;
    bp_api_devi_get_junction_temp_reply         devi_get_junction_temp_reply;
    bp_api_devi_set_low_power_reply             devi_set_low_power_reply;
    bp_api_devi_get_latest_fw_version_reply     devi_get_latest_fw_version_reply;
    bp_api_devi_update_fw_reply                 devi_update_fw_reply;
    bp_api_devi_get_autonomous_mode_config_reply devi_get_autonomous_mode_config_reply;
    bp_api_devi_set_autonomous_mode_config_reply devi_set_autonomous_mode_config_reply;

    // MOBILE INFORMATION
    bp_api_mobi_power_on_radio_reply            mobi_power_on_radio_reply;
    bp_api_mobi_power_off_radio_reply           mobi_power_off_radio_reply;
    bp_api_mobi_radio_start_notify_post         mobi_radio_start_notify_post;
    bp_api_mobi_get_radio_status_reply          mobi_get_radio_status_reply;
    bp_api_mobi_get_imei_reply                  mobi_get_imei_reply;
    bp_api_mobi_get_iccid_reply                 mobi_get_iccid_reply;
    bp_api_mobi_get_imsi_reply                  mobi_get_imsi_reply;
    bp_api_mobi_get_own_phone_number_reply      mobi_get_own_phone_number_reply;
    bp_api_mobi_get_apn_config_reply            mobi_get_apn_config_reply;
    bp_api_mobi_set_apn_config_reply            mobi_set_apn_config_reply;
    bp_api_mobi_set_sim_access_config_reply     mobi_set_sim_access_config_reply;
    bp_api_mobi_get_mobile_conn_status_reply    mobi_get_mobile_conn_status_reply;
    bp_api_mobi_mobile_conn_status_change_notify_post mobi_mobile_conn_status_change_notify_post;
    bp_api_mobi_get_mobile_date_time_reply      mobi_get_mobile_date_time_reply;

    // IP CONNECTIVITY
    bp_api_ip_tcp_init_connection_reply         ip_tcp_init_connection_reply;
    bp_api_ip_tcp_connection_established_notify_post ip_tcp_connection_established_notify_post;
    bp_api_ip_tcp_write_append_reply            ip_tcp_write_append_reply;
    bp_api_ip_tcp_write_send_reply              ip_tcp_write_send_reply;
    bp_api_ip_tcp_rx_notify_request             ip_tcp_rx_notify_request;
    bp_api_ip_http_init_connection_reply        ip_http_init_connection_reply;
    bp_api_ip_http_post_write_ready_notify_post	ip_http_post_write_ready_notify_post;
    bp_api_ip_http_post_write_reply             ip_http_post_write_reply;
    bp_api_ip_http_rx_notify_request            ip_http_rx_notify_request;
    bp_api_ip_connection_close_reply            ip_connection_close_reply;
    bp_api_ip_connection_close_notify_post      ip_connection_close_notify_post;

    // MQTT
    bp_api_mqtt_init_connection_reply           mqtt_init_connection_reply;
    bp_api_mqtt_connection_established_notify_post mqtt_connection_established_notify_post;
    bp_api_mqtt_subscribe_topics_reply          mqtt_subscribe_topics_reply;
    bp_api_mqtt_unsubscribe_topics_reply        mqtt_unsubscribe_topics_reply;
    bp_api_mqtt_publish_reply                   mqtt_publish_reply;
    bp_api_mqtt_publication_rx_notify_request   mqtt_publication_rx_notify_request;
    bp_api_mqtt_disconnect_reply                mqtt_disconnect_reply;
    bp_api_mqtt_disconnect_notify_post          mqtt_disconnect_notify_post;

    // SMS
    bp_api_sms_setup_transmission_reply         sms_setup_transmission_reply;
    bp_api_sms_message_append_send_reply        sms_message_append_send_reply;
    bp_api_sms_message_rx_notify_request        sms_message_rx_notify_request;
    bp_api_sms_message_part_notify_request      sms_message_part_notify_request;

} bp_api_callbacks;


/// @brief Application UART write function. A function of this type needs to be defined in your application, so the SDK can
///        write (i.e.: send data) to the BitPipe. The function should block until all data is written to BitPipe.
typedef void (*uart_write_fct)(const uint8_t* data, uint16_t data_len);


/// @brief SDK Initialization parameters. Stuff used by the SDK to handle UART communication with the BitPipe.
typedef struct __bp_api_init_params
{
    bp_api_callbacks*   app_callbacks;          ///< Application callbacks when receiving messages from the BitPipe.
    uart_write_fct      uart_write;             ///< Platform-specific function to handle writing on UART port.

    uint16_t            serialize_buf_len;      ///< Size, in bytes, of the 'serialize_buf' buffer.
    uint8_t*            serialize_buf;          ///< Buffer used for flattening out outgoing messages in a linear byte array (before writing to UART port).

    uint16_t            deserialize_buf_len;    ///< Size, in bytes, of the 'deserialize_buf' buffer.
    uint8_t*            deserialize_buf;        ///< Buffer used for reconstructing messages received from the BitPipe.

} bp_api_init_params;


// Global Variables ----------------------------------------------------------------------------------------------------

extern mp_seq_nr bp_req_seq_nr;


// Public Functions ----------------------------------------------------------------------------------------------------

uint8_t bp_api_init(bp_api_init_params* init_params);

uint8_t bp_api_process_incoming(const uint8_t* data, uint16_t data_len, uint16_t* consumed_len);

uint8_t bp_api_send_message(uint8_t msg_id, mp_seq_nr seq_nr, uint8_t status, const uint8_t* data, uint16_t data_len);


#endif // __BP_API_H__
