/** ********************************************************************************************************************
 * @file        bp_api_ip.h
 * @author      Briowireless, Inc.
 * @date        2016-11-03
 * @brief       API layer for IP CONNECTIVITY requests and replies to the BitPipe, and defining callback prototypes for
 *               requests, replies and asynchronous posts sent from the BitPipe.
 *              See the BitPipe Serial API document for more information on requests below.
 * @version     API V1
 * @copyright   Briowireless, Inc., 2016, Licensed under BSD license. See LICENSE file for more details.
 *
 ** *******************************************************************************************************************/
#ifndef __BP_API_IP_H__
#define __BP_API_IP_H__


// Include directives  -------------------------------------------------------------------------------------------------

#include <platform_defs.h>
#include <message_protocol/message_protocol.h>


// Preprocessor definitions --------------------------------------------------------------------------------------------

#define BP_API_IP_INVALID_CONNECTION_ID                                         0xFF    ///< Invalid IP connection Identifier.

#define BP_API_IP_TCP_INIT_CONNECTION__SECURE_NO                                0x00    ///< Doesn't use encryption
#define BP_API_IP_TCP_INIT_CONNECTION__SECURE_YES                               0x01    ///< Uses SSL/TLS encryption
#define BP_API_IP_TCP_INIT_CONNECTION__CONNCLASS_A                              0x00    ///< Connection class A. (The only connection class supported).
#define BP_API_IP_TCP_INIT_CONNECTION__STATUS_NO_TCP_CONN_AVAILABLE             0x0A    ///< Status: No more connections are available.
#define BP_API_IP_TCP_INIT_CONNECTION__STATUS_APN_NOT_CONFIGURED                0x0B    ///< Status: APN is not configured.
#define BP_API_IP_TCP_INIT_CONNECTION__STATUS_SIM_STATUS_INVALID                0x0C    ///< Status: SIM status is invalid.
#define BP_API_IP_TCP_INIT_CONNECTION__STATUS_SECURITY_NOT_AVAILABLE            0x0D    ///< Status: Transport layer security not available.
#define BP_API_IP_TCP_INIT_CONNECTION__STATUS_TCP_KA_NOT_SUPPORTED              0x0E    ///< Status: TCP keep-alive not supported.
#define BP_API_IP_TCP_INIT_CONNECTION__STATUS_NO_LINK_OR_BEARER                 0x0F    ///< Status: No link or bearer available
#define BP_API_IP_TCP_INIT_CONNECTION__STATUS_ERR_GENERIC                       0x10    ///< Status: Unknown error occurred.

#define BP_API_IP_HTTP_INIT_CONNECTION__SECURE_NO                               0x00    ///< Doesn't use encryption
#define BP_API_IP_HTTP_INIT_CONNECTION__SECURE_YES                              0x01    ///< Uses SSL/TLS encryption
#define BP_API_IP_HTTP_INIT_CONNECTION__METHOD_GET                              0x00    ///< HTTP method is 'GET'
#define BP_API_IP_HTTP_INIT_CONNECTION__METHOD_POST                             0x01    ///< HTTP method is 'POST'
#define BP_API_IP_HTTP_INIT_CONNECTION__METHOD_HEAD                             0x02    ///< HTTP method is 'HEAD'

#define BP_API_IP_HTTP_POST_WRITE__WRITEOP_WRITE_TO_BUFFER                      0x00    ///< Just write/append to internal BitPipe Buffer.
#define BP_API_IP_HTTP_POST_WRITE__WRITEOP_WRITE_TO_BUFFER_AND_SEND             0x01    ///< Write to buffer, and transmit on the network.
#define BP_API_IP_HTTP_POST_WRITE__WRITEOP_WRITE_TO_BUFFER_AND_SEND_COMPLETE    0x02    ///< Write to buffer, and transmit on the network, and we're done.

#define BP_API_IP_CONNECTION_CLOSE_NOTIFY__REASON_REMOTE_CLOSED                 0x00    ///< Remote host has closed the connection.
#define BP_API_IP_CONNECTION_CLOSE_NOTIFY__REASON_MOBILE_NETWORK_ERROR          0x01    ///< PDP error, APN error, mobile-network related error.
#define BP_API_IP_CONNECTION_CLOSE_NOTIFY__REASON_HOST_NOT_FOUND                0x02    ///< Host name not found/resolved.
#define BP_API_IP_CONNECTION_CLOSE_NOTIFY__REASON_HOST_UNREACHABLE              0x03    ///< Connection timeout to specified host/port
#define BP_API_IP_CONNECTION_CLOSE_NOTIFY__REASON_CONNECTION_RESET              0x04    ///< Host actively refused connection.
#define BP_API_IP_CONNECTION_CLOSE_NOTIFY__REASON_RESOURCE_NOT_FOUND            0x05    ///< (HTTP 404)
#define BP_API_IP_CONNECTION_CLOSE_NOTIFY__REASON_GENERIC_UNKNOWN               0xFE    ///< Any other error.


// Global Type definitions ---------------------------------------------------------------------------------------------

// TCP
typedef void (*bp_api_ip_tcp_init_connection_reply)(uint8_t seq_nr, uint8_t status, uint8_t tcp_conn_id);
typedef void (*bp_api_ip_tcp_connection_established_notify_post)(uint8_t seq_nr, uint8_t tcp_conn_id);
typedef void (*bp_api_ip_tcp_write_append_reply)(uint8_t seq_nr, uint8_t status, uint16_t buf_sz_current, uint16_t buf_sz_max);
typedef void (*bp_api_ip_tcp_write_send_reply)(uint8_t seq_nr, uint8_t status);
typedef void (*bp_api_ip_tcp_rx_notify_request)(uint8_t seq_nr, uint8_t tcp_conn_id, uint8_t data_len, uint8_t* data);

// HTTP
typedef void (*bp_api_ip_http_init_connection_reply)(uint8_t seq_nr, uint8_t status, uint8_t http_conn_id);
typedef void (*bp_api_ip_http_post_write_ready_notify_post)(uint8_t seq_nr, uint8_t http_conn_id);
typedef void (*bp_api_ip_http_post_write_reply)(uint8_t seq_nr, uint8_t status);
typedef void (*bp_api_ip_http_rx_notify_request)(uint8_t seq_nr, uint8_t http_conn_id, uint8_t data_len, uint8_t* data);

// COMMON (TCP / HTTP)
typedef void (*bp_api_ip_connection_close_reply)(uint8_t seq_nr, uint8_t status);
typedef void (*bp_api_ip_connection_close_notify_post)(uint8_t seq_nr, uint8_t ip_conn_id, uint8_t reason);


// Public Functions ----------------------------------------------------------------------------------------------------

// TCP
uint8_t bp_api_ip_tcp_init_connection(mp_seq_nr* seq_nr, uint8_t secure, uint8_t conn_class, uint8_t host_len, uint8_t* host,
                                      uint16_t port, uint16_t ka_ival);
uint8_t bp_api_ip_tcp_write_append(mp_seq_nr* seq_nr, uint8_t tcp_conn_id, uint8_t data_len, uint8_t* data);
uint8_t bp_api_ip_tcp_write_send(mp_seq_nr* seq_nr, uint8_t tcp_conn_id);
uint8_t bp_api_ip_tcp_rx_notify_ack(mp_seq_nr ack_seq_nr);

// HTTP
uint8_t bp_api_ip_http_init_connection(mp_seq_nr* seq_nr, uint8_t secure, uint8_t method, uint8_t host_len, uint8_t* host,
                                       uint16_t port, uint8_t path_q_len, uint8_t* path_q, uint8_t add_h_len, uint8_t* add_h,
                                       uint32_t post_content_len);
uint8_t bp_api_ip_http_post_write(mp_seq_nr* seq_nr, uint8_t http_conn_id, uint8_t write_op, uint8_t data_len, uint8_t* data);
uint8_t bp_api_ip_http_rx_notify_ack(mp_seq_nr ack_seq_nr);

// COMMON
uint8_t bp_api_ip_connection_close(mp_seq_nr* seq_nr, uint8_t ip_conn_id);


#endif // __BP_API_IP_H__
