/** ********************************************************************************************************************
 * @file        bp_api_mqtt.c
 * @author      Briowireless, Inc.
 * @date        2016-11-08
 * @brief       API layer for MQTT requests and replies to the BitPipe, and defining callback prototypes for
 *               requests, replies and asynchronous posts sent from the BitPipe.
 *              See the BitPipe Serial API document for more information on requests below.
 * @version     API V1
 * @copyright   Briowireless, Inc., 2016, Licensed under BSD license. See LICENSE file for more details.
 *
 ** *******************************************************************************************************************/

// Include directives  -------------------------------------------------------------------------------------------------

#include <bp_api_mqtt.h>
#include <bp_api.h>
#include <message_defs.h>


// Preprocessor definitions --------------------------------------------------------------------------------------------

// Global (extern) function  -------------------------------------------------------------------------------------------

/**
 * @defgroup api_mqtt BitPipe API: MQTT
 * Allows initiating and handling MQTT client connections to MQTT brokers or servers.
 * Allows subscription (and unsubscription) to topics, receiving messages, publishing messages, and basic QOS0 and QOS1
 * support.
 * @{
 */

/** ********************************************************************************************************************
 * @brief  Initiates a MQTT connection to a remote MQTT server.
 *         The 'MQTT CONNECT' operation is performed internally by the BitPipe after TCP connection is established.
 *
 * @param  seq_nr           The message sequence number of the request sent to the BitPipe, if successful.
 * @param  secure           Specifies if the MQTT connection should be encrypted using SSL/TLS. Values: [0,1]
 * @param  clean_session    Specifies whether we're starting a clean session, or resuming an exiting one. Values: [0,1]
 * @param  host_len         Length of the remote MQTT server hostname, or IP address (excluding NULL string terminator, if any).
 * @param  host             Remote MQTT server hostname, or IP address.
 * @param  port             MQTT remote port to connect to. If 0, 1883 or 8883 will be automatically chosen depending on the secure parameter.
 * @param  client_id_len    Length of the MQTT CONNECT client identifier parameter.
 * @param  client_id        Client identifier of the MQTT CONNECT operation.
 * @param  user_len         Length of the MQTT CONNECT user name for authentication.
 * @param  user             Username for authentication during the MQTT CONNECT OPERATION.
 * @param  pass_len         Length of the MQTT CONNECT password for authentication.
 * @param  pass             Password for authentication during MQTT CONNECT operation.
 * @param  ka_ival_5_secs   Keep-Alive interval (in units of 5 seconds). Values: [0 (disabled), 60-240]
 *
 * @retval uint8_t  0 indicates that the message was successfully serialized and sent to the BitPipe,
 *                  See bp_api_priv_send_message() for more information on other return codes.
 */
uint8_t bp_api_mqtt_init_connection(mp_seq_nr* seq_nr, uint8_t secure, uint8_t clean_session,
                                    uint8_t host_len, uint8_t* host, uint16_t port,
                                    uint8_t client_id_len, uint8_t* client_id,
                                    uint8_t user_len, uint8_t* user,
                                    uint8_t pass_len, uint8_t* pass,
                                    uint8_t ka_ival_5_secs)
{
    uint8_t linear_buf[8 + host_len + client_id_len + user_len + pass_len];
    uint8_t msg_seq_nr = bp_req_seq_nr++;

    linear_buf[0] = 0;
    if(secure)
    {
        linear_buf[0] |= (1<<0);        // SSL/TLS
    }
    if(clean_session)
    {
        linear_buf[0] |= (1<<1);        // Clean Session
    }

    linear_buf[1] = host_len;
    memcpy(&linear_buf[2], host, host_len);

    linear_buf[2+host_len] = ((port >> 8) & 0xFF);
    linear_buf[3+host_len] = (port & 0xFF);

    linear_buf[4+host_len] = client_id_len;
    memcpy(&linear_buf[5+host_len], client_id, client_id_len);

    linear_buf[5+host_len+client_id_len] = user_len;
    memcpy(&linear_buf[6+host_len+client_id_len], user, user_len);

    linear_buf[6+host_len+client_id_len+user_len] = pass_len;
    memcpy(&linear_buf[7+host_len+client_id_len+user_len], pass, pass_len);

    linear_buf[8+host_len+client_id_len+user_len+pass_len] = ka_ival_5_secs;

    P_WARN("[%02X] Initiating %sMQTT connection to %s:%d ...\n", msg_seq_nr, secure ? "secure ":"", host, port);

    uint8_t status = bp_api_send_message(MSG_ID_MQTT_INITIATE_CONNECTION, msg_seq_nr, 0x00, linear_buf, sizeof(linear_buf));
    if(status != 0)
    {
        P_ERR("Failed to send message!\n");
        return status;
    }

    if(seq_nr)  *seq_nr = msg_seq_nr;

    return 0;
}

/** ********************************************************************************************************************
 * @brief  Subscribes to 1, 2 or 3 topics on the MQTT server.
 *
 * @param  seq_nr       The message sequence number of the request sent to the BitPipe, if successful.
 * @param  mqtt_conn_id A valid MQTT connection identifier acquired in the reply to the MQTT INITIATE CONNECTION request.
 * @param  qos1         The requested QOS for subscribing to topic #1.
 * @param  qos2         The requested QOS for subscribing to topic #2.
 * @param  qos3         The requested QOS for subscribing to topic #3.
 * @param  topic1_len   The length of the first topic to subscribe to (excluding NULL string terminator, if any).
 * @param  topic1       The first topic to subscribe to.
 * @param  topic2_len   The length of the second topic to subscribe to (excluding NULL string terminator, if any).
 * @param  topic2       The first second to subscribe to.
 * @param  topic3_len   The length of the third topic to subscribe to (excluding NULL string terminator, if any).
 * @param  topic3       The third topic to subscribe to.
 *
 * @retval uint8_t  0 indicates that the message was successfully serialized and sent to the BitPipe,
 *                  See bp_api_priv_send_message() for more information on other return codes.
 */
uint8_t bp_api_mqtt_subscribe_topics(mp_seq_nr* seq_nr, uint8_t mqtt_conn_id, uint8_t qos1, uint8_t qos2, uint8_t qos3,
                                     uint8_t topic1_len, uint8_t* topic1, uint8_t topic2_len, uint8_t* topic2,
                                     uint8_t topic3_len, uint8_t* topic3)
{
    uint8_t linear_buf[5+topic1_len+topic2_len+topic3_len];
    uint8_t msg_seq_nr = bp_req_seq_nr++;

    linear_buf[0] = mqtt_conn_id;

    linear_buf[1] = 0;
    linear_buf[1] |= qos1 << 0;
    linear_buf[1] |= qos2 << 2;
    linear_buf[1] |= qos3 << 4;

    linear_buf[2] = topic1_len;
    memcpy(&linear_buf[3], topic1, topic1_len);

    linear_buf[3+topic1_len] = topic2_len;
    memcpy(&linear_buf[4+topic1_len], topic2, topic2_len);

    linear_buf[4+topic1_len+topic2_len] = topic3_len;
    memcpy(&linear_buf[5+topic1_len+topic2_len], topic3, topic3_len);

    P_WARN("[%02X] Subscribing to MQTT topics (conn_id: %02X) ...\n", msg_seq_nr, mqtt_conn_id);

    uint8_t status = bp_api_send_message(MSG_ID_MQTT_SUBSCRIBE_TOPICS, msg_seq_nr, 0x00, linear_buf, sizeof(linear_buf));
    if(status != 0)
    {
        P_ERR("Failed to send message!\n");
        return status;
    }

    if(seq_nr)  *seq_nr = msg_seq_nr;

    return 0;
}

/** ********************************************************************************************************************
 * @brief  Unubscribes from 1, 2 or 3 topics on the MQTT server.
 *
 * @param  seq_nr       The message sequence number of the request sent to the BitPipe, if successful.
 * @param  mqtt_conn_id A valid MQTT connection identifier acquired in the reply to the MQTT INITIATE CONNECTION request.
 * @param  topic1_len   The length of the first topic to unsubscribe from (excluding NULL string terminator, if any).
 * @param  topic1       The first topic to subscribe from.
 * @param  topic2_len   The length of the second topic to unsubscribe from (excluding NULL string terminator, if any).
 * @param  topic2       The first second to subscribe from.
 * @param  topic3_len   The length of the third topic to unsubscribe from (excluding NULL string terminator, if any).
 * @param  topic3       The third topic to subscribe from.
 *
 * @retval uint8_t  0 indicates that the message was successfully serialized and sent to the BitPipe,
 *                  See bp_api_priv_send_message() for more information on other return codes.
 */
uint8_t bp_api_mqtt_unsubscribe_topics(mp_seq_nr* seq_nr, uint8_t mqtt_conn_id, uint8_t topic1_len, uint8_t* topic1,
                                       uint8_t topic2_len, uint8_t* topic2, uint8_t topic3_len, uint8_t* topic3)
{
    uint8_t linear_buf[4+topic1_len+topic2_len+topic3_len];
    uint8_t msg_seq_nr = bp_req_seq_nr++;

    linear_buf[0] = mqtt_conn_id;

    linear_buf[1] = topic1_len;
    memcpy(&linear_buf[2], topic1, topic1_len);

    linear_buf[2+topic1_len] = topic2_len;
    memcpy(&linear_buf[3+topic1_len], topic2, topic2_len);

    linear_buf[3+topic1_len+topic2_len] = topic3_len;
    memcpy(&linear_buf[4+topic1_len+topic2_len], topic3, topic3_len);

    P_WARN("[%02X] UnSubscribing from MQTT topics (conn_id: %02X) ...\n", msg_seq_nr, mqtt_conn_id);

    uint8_t status = bp_api_send_message(MSG_ID_MQTT_UNSUBSCRIBE_TOPICS, msg_seq_nr, 0x00, linear_buf, sizeof(linear_buf));
    if(status != 0)
    {
        P_ERR("Failed to send message!\n");
        return status;
    }

    if(seq_nr)  *seq_nr = msg_seq_nr;

    return 0;
}

/** ********************************************************************************************************************
 * @brief  Publishes a message to the MQTT broker, with the specified flags, topic and data.
 * @note   When setupping to transfer a large paylaoad, you may need to call this function several times with
 *         the parameter 'send_indicator' at 0. When doing this, the parameter 'topic' is only necessary on the first
 *         call to this method. This allows you to maximize message data transfer by omitting the topic after the first
 *         invcation.
 *
 * @param  seq_nr           The message sequence number of the request sent to the BitPipe, if successful.
 * @param  mqtt_conn_id     A valid MQTT connection identifier acquired in the reply to the MQTT INITIATE CONNECTION request.
 * @param  qos              The QOS under which this message is published. Values: [0,1]
 * @param  retained         Specifies whether the message is marked as retained. Values: [0,1]
 * @param  send_indicator   Specifies whether all the message data has been transferred, and is ready to be sent.
 * @param  topic_len        Length of the topic string, under which this message is published (excluding NULL string terminator, if any).
 * @param  topic            Topic, under which this message is published.
 * @param  data_len         Length of the message data.
 * @param  data             Message data.
 *
 * @retval uint8_t  0 indicates that the message was successfully serialized and sent to the BitPipe,
 *                  See bp_api_priv_send_message() for more information on other return codes.
 */
uint8_t bp_api_mqtt_publish(mp_seq_nr* seq_nr, uint8_t mqtt_conn_id, uint8_t qos, uint8_t retained, uint8_t send_indicator,
                            uint8_t topic_len, uint8_t* topic,
                            uint8_t data_len, uint8_t* data)
{
    uint8_t linear_buf[4+topic_len+data_len];
    uint8_t msg_seq_nr = bp_req_seq_nr++;

    linear_buf[0] = mqtt_conn_id;

    linear_buf[1] = 0;
    linear_buf[1] |= ((uint8_t)qos);   // QOS on bits #0,1
    if(retained)
    {
        linear_buf[1] |= (1<<2);        // Retain indicator at bit #2
    }
    if(send_indicator)
    {
        linear_buf[1] |= (1<<7);        // Send indicator at bit #7
    }

    linear_buf[2] = topic_len;
    memcpy(&linear_buf[3], topic, topic_len);

    linear_buf[3+topic_len] = data_len;
    memcpy(&linear_buf[4+topic_len], data, data_len);

    P_WARN("[%02X] Publishing MQTT message (conn_id: %02X) ...\n", msg_seq_nr, mqtt_conn_id);

    uint8_t status = bp_api_send_message(MSG_ID_MQTT_PUBLISH, msg_seq_nr, 0x00, linear_buf, sizeof(linear_buf));
    if(status != 0)
    {
        P_ERR("Failed to send message!\n");
        return status;
    }

    if(seq_nr)  *seq_nr = msg_seq_nr;

    return 0;
}

/** ********************************************************************************************************************
 * @brief  Sends a reply (as acknowledgement) to the BitPipe, indicating that the MQTT PUBLICATION RX NOTIFY request
 *         was well received by the Host.
 *
 * @note   Failure to acknowledge MQTT PUBLICATION RX NOTIFY requests from the BitPipe will result in the BitPipe
 *         repeating the request, with the same data and sequence number, up to three times.
 *
 * @param  ack_seq_nr   The sequence number of the MQTT PUBLICATION RX NOTIFY request we want to acknowledge.
 *
 * @retval uint8_t  0 indicates that the message was successfully serialized and sent to the BitPipe,
 *                  See bp_api_priv_send_message() for more information on other return codes.
 */
uint8_t bp_api_mqtt_publication_rx_notify_ack(mp_seq_nr ack_seq_nr)
{
    P_WARN("[%02X]  -> Replying to MQTT Publication RX NOTIFY ...\n", ack_seq_nr);

    uint8_t status = bp_api_send_message(MSG_ID_MQTT_PUBLICATION_RX_NOTIFY, ack_seq_nr, 0x80, NULL, 0);
    if(status != 0)
    {
        P_ERR("Failed to send message!\n");
        return status;
    }

    return 0;
}

/** ********************************************************************************************************************
 * @brief  Gracefully disconnects the MQTT connection identified by 'mqtt_conn_id'.
 *
 * @note   Never specify a TCP or HTTP connection identifier in this function. Use bp_api_ip_connection_close() instead.
 *
 * @param  seq_nr       The message sequence number of the request sent to the BitPipe, if successful.
 * @param  mqtt_conn_id The MQTT connection identifier.
 *
 * @retval uint8_t  0 indicates that the message was successfully serialized and sent to the BitPipe,
 *                  See bp_api_priv_send_message() for more information on other return codes.
 */
uint8_t bp_api_mqtt_disconnect(mp_seq_nr* seq_nr, uint8_t mqtt_conn_id)
{
    uint8_t msg_seq_nr = bp_req_seq_nr++;

    P_WARN("[%02X] Disconnecting MQTT connection (conn_id: %02X) ...\n", msg_seq_nr, mqtt_conn_id);

    uint8_t status = bp_api_send_message(MSG_ID_MQTT_DISCONNECT, msg_seq_nr, 0x00, &mqtt_conn_id, sizeof(mqtt_conn_id));
    if(status != 0)
    {
        P_ERR("Failed to send message!\n");
        return status;
    }

    if(seq_nr)  *seq_nr = msg_seq_nr;

    return 0;
}


// ---------------------------------------------------------------------------------------------------------------------
// Doxygen documentation support.
// ---------------------------------------------------------------------------------------------------------------------


/**
 * @typedef void (*bp_api_mqtt_init_connection_reply)(uint8_t seq_nr, uint8_t status, uint8_t mqtt_conn_id)
 * @brief  Callback into application when the reply to MQTT INITIATE CONNECTION request is received from the BitPipe.
 *
 * @note   You will need to keep tack of the 'tcp_conn_id' for subsequent operations.
 *
 * @param  seq_nr       The sequence number of the message, matches the sequence number of the request sent previously.
 * @param  status       The status code of the operation. See the *BitPipe Serial API* documentation for possible values.
 * @param  mqtt_conn_id The connection identifier for this new connection.
 */

/**
 * @typedef void (*bp_api_mqtt_connection_established_notify_post)(uint8_t seq_nr, uint8_t mqtt_conn_id)
 * @brief  Callback into application when the MQTT CONNECTION ESTABLISHED NOTIFY message is posted from the BitPipe.
 *
 * @note   After connection establishment, you may now subscribe, unsubscribe, publish and receive publications.
 *
 * @param  seq_nr       The sequence number of the message.
 * @param  mqtt_conn_id The connection whose status changed to 'established'.
 */

/**
 * @typedef void (*bp_api_mqtt_subscribe_topics_reply)(uint8_t seq_nr, uint8_t status, uint8_t sub_status)
 * @brief  Callback into application when the reply to MQTT SUBSCRIBE TOPICS request is received from the BitPipe.
 *
 * @param  seq_nr       The sequence number of the message, matches the sequence number of the request sent previously.
 * @param  status       The status code of the operation. See the *BitPipe Serial API* documentation for possible values.
 * @param  sub_status   The subscription status for each topic requested (bit #0, #1, #2).
 */

/**
 * @typedef void (*bp_api_mqtt_unsubscribe_topics_reply)(uint8_t seq_nr, uint8_t status)
 * @brief  Callback into application when the reply to MQTT UNSUBSCRIBE TOPICS request is received from the BitPipe.
 *
 * @param  seq_nr       The sequence number of the message, matches the sequence number of the request sent previously.
 * @param  status       The status code of the operation. See the *BitPipe Serial API* documentation for possible values.
 */

/**
 * @typedef void (*bp_api_mqtt_publish_reply)(uint8_t seq_nr, uint8_t status)
 * @brief  Callback into application when the reply to MQTT PUBLISH request is received from the BitPipe.
 *
 * @param  seq_nr       The sequence number of the message, matches the sequence number of the request sent previously.
 * @param  status       The status code of the operation. See the *BitPipe Serial API* documentation for possible values.
 */

/**
 * @typedef void (*bp_api_mqtt_publication_rx_notify_request)(uint8_t seq_nr, uint8_t mqtt_conn_id, uint8_t qos, uint8_t retained, uint8_t complete, uint8_t topic_len, uint8_t* topic, uint8_t data_len, uint8_t* data)
 * @brief  Callback into application when the MQTT PUBLICATION RX NOTIFY message is requested from the BitPipe.
 *
 * This message is used whenever an MQTT message is received from the remote MQTT server by the BitPipe.
 * You **MUST** reply to this message.
 *
 * @note   For large message payloads, multiple MQTT PUBLICATION RX NOTIFY may be requested by the BitPipe. Check the 'complete'
 *         indicator to know when the message payload is complete. Also, the 'topic', 'qos', 'retained' parameters are only valid
 *         in the first message.
 *
 * @param  seq_nr       The sequence number of the message, which must be used in the reply to this request.
 * @param  mqtt_conn_id The MQTT connection identifier, from where the message came from.
 * @param  qos          The quality of service for this message.
 * @param  retained     Indicates whether this message was published because it was retained on the server.
 * @param  complete     Indicates whether this is the last MQTT PUBLICATION RX NOTIFY for the incoming publication.
 * @param  topic_len    The length of the topic, under which the message was published. This is only valid in the first MQTT PUBLICATION RX NOTIFY for the incoming publication.
 * @param  topic        The topic under which the message was published.
 * @param  data_len     Length of the message payload data this message contains.
 * @param  data         The message payload (or a part of it).
 */

/**
 * @typedef void (*bp_api_mqtt_disconnect_reply)(uint8_t seq_nr, uint8_t status)
 * @brief  Callback into application when the reply to MQTT DISCONNECT request is received from the BitPipe.
 *
 * @param  seq_nr       The sequence number of the message, matches the sequence number of the request sent previously.
 * @param  status       The status code of the operation. See the *BitPipe Serial API* documentation for possible values.
 */

/**
 * @typedef void (*bp_api_mqtt_disconnect_notify_post)(uint8_t seq_nr, uint8_t mqtt_conn_id, uint8_t reason)
 * @brief  Callback into application when the MQTT DISCONNECT NOTIFY message is posted from the BitPipe.
 *
 * You may not use the connection identifier 'mqtt_conn_id' in any subsequent operation.
 *
 * @param  seq_nr       The sequence number of the message.
 * @param  mqtt_conn_id The connection identifier that is now closed.
 * @param  reason       The reason why the connection was closed. See the *BitPipe Serial API* documentation for possible values.
 */


/**
 * @}       // Doxygen api_mqtt defgroup
 */
