/** ********************************************************************************************************************
 * @file        bp_api_sms.c
 * @author      Briowireless, Inc.
 * @date        2016-11-03
 * @brief       API layer for SMS requests and replies to the BitPipe, and defining callback prototypes for
 *               requests, replies and asynchronous posts sent from the BitPipe.
 *              See the BitPipe Serial API document for more information on requests below.
 * @version     API V1
 * @copyright   Briowireless, Inc., 2016, Licensed under BSD license. See LICENSE file for more details.
 *
 ** *******************************************************************************************************************/

// Include directives  -------------------------------------------------------------------------------------------------

#include <bp_api_sms.h>
#include <bp_api.h>
#include <message_defs.h>


// Preprocessor definitions --------------------------------------------------------------------------------------------

// Global (extern) function  -------------------------------------------------------------------------------------------

/**
 * @defgroup api_sms BitPipe API: SMS
 * Allows sending and receiving SMS messages to and from the BitPipe.
 * @{
 */

/** ********************************************************************************************************************
 * @brief  Informs the BitPipe of an upcoming SMS message transmission.
 *         This message configures the send options and the SMS message recipient (Phone Number).
 *
 * @note   If options and a recipient have been previously configured and don't need to be changed, this operation can
 *          be skipped. However, this information is NOT stored in volatile memory; it doesn't carry across reboots.
 *
 * @param  seq_nr           The message sequence number of the request sent to the BitPipe, if successful.
 * @param  concat_support   Indicates whether concatenation support is enabled for the upcoming SMS message.
 * @param  encoding         SMS message encding: 0=7-bits GSM03.38, 1=8-bits binary, 2=16-bits UCS2
 * @param  msisdn_len       Length of the mobile recipient phone number (excluding NULL string terminator, if any).
 * @param  msisdn           The mobile recipient phone number.
 *
 * @retval 0 indicates that the message was successfully serialized and sent to the BitPipe,
 *         See 'bp_api_priv_send_message' for more information on other return codes.
 */
uint8_t bp_api_sms_setup_transmission(mp_seq_nr* seq_nr, uint8_t concat_support, uint8_t encoding, uint8_t msisdn_len, uint8_t* msisdn)
{
    uint8_t linear_buf[2+msisdn_len];
    uint8_t msg_seq_nr = bp_req_seq_nr++;

    P_WARN("[%02X] SMS Setup Transmission ...\n", msg_seq_nr);

    linear_buf[0] = concat_support;
    linear_buf[0] |= (encoding << 1);

    linear_buf[1] = msisdn_len;
    memcpy(&linear_buf[2], msisdn, msisdn_len);

    uint8_t status = bp_api_send_message(MSG_ID_SMS_TX_SETUP, msg_seq_nr, 0x00, linear_buf, sizeof(linear_buf));
    if(status != 0)
    {
        P_ERR("Failed to send message!\n");
        return status;
    }

    if(seq_nr)  *seq_nr = msg_seq_nr;

    return 0;
}

/** ********************************************************************************************************************
 * @brief  This function allows to transfer a part of a message to the BitPipe and send it when the message is complete.
 *         If the message is short, it can be sent at once.
 *
 * @param  seq_nr           The message sequence number of the request sent to the BitPipe, if successful.
 * @param  send_indicator   Indicates whether the message is complete and should be sent. Values: [0,1]
 * @param  data_len         Length of the mobile recipient phone number (excluding NULL string terminator, if any).
 * @param  data             The mobile recipient phone number.
 *
 * @retval 0 indicates that the message was successfully serialized and sent to the BitPipe,
 *         See 'bp_api_priv_send_message' for more information on other return codes.
 */
uint8_t bp_api_sms_message_append_send(mp_seq_nr* seq_nr, uint8_t send_indicator, uint8_t data_len, uint8_t* data)
{
    uint8_t linear_buf[2+data_len];
    uint8_t msg_seq_nr = bp_req_seq_nr++;

    P_WARN("[%02X] SMS Append/Send send_ind=%d ...\n", msg_seq_nr, send_indicator);

    linear_buf[0] = send_indicator;

    linear_buf[1] = data_len;
    memcpy(&linear_buf[2], data, data_len);

    uint8_t status = bp_api_send_message(MSG_ID_SMS_TX_APPEND_SEND, msg_seq_nr, 0x00, linear_buf, sizeof(linear_buf));
    if(status != 0)
    {
        P_ERR("Failed to send message!\n");
        return status;
    }

    if(seq_nr)  *seq_nr = msg_seq_nr;

    return 0;
}

/** ********************************************************************************************************************
 * @brief  Sends a reply (as acknowledgement) to the BitPipe, indicating that the SMS MESSAGE RX NOTIFY request
 *         was well received by the Host.
 *
 * @note   Failure to acknowledge SMS MESSAGE RX NOTIFY requests from the BitPipe will result in the BitPipe
 *         repeating the request, with the same data and sequence number, up to three times.
 *
 * @param  ack_seq_nr   The sequence number of the SMS MESSAGE RX NOTIFY request we want to acknowledge.
 *
 * @retval uint8_t  0 indicates that the message was successfully serialized and sent to the BitPipe,
 *                  See bp_api_priv_send_message() for more information on other return codes.
 */
uint8_t bp_api_sms_message_rx_notify_ack(mp_seq_nr ack_seq_nr)
{
    P_WARN("[%02X]  -> Replying to SMS MESSAGE RX NOTIFY ...\n", ack_seq_nr);

    uint8_t status = bp_api_send_message(MSG_ID_SMS_RX_NOTIFY, ack_seq_nr, 0x80, NULL, 0);
    if(status != 0)
    {
        P_ERR("Failed to send message!\n");
        return status;
    }

    return 0;
}

/** ********************************************************************************************************************
 * @brief  Sends a reply (as acknowledgement) to the BitPipe, indicating that the SMS MESSAGE PART RX NOTIFY request
 *         was well received by the Host.
 *
 * @note   Failure to acknowledge SMS MESSAGE PART RX NOTIFY requests from the BitPipe will result in the BitPipe
 *         repeating the request, with the same data and sequence number, up to three times.
 *
 * @param  ack_seq_nr   The sequence number of the SMS MESSAGE PART RX NOTIFY request we want to acknowledge.
 *
 * @retval uint8_t  0 indicates that the message was successfully serialized and sent to the BitPipe,
 *                  See bp_api_priv_send_message() for more information on other return codes.
 */
uint8_t bp_api_sms_message_part_notify_ack(mp_seq_nr ack_seq_nr)
{
    P_WARN("[%02X]  -> Replying to SMS MESSAGE RX PART NOTIFY ...\n", ack_seq_nr);

    uint8_t status = bp_api_send_message(MSG_ID_SMS_RX_PART_NOTIFY, ack_seq_nr, 0x80, NULL, 0);
    if(status != 0)
    {
        P_ERR("Failed to send message!\n");
        return status;
    }

    return 0;
}


// ---------------------------------------------------------------------------------------------------------------------
// Doxygen documentation support.
// ---------------------------------------------------------------------------------------------------------------------

/**
 * @typedef void (*bp_api_sms_setup_transmission_reply)(uint8_t seq_nr, uint8_t status)
 * @brief  Callback into application when the reply to SMS SETUP TRANSMISSION request is received from the BitPipe.
 *
 * @param  seq_nr       The sequence number of the message, matches the sequence number of the request sent previously.
 * @param  status       The status code of the operation. See the *BitPipe Serial API* documentation for possible values.
 */

/**
 * @typedef void (*bp_api_sms_message_append_send_reply)(uint8_t seq_nr, uint8_t status)
 * @brief  Callback into application when the reply to SMS MESSAGE APPEND/SEND request is received from the BitPipe.
 *
 * @param  seq_nr       The sequence number of the message, matches the sequence number of the request sent previously.
 * @param  status       The status code of the operation. See the *BitPipe Serial API* documentation for possible values.
 */

/**
 * @typedef void (*bp_api_sms_message_rx_notify_request)(uint8_t seq_nr, uint8_t encoding, uint8_t year, uint8_t month, uint8_t day, uint8_t hour, uint8_t minute, uint8_t second, uint8_t tz_info_offset, uint8_t tz_info_sign, uint8_t msisdn_len, uint8_t* msisdn)
 * @brief  Callback into application when the SMS MESSAGE RX NOTIFY message is requested from the BitPipe.
 *
 * This message is used when a new SMS is received by the BitPipe.
 *
 * You should expect one or more bp_api_sms_message_part_notify_request() following the reply to this message.
 *
 * You **MUST** reply to this message.
 *
 * @param  seq_nr       The sequence number of the message, which must be used in the reply to this request.
 * @param  encoding     Encoding of the SMS message (0=7-bits GSM03.38; 1=8-bits binary; 2=16-bits UCS2)
 * @param  year         Sent local date year, minus 2000.
 * @param  month        Sent local date month, 0-based.
 * @param  day          Sent local date day, 0-based.
 * @param  hour         Sent local time, in 24-hour format.
 * @param  minute       Sent local time minute.
 * @param  second       Sent local time second.
 * @param  tz_info_offset   The offset, in units of 15 minutes, of the sender current timezone.
 * @param  tz_info_sign Specifies if the offset is negative(=1) or positive(=0).
 * @param  msisdn_len   Length of the sender mobile phone number.
 * @param  msisdn       The sender mobile phone number.
 */

/**
 * @typedef void (*bp_api_sms_message_part_notify_request)(uint8_t seq_nr, uint8_t last_part_indicator, uint8_t data_len, uint8_t* data)
 * @brief  Callback into application when the SMS MESSAGE PART NOTIFY message is requested from the BitPipe.
 *
 * This message is used to report SMS message data for an SMS received by the BitPipe.
 *
 * You **MUST** reply to this message.
 *
 * @param  seq_nr               The sequence number of the message, which must be used in the reply to this request.
 * @param  last_part_indicator  The connection identifier for this new connection.
 * @param  data_len             Length of SMS message data this message contains.
 * @param  data                 The SMS message data.
 */


/**
 * @}       // Doxygen api_sms defgroup
 */
