/** ********************************************************************************************************************
 * @file        message_defs.h
 * @author      Briowireless, Inc.
 * @date        2016-10-28
 * @brief       Message definitions for UART communication between BitPipe and Host.
 * @version     API V1
 * @copyright   Briowireless, Inc., 2016, Licensed under BSD license. See LICENSE file for more details.
 *
 **********************************************************************************************************************/
#ifndef __MESSAGE_DEFS_H__
#define __MESSAGE_DEFS_H__


/* Includes ----------------------------------------------------------------------------------------------------------*/

/* Global Defines ----------------------------------------------------------------------------------------------------*/

#define MSG_STATUS_REQUEST                          0x00        ///< Message type: Request.
#define MSG_STATUS_REPLY                            0x80        ///< Message type: Reply.
#define MSG_STATUS_GET_MSGTYPE(status)              (status & 0x80) ///< Allows retrieving the message type from a status code.
#define MSG_STATUS_GET_CODE(status)                 (status & 0x7F) ///< Allows retrieving the code from the status code.

// Common message status, applicable to all messages.
#define MSG_STATUS_CODE_SUCCESS                     0           ///< Operation successful.
#define MSG_STATUS_ERR_MESSAGE_MALFORMED            1           ///< Error, the message is badly formatted.
#define MSG_STATUS_ERR_INVALID_PARAMETER            2           ///< Error, a parameter value has failed validation.
#define MSG_STATUS_WARN_BITPIPE_BUSY                3           ///< Warning, the BitPipe is currently busy. Try again later.
#define MSG_STATUS_WARN_BITPIPE_INTERNAL_TIMEOUT    4           ///< Warning, an internal timeout occurred (often because of bad network coverage).
#define MSG_STATUS_WARN_BITPIPE_WLS_SLAVE_OFFLINE   5           ///< Warning, this operation requires the wireless slave to be powered on and ready.


/* Global type definitions -------------------------------------------------------------------------------------------*/

/// @brief Message command ID definition for BitPipe.
typedef enum __bp_msg_id
{
    // SYSTEM
    MSG_ID_ERROR_REPORT                     = 0x00,     // Request: BitPipe -> Host
    MSG_ID_SET_MAX_MSG_PAYSIZE              = 0x01,
    MSG_ID_SET_UART_SPEED                   = 0x02,

    // DEVICE INFORMATION
    MSG_ID_GET_HW_VERSION                   = 0x10,
    MSG_ID_GET_FW_VERSION                   = 0x11,
    MSG_ID_GET_PRODUCT_NAME                 = 0x12,
    MSG_ID_GET_VIN_VIO_VOLTAGE              = 0x13,
    MSG_ID_GET_JUNCTION_TEMP                = 0x14,
    MSG_ID_SET_LOW_POWER_MODE               = 0x15,
    // ...
    MSG_ID_GET_LATEST_FW_VERSION            = 0x18,
    MSG_ID_UPDATE_FIRMWARE                  = 0x19,
    MSG_ID_GET_AUTONOMOUS_CONFIG            = 0x1A,
    MSG_ID_SET_AUTONOMOUS_CONFIG            = 0x1B,

    // MOBILE INFORMATION and STATUS
    MSG_ID_POWER_ON_RADIO                   = 0x20,
    MSG_ID_POWER_OFF_RADIO                  = 0x21,
    MSG_ID_RADIO_START_NOTIFY               = 0x22,     // Request: BitPipe -> Host
    MSG_ID_GET_RADIO_STATUS                 = 0x23,
    MSG_ID_GET_IMEI                         = 0x24,
    MSG_ID_GET_ICCID                        = 0x25,
    MSG_ID_GET_IMSI                         = 0x26,
    MSG_ID_GET_OWN_PHONE_NUMBER             = 0x27,
    MSG_ID_GET_APN_CONFIG                   = 0x28,
    MSG_ID_SET_APN_CONFIG                   = 0x29,
    MSG_ID_SET_SIM_ACCESS_CONFIG            = 0x2A,
    MSG_ID_GET_MOBILE_CONN_STATUS           = 0x2B,
    MSG_ID_MOBILE_CONN_STATUS_NOTIFY        = 0x2C,     // Request: BitPipe -> Host
    MSG_ID_GET_MOBILE_DATE_AND_TIME         = 0x2D,

    // WiFi INFORMATION and STATUS
    // ..

    // LoRa INFORMATION and STATUS
    // ..

    // ..

    // IP SOCKET
    MSG_ID_TCP_INITIATE_CONNECTION          = 0x80,
    MSG_ID_TCP_CONNECTION_ESTABLISHED_NOTIFY= 0x81,     // Request: BitPipe -> Host
    MSG_ID_TCP_WRITE_APPEND                 = 0x82,
    MSG_ID_TCP_WRITE_SEND                   = 0x83,
    MSG_ID_TCP_RX_NOTIFY                    = 0x84,     // Request: BitPipe -> Host
    MSG_ID_HTTP_INITIATE_CONNECTION         = 0x85,
    MSG_ID_HTTP_POST_WRITE_READY_NOTIFY     = 0x86,     // Request: BitPipe -> Host
    MSG_ID_HTTP_POST_WRITE                  = 0x87,
    MSG_ID_HTTP_RX_NOTIFY                   = 0x88,     // Request: BitPipe -> Host
    // ..
    MSG_ID_INET_CLOSE                       = 0x8E,
    MSG_ID_INET_CLOSE_NOTIFY                = 0x8F,     // Request: BitPipe -> Host

    // MQTT
    MSG_ID_MQTT_INITIATE_CONNECTION         = 0x90,
    MSG_ID_MQTT_CONNECTION_ESTABLISHED_NOTIFY=0x91,     // Request: BitPipe -> Host
    MSG_ID_MQTT_SUBSCRIBE_TOPICS            = 0x92,
    MSG_ID_MQTT_UNSUBSCRIBE_TOPICS          = 0x93,
    MSG_ID_MQTT_PUBLISH                     = 0x94,
    MSG_ID_MQTT_PUBLICATION_RX_NOTIFY       = 0x95,     // Request: BitPipe -> Host
    MSG_ID_MQTT_DISCONNECT                  = 0x9E,
    MSG_ID_MQTT_DISCONNECT_NOTIFY           = 0x9F,     // Request: BitPipe -> Host

    // SMS
    MSG_ID_SMS_TX_SETUP                     = 0xD0,
    MSG_ID_SMS_TX_APPEND_SEND               = 0xD1,
    // ..
    MSG_ID_SMS_RX_NOTIFY                    = 0xD4,     // Request: BitPipe -> Host
    MSG_ID_SMS_RX_PART_NOTIFY               = 0xD5,     // Request: BitPipe -> Host
} bp_msg_id;


/* Global Variables --------------------------------------------------------------------------------------------------*/

/* Global Functions Prototypes ---------------------------------------------------------------------------------------*/

/* -------------------------------------------------------------------------------------------------------------------*/

#endif // __MESSAGE_DEFS_H__
