/** ********************************************************************************************************************
 * @file        message_protocol.c
 * @author      Brioconcept Consulting, Inc
 * @date        2015-08-05
 * @brief       Protocol to send and receive messages from point to point
 * @copyright   Brioconcept Consulting, Inc., 2016, Licensed under BSD license. See LICENSE file for more details.
 *
 ** *******************************************************************************************************************/

/* Includes ------------------------------------------------------------------*/
#include <message_protocol/crc16_ccitt.h>

/**
 * @defgroup api_message_protocol Message Protocol Low-level API
 * Definition of types and implementation of the message_protocol loow-level API.
 * @{
 */
#include <message_protocol/message_protocol.h>


/* Local Constants -----------------------------------------------------------*/

#ifndef NULL
#define NULL                                     ((void*)0)     ///< NULL
#endif

#define MESSAGE_PROTOCOL_WORKBUF_SIZE_MIN        32             ///< Minimum buffer size allowed during init.
#define MESSAGE_PROTOCOL_SERIALIZE_BUF_SIZE_MIN  7              ///< SOF, CMD_ID, STATUS, SEQ_NR, CRC_1, CRC_2, EOF

#define MESSAGE_PROTOCOL_SOF                     0x7E           ///< message protocol Start-Of-Frame
#define MESSAGE_PROTOCOL_EOF                     0xCE           ///< message protocol End-Of-Frame
#define MESSAGE_PROTOCOL_ESC                     0x9E           ///< message protocol escape character


/* Local Typedef -------------------------------------------------------------*/


/* Local Functions -----------------------------------------------------------*/

static inline int need_escaping(const uint8_t b)
{
    return (b == MESSAGE_PROTOCOL_SOF || b == MESSAGE_PROTOCOL_EOF || b == MESSAGE_PROTOCOL_ESC) ? 1 : 0;
}

static mp_status encode_byte(const uint8_t b, uint8_t * const buffer, uint16_t * const idx, const uint16_t buffer_len)
{
    if(need_escaping(b))
    {
        if(*idx >= buffer_len)
        {
            return MESSAGE_PROTOCOL_ERROR_BUF_TOO_SMALL;
        }

        buffer[(*idx)++] = MESSAGE_PROTOCOL_ESC;
    }

    if((*idx) >= buffer_len)
    {
        return MESSAGE_PROTOCOL_ERROR_BUF_TOO_SMALL;
    }
    buffer[(*idx)++] = b;

    return MESSAGE_PROTOCOL_OK;
}


// Public Functions -------------------------------------------------------------------------------------------------

/**
 * @brief This function initializes the message protocol internals.
 *        It takes ownership of the buffer provided for context, message decoding and deserialization.
 *
 *        The size of the buffer you need to pass here depends on the length of messages you expect to receive
 *        from the other endpoint. This is project-specific and should be carefully tuned depending on your needs.
 *
 * Buffer usage example, for a buffer of 44 bytes long:
 *
 *      [CONTEXT----][UART RX MESSAGE DATA---------]
 *     {............................................}
 *      |            |                             |
 *      0            12                            43
 *
 * Context takes ~12 bytes and is placed at the beginning of the buffer. The rest is used for message byte reception.
 *
 * @param buffer the buffer to use internally for decoding and deserialization.
 * @param buffer_len the length of the buffer.
 * @param ctx [OUT] the context created.
 *
 * @retval MESSAGE_PROTOCOL_OK Initialization successful.
 * @retval MESSAGE_PROTOCOL_ERROR_ARGS The buffer provided or context pointer is invalid.
 * @retval MESSAGE_PROTOCOL_ERROR_BUF_TOO_SMALL The buffer provided is too small.
 */
mp_status message_protocol_init(uint8_t* buffer, const uint16_t buffer_len, mp_context* ctx)
{
    mp_context_internal* int_ctx = NULL;

    if(NULL == buffer || NULL == ctx)
    {
        return MESSAGE_PROTOCOL_ERROR_ARGS;
    }

    // Check for minimum size.
    if((MESSAGE_PROTOCOL_WORKBUF_SIZE_MIN + sizeof(mp_context_internal)) > buffer_len)
    {
        return MESSAGE_PROTOCOL_ERROR_BUF_TOO_SMALL;
    }
    int_ctx = (mp_context_internal*)buffer;

    int_ctx->work_buf = buffer + sizeof(mp_context_internal);
    int_ctx->work_buf_len = buffer_len - sizeof(mp_context_internal);
    int_ctx->work_buf_idx = 0;
    int_ctx->work_buf_crc = 0;
    int_ctx->parser_state = STATE_IDLE;

    *ctx = (mp_context)int_ctx;

    return MESSAGE_PROTOCOL_OK;
}


/**
 * @brief Pushes one byte of received message data into the internal message parser.
 *
 * @param ctx message protocol context.
 * @param message_byte received message data byte
 * @retval MESSAGE_PROTOCOL_OK : Indicates the message byte has been processed.
 * @retval MESSAGE_PROTOCOL_MESSAGE_READY Indicates the message byte has been processed, and a message is ready and can be retrieved.
 * @retval MESSAGE_PROTOCOL_ERROR_ARGS Invalid context.
 * @retval MESSAGE_PROTOCOL_ERROR_INDEX Indicates the sequence of message bytes at this point is invalid. The message bytes are discarded,
 *                                      and the message parser is reset.
 * @retval MESSAGE_PROTOCOL_ERROR_CHECKSUM Indicates a message is ready but its checksum is invalid. The message bytes are discarded,
 *                                         and the message parser is reset.
 * @retval MESSAGE_PROTOCOL_ERROR_BUF_TOO_SMALL Indicates the internal work buffer is too small to continue processing more bytes. The message
 *                                              bytes are discarded, and the message parser is reset.
 * @retval MESSAGE_PROTOCOL_ERROR_UNKOWN_STATE Indicates an unexpected internal state. A bug is at work here. This error is not recoverable.
 */
mp_status message_protocol_push_data(mp_context ctx, uint8_t message_byte)
{
    mp_context_internal* int_ctx = NULL;

    if(NULL == ctx)
    {
        return MESSAGE_PROTOCOL_ERROR_ARGS;
    }
    int_ctx = (mp_context_internal*)ctx;

    switch(int_ctx->parser_state)
    {
        case STATE_IDLE:
            if (message_byte == MESSAGE_PROTOCOL_ESC)
            {
                // If we're receiving an escape char while waiting for resync,
                // drop next character (we don't want to sync on an escaped char)
                int_ctx->parser_state = STATE_IDLE_ESCAPE_DATA;
            }
            else if (message_byte == MESSAGE_PROTOCOL_SOF)
            {
                int_ctx->work_buf_idx = 0;
                int_ctx->work_buf_crc = CRC16_CCITT_SEED_VALUE;
                int_ctx->parser_state = STATE_WAIT_DATA;
            }
            break;

        case STATE_IDLE_ESCAPE_DATA:
            int_ctx->parser_state = STATE_IDLE;
            break;

        case STATE_WAIT_DATA:
            if (message_byte == MESSAGE_PROTOCOL_ESC)
            {
                int_ctx->parser_state = STATE_ESCAPE_DATA;
            }
            else if (message_byte == MESSAGE_PROTOCOL_SOF)
            {
                int_ctx->work_buf_crc = CRC16_CCITT_SEED_VALUE;
                int_ctx->work_buf_idx = 0;
            }
            else if (message_byte == MESSAGE_PROTOCOL_EOF)
            {
                // If we received an EOF, at least *SOF, CMD, ID, STATUS and CRC(2) must
                // have been received already. * is not accounted in the index.
                if (int_ctx->work_buf_idx < 5)
                {
                    int_ctx->parser_state = STATE_IDLE;
                    return MESSAGE_PROTOCOL_ERROR_INDEX;
                }

                if (int_ctx->work_buf_crc == 0)
                {
                    int_ctx->parser_state = STATE_MESSAGE_READY;
                    return MESSAGE_PROTOCOL_MESSAGE_READY;
                }
                else
                {
                    int_ctx->parser_state = STATE_IDLE;
                    return MESSAGE_PROTOCOL_ERROR_CHECKSUM;
                }
            }
            else
            {
                if (int_ctx->work_buf_idx >= int_ctx->work_buf_len)
                {
                    int_ctx->parser_state = STATE_IDLE;
                    return MESSAGE_PROTOCOL_ERROR_BUF_TOO_SMALL;
                }
                int_ctx->work_buf[int_ctx->work_buf_idx++] = message_byte;
                int_ctx->work_buf_crc = crc16_ccitt_update(int_ctx->work_buf_crc, message_byte);
            }
            break;

        case STATE_ESCAPE_DATA:
            if (int_ctx->work_buf_idx >= int_ctx->work_buf_len)
            {
                int_ctx->parser_state = STATE_IDLE;
                return MESSAGE_PROTOCOL_ERROR_BUF_TOO_SMALL;
            }
            int_ctx->work_buf[int_ctx->work_buf_idx++] = message_byte;
            int_ctx->work_buf_crc = crc16_ccitt_update(int_ctx->work_buf_crc, message_byte);
            int_ctx->parser_state = STATE_WAIT_DATA;
            break;

        case STATE_MESSAGE_READY:
            // message_byte is lost.
            return MESSAGE_PROTOCOL_MESSAGE_READY;

        // Not supposed to happen
        default:
            // This condition is not recoverable.
            // Something outside has corrupted our memory.
            // Check for buffer overflows.
            return MESSAGE_PROTOCOL_ERROR_UNKOWN_STATE;
    }

    return MESSAGE_PROTOCOL_OK;
}


/**
 * @brief Retrieves the message received.
 *
 * @param ctx message protocol context.
 * @param message will contain the received message.
 *
 * @retval MESSAGE_PROTOCOL_OK Indictes the message was retrieved successfully.
 * @retval MESSAGE_PROTOCOL_ERROR_ARGS Invalid context or message parameter.
 * @retval MESSAGE_PROTOCOL_WARN_NOT_AVAILABLE No message is ready to be retrieved.
 * @retval MESSAGE_PROTOCOL_ERROR_INDEX The message is invalid. Should not occur.
 */
mp_status message_protocol_retrieve(mp_context ctx, mp_message* const message)
{
    mp_context_internal* int_ctx = NULL;

    if(NULL == ctx || NULL == message)
    {
        return MESSAGE_PROTOCOL_ERROR_ARGS;
    }
    int_ctx = (mp_context_internal*)ctx;

    if (STATE_MESSAGE_READY != int_ctx->parser_state)
    {
        return MESSAGE_PROTOCOL_WARN_NOT_AVAILABLE;
    }

    int_ctx->parser_state = STATE_IDLE;

    message->cmd_id = int_ctx->work_buf[0];
    message->seq_nr = int_ctx->work_buf[1];
    message->status = int_ctx->work_buf[2];
    message->data = &int_ctx->work_buf[3];
    message->data_len = int_ctx->work_buf_idx-5;    // Remove: ID, SEQ_NR, STATUS, CRC[2]

    return MESSAGE_PROTOCOL_OK;
}


/**
 * @brief This function serializes the specified message in the buffer provided.
 *        Message header, CRC and footer are attached.
 *        Upon success, 'buffer' is ready to be sent to the other endpoint using UART.
 *
 * @param message the message to serialize in the buffer.
 * @param buffer where the message will be serialized in.
 * @param buffer_len length of the buffer provided. Used to prevent buffer overflow.
 * @param buffer_used_len [OUT]length of the buffer used for serialization.
 *
 * @retval MESSAGE_PROTOCOL_OK Serialization successful.
 * @retval MESSAGE_PROTOCOL_ERROR_ARGS Invalid input parameters.
 * @retval MESSAGE_PROTOCOL_ERROR_BUF_TOO_SMALL Buffer provided is too small to serialize message.
 */
mp_status message_protocol_serialize(const mp_message* message, uint8_t* buffer, uint16_t buffer_len, uint16_t* buffer_used_len)
{
    uint16_t buf_idx  = 0;
    uint16_t dat_idx  = 0;
    uint16_t crc;
    mp_status encode_status;


    if(NULL == message  ||  NULL == buffer  ||  NULL == buffer_used_len)
    {
        return MESSAGE_PROTOCOL_ERROR_ARGS;
    }

    if(MESSAGE_PROTOCOL_SERIALIZE_BUF_SIZE_MIN > buffer_len)
    {
        return MESSAGE_PROTOCOL_ERROR_BUF_TOO_SMALL;
    }

    // SOF
    buffer[buf_idx++] = MESSAGE_PROTOCOL_SOF;

    // Command Id
    encode_status = encode_byte(message->cmd_id, buffer, &buf_idx, buffer_len);
    if(MESSAGE_PROTOCOL_OK != encode_status)
    {
        return encode_status;
    }
    crc = crc16_ccitt_update(CRC16_CCITT_SEED_VALUE, message->cmd_id);

    // Sequence Number
    encode_status = encode_byte(message->seq_nr, buffer, &buf_idx, buffer_len);
    if(MESSAGE_PROTOCOL_OK != encode_status)
    {
        return encode_status;
    }
    crc = crc16_ccitt_update(crc, message->seq_nr);

    // Status
    encode_status = encode_byte(message->status, buffer, &buf_idx, buffer_len);
    if(MESSAGE_PROTOCOL_OK != encode_status)
    {
        return encode_status;
    }
    crc = crc16_ccitt_update(crc, message->status);

    // Data payload
    for(dat_idx=0; dat_idx < message->data_len; dat_idx++)
    {
        encode_status = encode_byte(message->data[dat_idx], buffer, &buf_idx, buffer_len);
        if(MESSAGE_PROTOCOL_OK != encode_status)
        {
            return encode_status;
        }
        crc = crc16_ccitt_update(crc, message->data[dat_idx]);
    }

    // CRC
    encode_status = encode_byte((crc >> 8) & 0xFF, buffer, &buf_idx, buffer_len);
    if(MESSAGE_PROTOCOL_OK != encode_status)
    {
        return encode_status;
    }
    encode_status = encode_byte(crc & 0xFF, buffer, &buf_idx, buffer_len);
    if(MESSAGE_PROTOCOL_OK != encode_status)
    {
        return encode_status;
    }

    // EOF
    if(buf_idx >= buffer_len)
    {
        return MESSAGE_PROTOCOL_ERROR_BUF_TOO_SMALL;
    }
    buffer[buf_idx++] = MESSAGE_PROTOCOL_EOF;

    *buffer_used_len = buf_idx;

    return MESSAGE_PROTOCOL_OK;
}


/**
 * @brief Performs a deep copy of the message.
 *        NOTE: clone->data MUST point to a buffer of at least 'from->data_len' size.
 *
 * @param from the copy source
 * @param clone the copy destination
 * @param clone_data_len length of the buffer at clone->data. This is used to prevent a buffer overflow.
 *
 * @retval MESSAGE_PROTOCOL_OK The cloning was successful.
 * @retval MESSAGE_PROTOCOL_ERROR_ARGS Invalid arguments to function.
 */
mp_status message_protocol_clone(const mp_message* from, mp_message* const clone, uint16_t clone_data_len)
{
    uint16_t dat_idx;

    if(NULL == from  ||  NULL == clone  ||  clone_data_len < from->data_len)
    {
        return MESSAGE_PROTOCOL_ERROR_ARGS;
    }

    clone->cmd_id = from->cmd_id;
    clone->seq_nr = from->seq_nr;
    clone->status = from->status;
    clone->data_len = from->data_len;
    for(dat_idx=0; dat_idx<clone->data_len; dat_idx++)
    {
        clone->data[dat_idx] = from->data[dat_idx];
    }

    return MESSAGE_PROTOCOL_OK;
}

/**
 * @}   // defgroup api_message_protocol
 */
