/** ********************************************************************************************************************
 * @file        platform.c
 * @author      Briowireless, Inc.
 * @date        2017-01-03
 * @brief       Implementation of the required functions to allow running the SDK and samples on Linux.
 * @version     LINUX V1
 * @copyright   Briowireless, Inc., 2017, Licensed under BSD license. See LICENSE file for more details.
 *
 ** *******************************************************************************************************************/

// Include directives  -------------------------------------------------------------------------------------------------

#include <../platform.h>

#include <errno.h>
#include <fcntl.h> 
#include <string.h>
#include <termios.h>
#include <unistd.h>


// Preprocessor definitions --------------------------------------------------------------------------------------------

// Local variables -----------------------------------------------------------------------------------------------------

static int g_fdesc;


// Local functions -----------------------------------------------------------------------------------------------------

static int set_interface_attribs (int fd, int speed, int parity)
{
    struct termios tty;
    memset (&tty, 0, sizeof tty);
    if (tcgetattr (fd, &tty) != 0)
    {
        P_ERR("error %d from tcgetattr", errno);
        return -1;
    }

    cfsetospeed (&tty, speed);
    cfsetispeed (&tty, speed);

    tty.c_cflag = (tty.c_cflag & ~CSIZE) | CS8;     // 8-bit chars
    // disable IGNBRK for mismatched speed tests; otherwise receive break
    // as \000 chars
    tty.c_iflag &= ~IGNBRK;         // disable break processing
    tty.c_lflag = 0;                // no signaling chars, no echo,
                                    // no canonical processing
    tty.c_oflag = 0;                // no remapping, no delays
    tty.c_cc[VMIN]  = 0;            // read doesn't block
    tty.c_cc[VTIME] = 5;            // 0.5 seconds read timeout

    tty.c_iflag &= ~(IXON | IXOFF | IXANY); // shut off xon/xoff ctrl

    tty.c_cflag |= (CLOCAL | CREAD);// ignore modem controls,
                                    // enable reading
    tty.c_cflag &= ~(PARENB | PARODD);      // shut off parity
    tty.c_cflag |= parity;
    tty.c_cflag &= ~CSTOPB;
    tty.c_cflag &= ~CRTSCTS;

    if (tcsetattr (fd, TCSANOW, &tty) != 0)
    {
        P_ERR("error %d from tcsetattr", errno);
        return -1;
    }
    return 0;
}

static int set_blocking (int fd, int should_block)
{
    struct termios tty;
    memset (&tty, 0, sizeof tty);
    if (tcgetattr (fd, &tty) != 0)
    {
        P_ERR("error %d from tggetattr", errno);
        return -1;
    }

    tty.c_cc[VMIN]  = should_block ? 1 : 0;
    tty.c_cc[VTIME] = 5;            // 0.5 seconds read timeout

    if (tcsetattr (fd, TCSANOW, &tty) != 0)
    {
        P_ERR("error %d setting term attributes", errno);
        return -1;
    }
    
    return 0;
}


// Global (extern) function  -------------------------------------------------------------------------------------------

/**
 * @defgroup platform_linux Platform Implementation for SDK: Linux
 * Implementation of required functions to support running the SDK and samples on Linux-based machines.
 * @{
 */

/** ********************************************************************************************************************
 * @brief  Initializes the platform internals, and opens the COM prt for communication with the BitPipe.
 *
 * @param  bp_com_port  The name of the COM port (Full path to TTY device)
 *
 * @retval uint8_t      0 indicates success;
 *                      1 indicates the COM port could not be opened.
 *                      2 indicates an error occurred while getting/changing COM port settings.
 *                      3 indicates an error occurred while configuring non-blocking read on COM port.
 */
uint8_t platform_init(const char* bp_com_port)
{
    g_fdesc = open(bp_com_port, O_RDWR | O_NOCTTY | O_SYNC);
    if(g_fdesc < 0)
    {
        P_ERR("error %d opening %s: %s", errno, bp_com_port, strerror(errno));
        return 1;
    }

    if(set_interface_attribs (g_fdesc, B115200, 0) != 0)    // set speed to 115,200 bps, 8n1 (no parity)
    {
        return 2;
    }
    
    if(set_blocking (g_fdesc, 0) != 0)                      // set no blocking
    {
        return 3;
    }

    return 0;
}

/** ********************************************************************************************************************
 * @brief  Closes and releases resources obtained during the platform_init call.
 *
 */
void platform_deinit(void)
{
    close(g_fdesc);
}

/** ********************************************************************************************************************
 * @brief  Closes and releases resources obtained during the platform_init call.
 *
 * @param  data     Binary data to write to the COM port.
 * @param  data_len Length of the binary data to write to the COM port.
 *
 */
void uart_write(const uint8_t* data, uint16_t data_len)
{
    int written_total = 0;
    int written;
    
    while(written_total < data_len)
    {
        written = write(g_fdesc, data + written_total, data_len - written_total);
        if(written == -1)
        {
            P_ERR("Error in write: errno=%d\n", errno);
            return;
        }

        written_total += written;
    }

    /* Uncomment this to get printout of the bytes written to UART.
    uint16_t i;
    P_DEBUG(" >> ");
    for(i=0; i<data_len; i++)
        P_DEBUG("%02X ", data[i]);
    P_DEBUG("\n");
    //*/
}

/** ********************************************************************************************************************
 * @brief  Closes and releases resources obtained during the platform_init call.
 *
 * @param  buffer     buffer where to write the bytes received from the COM port.
 * @param  buffer_len Length of the buffer above.
 *
 */
uint16_t uart_read(uint8_t* buffer, uint16_t buffer_len)
{
    int n_read;
    
    n_read = read(g_fdesc, buffer, buffer_len);
    if(n_read == -1)
    {
        P_ERR("Error in read: errno=%d\n", errno);
        return 0;
    }

    return (uint16_t)n_read;
}

/**
 * @}       // Doxygen platform_linux defgroup
 */
