/** ********************************************************************************************************************
 * @file        platform.h
 * @author      Briowireless, Inc.
 * @date        2016-11-02
 * @brief       Platform file definition for BitPipe SDK.
 *
 * To support your platform, you should implement the following functions in your project.
 * For example platform support, check out the *win32* platform. It is very straight-forward.
 *
 * Only the function @ref uart_write() is used internally in the SDK. The other functions will allow the samples
 *  to run out-of the box; they are not required to be implemented as-is.
 *
 * @version     API V1
 * @copyright   Briowireless, Inc., 2016, Licensed under BSD license. See LICENSE file for more details.
 *
 ** *******************************************************************************************************************/
#ifndef __PLATFORM_H__
#define __PLATFORM_H__


// Include directives  -------------------------------------------------------------------------------------------------

#include <platform_defs.h>


// Public Functions ----------------------------------------------------------------------------------------------------

/**
 * @defgroup platform_api Platform Implementation for SDK
 * Defines the functions to be implemented to support your platform.
 *
 * Only the function @ref uart_write() is used internally in the SDK. The other functions will allow the samples
 *  to run out-of the box; they are not required to be implemented as-is.
 * @{
 */

/**
 * @brief This function should open the UART/COM port to communicate with the BitPipe, set port settings, and make the
 *        read operation non-blocking.
 *
 * @param bp_com_port   The COM port (or UART device) name where the BitPipe is plugged. This is the device that will
 *                      be used to communicate using the binary serial protocol with the BitPipe.
 *
 * @return 0 indicates success, any other value indicate an error.
 */
uint8_t platform_init(const char* bp_com_port);

/**
 * @brief Releases any resources acquired and opened in the platform_init() function.
 */
void platform_deinit(void);

/**
 * @brief Writes data to the UART/COM port, to the BitPipe.
 *
 * @note  This function is not required to block until all data is sent. However, keep in mind that there are no retry
 *        mechanism in the SDK in case an error occurs during write.
 *
 * @param data      The binary data to send to the BitPipe. Should be a message formatted by the message_protocol low-level API.
 * @param data_len  Length pf the data to send.
 */
void uart_write(const uint8_t* data, uint16_t data_len);

/**
 * @brief Reads any data sent from the BitPipe to this endpoint. If no data has been received, this function should not block.
 * @param buffer    Where to store the bytes received.
 * @param buffer_len The length of the buffer supplied to store the bytes received.
 *
 * @return The actual number of bytes received, which can be pushed to @ref bp_api_process_incoming().
 */
uint16_t uart_read(uint8_t* buffer, uint16_t buffer_len);

/**
 * @}   // Doxygen defgroup platform_api
 */

#endif  // __PLATFORM_H__
