/** ********************************************************************************************************************
 * @file        platform.c
 * @author      Briowireless, Inc.
 * @date        2016-11-02
 * @brief       Implementation of the required functions to allow running the SDK and samples on Windows.
 * @version     WIN32 V1
 * @copyright   Briowireless, Inc., 2016, Licensed under BSD license. See LICENSE file for more details.
 *
 ** *******************************************************************************************************************/

// Include directives  -------------------------------------------------------------------------------------------------

#include <../platform.h>
#include <windows.h>


// Preprocessor definitions --------------------------------------------------------------------------------------------

// Local variables -----------------------------------------------------------------------------------------------------

static HANDLE hComPort = INVALID_HANDLE_VALUE;          ///< COM port object system handle.


// Global (extern) function  -------------------------------------------------------------------------------------------

/**
 * @defgroup platform_win32 Platform Implementation for SDK: Win32
 * Implementation of required functions to support running the SDK and samples on Windows-based machines.
 * @{
 */

/** ********************************************************************************************************************
 * @brief  Initializes the platform internals, and opens the COM prt for communication with the BitPipe.
 *
 * @param  bp_com_port  The name of the COM port (Windows-style "\\.\COM<id>")
 *
 * @retval uint8_t      0 indicates success;
 *                      1 indicates the COM port could not be opened.
 *                      2 and 3 indicates an error occurred while getting/changing COM port settings.
 *                      4 indicates an error occurred while configuring non-blocking read on COM port.
 */
uint8_t platform_init(const char* bp_com_port)
{
    // Open the COM port.
    hComPort = CreateFileA(bp_com_port,
                          GENERIC_READ | GENERIC_WRITE,
                          0,
                          NULL,
                          OPEN_EXISTING,
                          0,
                          NULL);
    if(INVALID_HANDLE_VALUE == hComPort)
    {
        DWORD dwErr = GetLastError();
        P_ERR("platform_init: Failed to open COM port %s. LastError=%ld\n", bp_com_port, dwErr);

        // Provide info for most common errors:
        if(ERROR_ACCESS_DENIED == dwErr)
        {
            P_ERR("ACCESS DENIED (Is the port already open in another application?)\n");
        }
        else if(ERROR_FILE_NOT_FOUND == dwErr)
        {
            P_ERR("FILE NOT FOUND (Are you sure about the name of the COM port? On Windows, the COM port name must be \"\\\\.\\COM<instance>\")\n");
        }

        return 1;
    }


    // Make sure to target BitPipe COM settings (115200bps, 8-N-1 by default).
    DCB serialParams;
    serialParams.DCBlength = sizeof(serialParams);
    if(GetCommState(hComPort, &serialParams) == FALSE)
    {
        platform_deinit();

        P_ERR("platform_init: Failed to get communication state on COM port. LastError=%ld\n", GetLastError());
        return 2;
    }

    serialParams.BaudRate = CBR_115200;
    serialParams.ByteSize = 8;
    serialParams.StopBits = ONESTOPBIT;
    serialParams.Parity   = NOPARITY;

    if(SetCommState(hComPort, &serialParams) == FALSE)
    {
        platform_deinit();

        P_ERR("platform_init: Failed to set communication state on COM port. LastError=%ld\n", GetLastError());
        return 3;
    }


    // Make the 'ReadFile' function non-blocking, in case no data is available.
    // Make the 'WriteFile' function blocking until all data is written.
    COMMTIMEOUTS timeouts = { 0 };
    timeouts.ReadIntervalTimeout = MAXDWORD;
    timeouts.ReadTotalTimeoutConstant = 1;
    timeouts.ReadTotalTimeoutMultiplier = 0;
    timeouts.WriteTotalTimeoutConstant = 0;
    timeouts.WriteTotalTimeoutMultiplier = 0;
    if(SetCommTimeouts(hComPort, &timeouts) == FALSE)
    {
        platform_deinit();

        P_ERR("platform_init: Failed to set communication timeouts on COM port. LastError=%ld\n", GetLastError());
        return 4;
    }

    return 0;
}

/** ********************************************************************************************************************
 * @brief  Closes and releases resources obtained during the platform_init call.
 *
 */
void platform_deinit(void)
{
    CloseHandle(hComPort);
}

/** ********************************************************************************************************************
 * @brief  Closes and releases resources obtained during the platform_init call.
 *
 * @param  data     Binary data to write to the COM port.
 * @param  data_len Length of the binary data to write to the COM port.
 *
 */
void uart_write(const uint8_t* data, uint16_t data_len)
{
    DWORD written_total = 0;
    DWORD written;
    while(written_total < data_len)
    {
        if(WriteFile(hComPort, data + written_total, data_len - written_total, &written, NULL) == FALSE)
        {
            P_ERR("Error in WriteFile. LastError: %ld\n", GetLastError());
            return;
        }
        written_total += written;
    }

    /* Uncomment this to get printout of the bytes written to UART.
    uint16_t i;
    P_DEBUG(" >> ");
    for(i=0; i<data_len; i++)
        P_DEBUG("%02X ", data[i]);
    P_DEBUG("\n");
    //*/
}

/** ********************************************************************************************************************
 * @brief  Closes and releases resources obtained during the platform_init call.
 *
 * @param  buffer     buffer where to write the bytes received from the COM port.
 * @param  buffer_len Length of the buffer above.
 *
 */
uint16_t uart_read(uint8_t* buffer, uint16_t buffer_len)
{
    DWORD read;
    if(ReadFile(hComPort, buffer, buffer_len, &read, NULL) == FALSE)
    {
        P_ERR("Error in ReadFile. LastError: %ld\n", GetLastError());
    }

    return (uint16_t)read;
}

/**
 * @}       // Doxygen platform_win32 defgroup
 */
